/* Copyright (C) 2006 Pierre Nicolas
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "mutmodel.h"
#include "coalnode.h"

#include <iostream>
#include <gsl/gsl_randist.h>

using namespace std;

int Coalnode::_count = 0;
int Coalnode::_id = 0;

Coalnode::Coalnode(int node_status, Coalnode* prev, Coalnode* next, 
		   Coalnode* child_right, Coalnode* child_left,
		   Coalnode* ancestor, double time, int lseq,
		   Mutmodel **mut_model, gsl_rng* rng)
{
  _node_id = _id;

  _count++;
  _id++;
  
  _node_status = node_status;
  _prev = prev;
  _next = next;
  _child_right = child_right;
  _child_left = child_left;
  _ancestor = ancestor;
  _time = time;
  _lseq = lseq; 
  _mut_model = mut_model;
  _nmodal = (*_mut_model)->nmodal();
  _seq = new int [_lseq];
  _seqbis = new int [_lseq];
  _rng = rng;
  _nb_mutclass = (*_mut_model)->nbmutclass();

  if (_ancestor != NULL)
    {
      
    }
  if (_child_right != NULL)
    {
      _child_right->_ancestor = this;
    }
  if (_child_left != NULL)
    {
      _child_left->_ancestor = this;      
    }
  
  if (_prev != NULL) 
    {
      _prev->_next = this;
    }
  if (_next != NULL) 
    {
      _next->_prev = this;
    }

  _posterior_probas_mutclass = new double [_nb_mutclass];
  _probatrans = new double** [_nb_mutclass];
  for (int imut=0 ; imut < _nb_mutclass ; imut++)
    {
      _probatrans[imut] = new double* [_nmodal];
      for (int i=0 ; i<_nmodal ; i++) 
	{
	  _probatrans[imut][i] = new double [_nmodal];
	}
    }
  
  _probaswitch = new double* [_nb_mutclass];
  for (int imut=0 ; imut < _nb_mutclass ; imut++)
    {
      _probaswitch[imut] = new double [_nb_mutclass];
    }

  _posterior_probas = new double** [_nb_mutclass];
  for (int imut=0 ; imut < _nb_mutclass ; imut++)
    {
      _posterior_probas[imut] = new double* [_lseq];
      for (int i=0 ; i<_lseq ; i++) 
	{
	  _posterior_probas[imut][i] = new double [_nmodal];
	}
    }
  
  _peeling_seq = new double* [_lseq];
  for (int i=0 ; i<_lseq ; i++) 
    {
      _peeling_seq[i] = new double [_nmodal];
    }
  
  _peeling_mutclass = new double [_nb_mutclass];

  //cout << "*-> new node created id: " << _node_id << " time: " << _time << endl;
  if (node_status == 0) 
    {
      //cout << "leaf node ";
    }
  if (node_status == 1) 
    {
      //cout << "internal node ";
    }
  if (node_status == 2) 
    {
      //cout << "root node ";
    }
  if (node_status == 1) 
    {
      //cout << "childs left " << _child_left->_node_id  << " right " << _child_right->_node_id;
    }
  if (node_status == 2) 
    {
      //cout << "childs left " << _child_left->_node_id;
    }
  //cout << endl;

  _countseq = new int* [_lseq];
  for (int i=0 ; i<_lseq ; i++) 
    {
      _countseq[i] = new int [_nmodal];
    }
  _countmut = new int [_nb_mutclass];

  _contextdep = (*_mut_model)->contextdep();
}


Coalnode::~Coalnode()
{
  _count--;
  delete _seq;
  delete _seqbis;
  for (int imut=0; imut<_nb_mutclass; imut++)
    {
      for (int i=0; i<_lseq; i++)
	{
	  delete _posterior_probas[imut][i];
	}
      delete _posterior_probas[imut];

      for (int i=0 ; i<_nmodal ; i++) 
	{
	  delete _probatrans[imut][i];
	}
      delete _probatrans[imut];

      delete _probaswitch[imut];
    }
  delete _posterior_probas;

  for (int i=0 ; i<_lseq ; i++) 
    {
      delete _peeling_seq[i];
    }
  delete _peeling_seq;
  delete _peeling_mutclass;
  
  delete _probatrans;
  delete _probaswitch;
  delete _posterior_probas_mutclass;
  
  for (int i=0 ; i<_lseq ; i++) 
    {
      delete _countseq[i];
    }
  delete _countseq;
  delete _countmut;
}

void Coalnode::resetCountSeq()
{
  for (int ipos=0 ; ipos<_lseq ; ipos++) 
    {
      for (int i=0; i<_nmodal; i++)
	{
	  _countseq[ipos][i] = 0;
	}
    }
  for (int imut=0 ; imut<_nb_mutclass ; imut++)
    {
      _countmut[imut] = 0;
    }
  _ncounts = 0;  
  if (_next != NULL)
    {
      _next->resetCountSeq();
    }
}

void Coalnode::countSeq()
{
  for (int ipos=0 ; ipos<_lseq ; ipos++) 
    {
      _countseq[ipos][_seq[ipos]]++;      
    }
  _countmut[_mut_class]++;
  _ncounts++;  
  if (_next != NULL)
    {
      _next->countSeq();
    }
}

void Coalnode::printCountSeq()
{
  cout << "Node " << _node_id << " post_proba_mut";
  for (int imut=0 ; imut<_nb_mutclass ; imut++)
    {
      cout << " " << (double)_countmut[imut]/(double)_ncounts;
    }
  cout << endl;
  cout << "post_proba_seq" << endl;
  for (int ipos=0 ; ipos<_lseq ; ipos++) 
    {
      cout << "ipos " << ipos;
      for (int i=0; i<_nmodal ; i++)
	{
	  cout << " " << (double)_countseq[ipos][i]/(double)_ncounts;
	}
      cout << endl;
    }
  if (_next != NULL)
    {
      _next->printCountSeq();
    }
}



void Coalnode::setiseq(int iseq, int* obs_seq)
{
  if (_node_status != 0) 
    {
      cerr << "Pb in void Coalnode::setiseq(int iseq)" << endl;
      exit(1);
    }
  _iseq = iseq;
  _obs_seq = obs_seq;
}

void Coalnode::setseq(int* seq)
{
  for (int i=0; i<_lseq; i++)
    {
      _seq[i] = seq[i];
    }
}

int Coalnode::iseq()
{
  return _iseq;
}

void Coalnode::moveafter(Coalnode* previous_node)
{
  //cout << "moveafter" << endl;
  if (this->_prev != NULL) 
    {
      this->_prev->_next = this->_next;
    }
  if (this->_next != NULL) 
    {
      this->_next->_prev = this->_prev;
    }
  if (previous_node->_next != NULL)
    {
      previous_node->_next->_prev = this;
    }
  this->_next = previous_node->_next;
  this->_prev = previous_node;
  previous_node->_next = this; 
}

void Coalnode::movebefore(Coalnode* next_node)
{
  //cout << "movebefore" << endl;
  if (this->_prev != NULL) 
    {
      this->_prev->_next = this->_next;
    }
  if (this->_next != NULL) 
    {
      this->_next->_prev = this->_prev;
    }
  if (next_node->_prev != NULL)
    {
      next_node->_prev->_next = this;
    }
  this->_prev = next_node->_prev;
  this->_next = next_node;
  next_node->_prev = this; 
}

void Coalnode::increaseNbTermStemCell()
{
  if (this->_ancestor != this->_next) 
    {
      cout << "Pb 1 in increaseNbTermStemCell()" << endl;
      exit(1);
    }
  if (_next->_child_left != this) 
    {
      cout << "Pb 2 in increaseNbTermStemCell()" << endl;
      exit(1);
    }

  _next->_child_left = this->_child_right;
  this->_child_right = NULL;
  this->_ancestor = NULL;
  _next->_child_left->_ancestor = _next;
  this->_time = _next->_time;
  this->_node_status = 2;
}

void Coalnode::decreaseNbTermStemCell(double new_coaltime)
{
  this->_time = new_coaltime;
  this->_node_status = 1;
  this->_ancestor = this->_next;
  this->_child_right = _next->_child_left;
  _next->_child_left = this;
  this->_child_right->_ancestor = this;
}

void Coalnode::increaseNbInitStemCell(Coalnode* &node_collector, double coaltime)
{
  Coalnode* node1 = this;
  Coalnode* node2;
  Coalnode* node3;
  Coalnode* node4 = this->_next;
  
  if (node_collector != NULL) 
    {
      node2 = node_collector;
      node3 = node2->_next;
      node_collector = node3->_next;
    } else {
      node2 = new Coalnode(-1, NULL, NULL,
			   NULL, NULL,
			   NULL, -1, this->_lseq, this->_mut_model, this->_rng);
      node3 = new Coalnode(-1, NULL, NULL,
			   NULL, NULL,
			   NULL, -1, this->_lseq, this->_mut_model, this->_rng);
    }
  
  node1->_next = node2;
  node2->_prev = node1;
  node2->_next = node3;
  node3->_prev = node2;
  node3->_next = node4;
  node4->_prev = node3;
    
  for (int ipos=0 ; ipos<_lseq ; ipos++)
    {
      node2->_seq[ipos] = node1->_seq[ipos];
    }
  node2->_iseq = node1->_iseq;
  node2->_node_status = 0;
  node2->_ancestor = node3;
  node2->_child_right = NULL;
  node2->_child_left = NULL;
  node2->_time = this->_time;
  node2->_mut_class = this->_mut_class;

  node3->_node_status = 1;
  node3->_ancestor = node1->_ancestor;
  node3->_ancestor->_child_left = node3;
  node3->_child_left = node1;
  node3->_child_right = node2;
  node3->_time = coaltime;

  node1->_ancestor = node3;
}

void Coalnode::decreaseNbInitStemCell(Coalnode* &node_collector)
{
  Coalnode* node1 = this;
  Coalnode* node2 = this->_next;
  Coalnode* node3 = node2->_next;
  Coalnode* node4 = node3->_next;

  // take node 2 & 3 and put them in the node collector
  node3->_next = node_collector;
  node_collector = node2;

  // rewiring
  node1->_next = node4;
  node4->_prev = node1;
  node1->_ancestor = node3->_ancestor;
  node1->_ancestor->_child_left = node1;  
}


void Coalnode::printInfo()
{
  cout << "node_id " << _node_id << " node_status " << _node_status << " time " << _time << "_mut_class " << _mut_class << endl;
  if (_node_status == 0)
    {
      cout << "_ancestor " <<  _ancestor->_node_id << endl;
      cout << "_iseq " <<  _iseq << endl;
      for (int ipos=0; ipos<_lseq ; ipos++)
	{
	  cout << _seq[ipos];
	}
      cout << endl;
    }
  if (_node_status == 1)
    {
      cout << "_ancestor " <<  _ancestor->_node_id << endl;
      cout << "_child_left " <<  _child_left->_node_id 
	   << " _child_right " <<  _child_right->_node_id << endl;
      cout << "_iseq " <<  _iseq << endl;
      for (int ipos=0; ipos<_lseq ; ipos++)
	{
	  cout << _seq[ipos];
	}
      cout << endl;
    }
  if (_node_status == 2)
    {
      cout << "_child_left " <<  _child_left->_node_id << endl;
      cout << "_iseq " <<  _iseq << endl;
      for (int ipos=0; ipos<_lseq ; ipos++)
	{
	  cout << _seq[ipos];
	}
      cout << endl;
    }
}

void Coalnode::computeProbaTrans()
{
  if (_node_status == 2)
    {
      cerr << "void Coalnode::computeProbaTrans() _node_status==2" << endl;
      exit(1);
    }
  
  double branch_length = (*_mut_model)->computeBranchLength(_time,_ancestor->_time);
  //cout << "branch_length " << branch_length << endl;
  for (int imut=0; imut<_nb_mutclass; imut++)
    {
      for (int jmut=0; jmut<_nb_mutclass; jmut++)
	{
	  _probaswitch[imut][jmut] 
	    = (*_mut_model)->probaswitch(branch_length, imut, jmut);
	  //cout << "_probaswitch[" << imut << "][" << jmut <<"] " << _probaswitch[imut][jmut] << " ";
	}
      //cout << endl;
      for (int i=0 ; i<_nmodal ; i++) 
	{
	  for (int j=0 ; j<_nmodal ; j++) 
	    {	      
	      _probatrans[imut][i][j] 
		= (*_mut_model)->proba(branch_length, imut, i, j);
	      //cout << "_probatrans["<<imut<<"]["<<i<<"]["<<j<<"] " << _probatrans[imut][i][j] << " ";
	    }
	  //cout << endl;
	}
    }
}


void Coalnode::drawSeq()
{
  if (_node_status == 2) 
    {
      for (int ipos=0 ; ipos<_lseq ; ipos++)
	{
	  _seq[ipos] = 0;
	}
      _mut_class = 0;
    } else if (!_contextdep) {  
      this->computeProbaTrans();
      if (_nb_mutclass>1)
	{
	  double rm = gsl_ran_flat(_rng, 0, 1);
	  _mut_class = 0;
	  double sumpm = _probaswitch[_ancestor->_mut_class][_mut_class];
	  while (sumpm < rm && _mut_class+1<_nb_mutclass) {
	    _mut_class++;
	    sumpm += _probaswitch[_ancestor->_mut_class][_mut_class];
	  }      
	} else {
	  _mut_class = 0;
	}

      for (int ipos=0 ; ipos<_lseq ; ipos++)
	{
	  double r = gsl_ran_flat(_rng, 0, 1);
	  int outcome = 0;
	  double sump = _probatrans[_mut_class][_ancestor->_seq[ipos]][outcome];
	  while (sump < r && outcome<_nmodal) {
	    outcome++;
	    sump += _probatrans[_mut_class][_ancestor->_seq[ipos]][outcome];
	  }
	  _seq[ipos] = outcome;
	}
    } else {
      _mut_class = 0;
      double blength = (*_mut_model)->computeBranchLength(_time,_ancestor->_time);
      (*_mut_model)->contextDepDrawSeq(blength, _ancestor->_seq, _seq);    
    }
}


double Coalnode::likelihoodNode()
{
  double ret_val =1;
  if (_node_status == 1) 
    {
      if (!_contextdep)
	{
	  this->computeProbaTrans();
	  _child_left->computeProbaTrans();
	  _child_right->computeProbaTrans();
	  
	  ret_val *= _probaswitch[_ancestor->_mut_class][_mut_class]
	    *_child_left->_probaswitch[_mut_class][_child_left->_mut_class]
	    *_child_right->_probaswitch[_mut_class][_child_right->_mut_class];
	  for (int ipos=0 ; ipos<_lseq ; ipos++)
	    {
	      ret_val *= _probatrans[_mut_class][_ancestor->_seq[ipos]][_seq[ipos]]
		*_child_left->_probatrans[_child_left->_mut_class][_seq[ipos]][_child_left->_seq[ipos]]
		*_child_right->_probatrans[_child_right->_mut_class][_seq[ipos]][_child_right->_seq[ipos]];
	    }
	} else {
	  double bl = (*_mut_model)->computeBranchLength(_child_left->_time, _time);
	  double br = (*_mut_model)->computeBranchLength(_child_right->_time, _time);
	  double ba = (*_mut_model)->computeBranchLength(_time, _ancestor->_time);
	  ret_val *= (*_mut_model)->probatranscontext(ba, _ancestor->_seq, _seq)
	    *(*_mut_model)->probatranscontext(bl, _seq, _child_left->_seq)
	    *(*_mut_model)->probatranscontext(br, _seq, _child_right->_seq);
	}
    }
  //cout << "likelihoodNode()=" << ret_val << endl;;
  return ret_val;
}

double Coalnode::loglikSeqChildren()
{
  double ret_val =0;

  if (_node_status == 0) 
    {
      int ne = 0;
      int n = 0;
      for (int ipos=0; ipos<_lseq ; ipos++) 
	{
	  if (_obs_seq[ipos] >= 0) 
	    {
	      n++;
	      if (_obs_seq[ipos] == _seq[ipos]) 
		{
		  ne++;
		}
	    }
	}
      ret_val = ne*log((*_mut_model)->perror()) + (n-ne)*log(1-(*_mut_model)->perror());
    } else if (!_contextdep) {
      if (_node_status == 1)
	{
	  _child_left->computeProbaTrans();
	  _child_right->computeProbaTrans();	  
	  ret_val += log(_child_left->_probaswitch[_mut_class][_child_left->_mut_class])
	    + log(_child_right->_probaswitch[_mut_class][_child_right->_mut_class]);
	  for (int ipos=0 ; ipos<_lseq ; ipos++)
	    {
	      ret_val += log(_child_left->_probatrans[_child_left->_mut_class][_seq[ipos]][_child_left->_seq[ipos]])
		+ log(_child_right->_probatrans[_child_right->_mut_class][_seq[ipos]][_child_right->_seq[ipos]]);
	    }      	  
	} else if (_node_status == 2) {
	  _child_left->computeProbaTrans();	  
	  ret_val += log(_child_left->_probaswitch[0][_child_left->_mut_class]);
	  for (int ipos=0 ; ipos<_lseq ; ipos++)
	    {
	      ret_val += log(_child_left->_probatrans[_child_left->_mut_class][0][_child_left->_seq[ipos]]);
	    }      
	} 
    } else {
      if (_node_status == 1)
	{
	  double bl = (*_mut_model)->computeBranchLength(_child_left->_time, _time);
	  double br = (*_mut_model)->computeBranchLength(_child_right->_time, _time);
	  ret_val += log((*_mut_model)->probatranscontext(bl, _seq, _child_left->_seq))
	    +log((*_mut_model)->probatranscontext(br, _seq, _child_right->_seq));
	} else if (_node_status == 2) {
	  double bl = (*_mut_model)->computeBranchLength(_child_left->_time, _time);
	  ret_val += log((*_mut_model)->probatranscontext(bl, _seq, _child_left->_seq));
	}
    }
  return ret_val;  
}


double Coalnode::loglikAncestor()
{
  double ret_val =0;

  if (!_contextdep)
    {
      if (_node_status != 2)
	{
	  this->computeProbaTrans();
	  if (_ancestor->_node_status != 2)
	    {
	      ret_val += log(_probaswitch[_ancestor->_mut_class][_mut_class]);
	      for (int ipos=0 ; ipos<_lseq ; ipos++)
		{
		  ret_val += log(_probatrans[_mut_class][_ancestor->_seq[ipos]][_seq[ipos]]);
		}      
	    } else {
	      ret_val += log(_probaswitch[0][_mut_class]);
	      for (int ipos=0 ; ipos<_lseq ; ipos++)
		{
		  ret_val += log(_probatrans[_mut_class][0][_seq[ipos]]);
		}      
	    }
	}
    } else  {
      if (_node_status != 2)
	{
	  double ba = (*_mut_model)->computeBranchLength(_time, _ancestor->_time);
	  ret_val += log((*_mut_model)->probatranscontext(ba, _ancestor->_seq, _seq));
	}
    }
  return ret_val;  
}

void Coalnode::peelingSeqBackward()
{
  if (_contextdep)
    {
      cerr << "Error: void Coalnode::peelingSeqBackward() is not implemented for the case _contextdep==true" << endl;
      exit(1);
    }
  // Given mutation model (mutclass) at the nodes of the tree
  if (_node_status == 0) 
    {
      for (int ipos=0; ipos<_lseq ; ipos++) 
	{
	  for (int i=0; i<_nmodal; i++) 
	    {
	      if (_seq[ipos]==i) 
		{
		  _peeling_seq[ipos][i]=1;
		} else {
		  _peeling_seq[ipos][i]=0;
		}
	    }
	}      
    } else if (_node_status == 1) {
      _child_left->computeProbaTrans();
      _child_right->computeProbaTrans();      
      for (int ipos=0; ipos<_lseq ; ipos++) 
	{
	  for (int i=0; i<_nmodal; i++) 
	    {
	      double left_term = 0;
	      double right_term = 0;
	      for (int j=0; j<_nmodal; j++) 
		{
		  left_term += _child_left->_probatrans[_child_left->_mut_class][i][j]
		    *_child_left->_peeling_seq[ipos][j];
		  right_term += _child_right->_probatrans[_child_right->_mut_class][i][j]
		    *_child_right->_peeling_seq[ipos][j];
		}
	      _peeling_seq[ipos][i] = left_term*right_term;	      
	    }
	}            
    } else if (_node_status == 2) {
      _child_left->computeProbaTrans();
      for (int ipos=0; ipos<_lseq ; ipos++) 
	{
	  for (int i=0; i<_nmodal; i++) 
	    {
	      double left_term = 0;
	      for (int j=0; j<_nmodal; j++) 
		{
		  left_term += _child_left->_probatrans[_child_left->_mut_class][i][j]
		    *_child_left->_peeling_seq[ipos][j];
		}
	      _peeling_seq[ipos][i] = left_term;	      
	    }
	}            
    }
  if (_next != NULL) 
    {
      _next->peelingSeqBackward();
    }
}

void Coalnode::simulSeqForward()
{
  if (_contextdep)
    {
      cerr << "Error: void Coalnode::simulSeqForward() is not implemented for the case _contextdep==true" << endl;
      exit(1);
    }
  if (_node_status == 2)
    {
      for (int ipos=0; ipos<_lseq; ipos++)
	{
	  _seq[ipos] = 0;	  
	}
    } else if (_node_status == 1) {
      for (int ipos=0; ipos<_lseq; ipos++)
	{
	  double sum_pos = 0;
	  for (int i=0; i<_nmodal; i++) 
	    {
	      sum_pos += _probatrans[_mut_class][_ancestor->_seq[ipos]][i]
		*_peeling_seq[ipos][i];
	    }	  
	  double rm = gsl_ran_flat(_rng, 0, sum_pos);
	  int ri = 0;
	  double sum_pm = _probatrans[_mut_class][_ancestor->_seq[ipos]][0]
	    *_peeling_seq[ipos][0];
	  while (sum_pm < rm && ri+1<_nmodal) 
	    {
	      ri++;
	      sum_pm += _probatrans[_mut_class][_ancestor->_seq[ipos]][ri]
		*_peeling_seq[ipos][ri];
	    }
	  _seq[ipos] = ri;
	}	  
    } else if (_node_status == 0) {
      // Nothing to do
    }
  if (_prev != NULL) 
    {
      _prev->simulSeqForward();
    }
}

void Coalnode::peelingMutClassBackward()
{
  if (_contextdep)
    {
      cerr << "Error: void Coalnode::peelingMutClassBackward() is not implemented for the case _contextdep==true" << endl;
      exit(1);
    }
  // Given sequences at the nodes of the tree
  if (_node_status == 0) 
    {
      for (int imut=0; imut<_nb_mutclass; imut++) 
	{
	  if (_mut_class==imut) 
	    {
	      _peeling_mutclass[imut] = 1;
	    } else {
	      _peeling_mutclass[imut] = 1;
	    }
	}
    } else if (_node_status == 1) {
      //      _child_left->computeProbaTrans();
      // _child_right->computeProbaTrans(); 
      // WE DON'T DO THAT BECAUSE WE ALREADY DID IT IN peelingSeqBackward()      

      for (int imut=0; imut<_nb_mutclass; imut++) 
	{
	  double left_term = 0;
	  double right_term = 0;
	  for (int jmut=0; jmut<_nb_mutclass; jmut++) 
	    {
	      double jleft_term 
		= _child_left->_probaswitch[imut][jmut]*_child_left->_peeling_mutclass[jmut];
	      double jright_term 
		= _child_right->_probaswitch[imut][jmut]*_child_right->_peeling_mutclass[jmut];
	      for (int ipos=0; ipos<_lseq ; ipos++) 
		{
		  jleft_term 
		    *= _child_left->_probatrans[jmut][_seq[ipos]][_child_left->_seq[ipos]];
		  jright_term 
		    *= _child_right->_probatrans[jmut][_seq[ipos]][_child_right->_seq[ipos]];
		}
	      left_term += jleft_term;
	      right_term += jright_term;
	    }
	  _peeling_mutclass[imut] = left_term*right_term;
	}            
    } else if (_node_status == 2) {
      //      _child_left->computeProbaTrans();
      // WE DON'T DO THAT BECAUSE WE ALREADY DID IT IN peelingSeqBackward()      
      for (int imut=0; imut<_nb_mutclass; imut++) 
	{
	  double left_term = 0;
	  for (int jmut=0; jmut<_nb_mutclass; jmut++) 
	    {
	      double jleft_term 
		= _child_left->_probaswitch[imut][jmut]*_child_left->_peeling_mutclass[jmut];
	      for (int ipos=0; ipos<_lseq ; ipos++) 
		{
		  jleft_term 
		    *= _child_left->_probatrans[jmut][_seq[ipos]][_child_left->_seq[ipos]];
		}
	      left_term += jleft_term;
	    }
	  _peeling_mutclass[imut] = left_term;
	}            
    }
  if (_next != NULL) 
    {
      _next->peelingMutClassBackward();
    }
}

void Coalnode::simulMutClassForward()
{
  if (_contextdep)
    {
      cerr << "Error: void Coalnode::simulMutClassForward() is not implemented for the case _contextdep==true" << endl;
      exit(1);
    }
  if (_node_status == 2)
    {
      _mut_class = 0;
    } else if (_node_status == 1) {
      double sum = 0;
      for (int imut=0; imut<_nb_mutclass; imut++) 
	{
	  sum += _probaswitch[_ancestor->_mut_class][imut]
	    *_peeling_mutclass[imut];
	}	  
      double rm = gsl_ran_flat(_rng, 0, sum);
      int ri = 0;
      double sum_pm = _probaswitch[_ancestor->_mut_class][ri]
	*_peeling_mutclass[ri];      
      while (sum_pm < rm && ri+1<_nb_mutclass) 
	{
	  ri++;
	  sum_pm += _probaswitch[_ancestor->_mut_class][ri]
	    *_peeling_mutclass[ri];
	}
      _mut_class = ri;
    } else if (_node_status == 0) {
      double sum = 0;
      for (int imut=0; imut<_nb_mutclass; imut++) 
	{
	  sum += _probaswitch[_ancestor->_mut_class][imut]
	    *_peeling_mutclass[imut];
	}	  
      double rm = gsl_ran_flat(_rng, 0, sum);
      int ri = 0;
      double sum_pm = _probaswitch[_ancestor->_mut_class][ri]
	*_peeling_mutclass[ri];      
      while (sum_pm < rm && ri+1<_nb_mutclass) 
	{
	  ri++;
	  sum_pm += _probaswitch[_ancestor->_mut_class][ri]
	    *_peeling_mutclass[ri];
	}
      _mut_class = ri;
      // Nothing to do
    }  
  if (_prev != NULL) 
    {
      _prev->simulMutClassForward();
    }
}


double Coalnode::sumLikelihoodNode()
{
  double ret_val = 0;

  if (!_contextdep) {
    if (_node_status == 0) {
      this->computeProbaTrans();      
      double sum_mut = 0;
      for (int imut=0; imut<_nb_mutclass; imut++)
	{
	  _posterior_probas_mutclass[imut] = _probaswitch[_ancestor->_mut_class][imut];
	  for (int ipos=0 ; ipos<_lseq ; ipos++)
	    {
	      double ipos_term = 0;
	      for (int i=0 ; i<_nmodal ; i++) 
		{
		  double error_term;
		  if (_obs_seq[ipos]<0) 
		    {
		      error_term = 1;
		    } else if (i!=_obs_seq[ipos]) {
		      error_term = (*_mut_model)->perror();
		    } else {
		      error_term = 1-(*_mut_model)->perror(); 
		    }
		
		  _posterior_probas[imut][ipos][i] 
		    = _probatrans[imut][_ancestor->_seq[ipos]][i]*error_term;
		  
		  ipos_term += _posterior_probas[imut][ipos][i];
		}	      
	      for (int i=0 ; i<_nmodal ; i++) 
		{
		  _posterior_probas[imut][ipos][i] /= ipos_term; 
		  // by the way we scale the posterior proba
		}
	      _posterior_probas_mutclass[imut] *= ipos_term;	
	    }
	  ret_val += _posterior_probas_mutclass[imut];
	  sum_mut += _posterior_probas_mutclass[imut];
	}
      for (int imut=0; imut<_nb_mutclass; imut++)
	{
	  _posterior_probas_mutclass[imut] /= sum_mut;
	}	  
    } else if (_node_status == 1) {
      this->computeProbaTrans();
      _child_left->computeProbaTrans();
      _child_right->computeProbaTrans();
      double sum_mut = 0;
      for (int imut=0; imut<_nb_mutclass; imut++)
	{
	  _posterior_probas_mutclass[imut] = _probaswitch[_ancestor->_mut_class][imut]
	    *_child_left->_probaswitch[imut][_child_left->_mut_class]
	    *_child_right->_probaswitch[imut][_child_right->_mut_class];
	  for (int ipos=0 ; ipos<_lseq ; ipos++)
	    {
	      double ipos_term = 0;
	      for (int i=0 ; i<_nmodal ; i++) 
		{
		  _posterior_probas[imut][ipos][i] 
		    = _probatrans[imut][_ancestor->_seq[ipos]][i]
		    *_child_left->_probatrans[_child_left->_mut_class][i][_child_left->_seq[ipos]]
		    *_child_right->_probatrans[_child_right->_mut_class][i][_child_right->_seq[ipos]];
		  ipos_term += _posterior_probas[imut][ipos][i];
		}	    
	      for (int i=0 ; i<_nmodal ; i++) 
		{
		  _posterior_probas[imut][ipos][i] /= ipos_term; 
		  // by the way we scale the posterior proba
		}
	      _posterior_probas_mutclass[imut] *= ipos_term;
	    }
	  ret_val += _posterior_probas_mutclass[imut];
	  sum_mut += _posterior_probas_mutclass[imut];
	}
      for (int imut=0; imut<_nb_mutclass; imut++)
	{
	  _posterior_probas_mutclass[imut] /= sum_mut;
	}	  
    } else if (_node_status == 2) {
      ret_val = 1;
      _child_left->computeProbaTrans();
      ret_val *= _child_left->_probaswitch[0][_child_left->_mut_class];
      for (int ipos=0 ; ipos<_lseq ; ipos++)
	{
	  ret_val *= _child_left->_probatrans[_child_left->_mut_class][0][_child_left->_seq[ipos]];
	}      
    } else {
      cout << "Pb 1 in double Coalnode::sumLikelihoodNode()" << endl;
      exit(1);
    }
  } else { // _contextdep == TRUE
    if (_node_status == 0) {          
      double bl_a = (*_mut_model)->computeBranchLength(_time, _ancestor->_time);
      ret_val = (*_mut_model)->contextSumLikelihoodNodeType0(bl_a, _ancestor->_seq, _obs_seq);
    } else if (_node_status == 1) {
      double bl_a = (*_mut_model)->computeBranchLength(_time, _ancestor->_time);
      double bl_cl = (*_mut_model)->computeBranchLength(_child_left->_time, _time);
      double bl_cr = (*_mut_model)->computeBranchLength(_child_right->_time, _time);
      ret_val = 
	(*_mut_model)->contextSumLikelihoodNode(bl_a,bl_cl,bl_cr,
						_ancestor->_seq, 
						_child_left->_seq,
						_child_right->_seq);
    } else if (_node_status == 2) {
      double bl_cl = (*_mut_model)->computeBranchLength(_child_left->_time, _time);
      ret_val = 
	(*_mut_model)->probatranscontext(bl_cl, 0, _child_left->_seq);
    } 
  }
  return ret_val;
}

void Coalnode::posteriorSimulNode()
{
  if (!_contextdep) {
    if (_node_status == 1 || _node_status == 0) {
      // to be used after double Coalnode::sumLikelihoodNode()
      this->sumLikelihoodNode();

      if (_nb_mutclass==1)
	{
	  _mut_class = 0;
	} else {
	  double rm = gsl_ran_flat(_rng, 0, 1);
	  _mut_class = 0;
	  double sum_pm = _posterior_probas_mutclass[0];
	  while (sum_pm < rm && _mut_class+1<_nb_mutclass) 
	    {
	      _mut_class++;
	      sum_pm += _posterior_probas_mutclass[_mut_class];
	    }
	}

      for (int ipos=0 ; ipos<_lseq ; ipos++)
	{
	  double r = gsl_ran_flat(_rng, 0, 1);
	  int outcome = 0;
	  double sump = _posterior_probas[_mut_class][ipos][outcome];
	  while (sump < r && outcome<_nmodal) 
	    {
	      outcome++;
	      sump += _posterior_probas[_mut_class][ipos][outcome];
	    }
	  _seq[ipos] = outcome;
	}

    } else if (_node_status == 2) {
      _mut_class = 0;
      for (int ipos=0 ; ipos<_lseq ; ipos++)
	{
	  _seq[ipos] = 0;
	}
    } else {
      cerr << "pb in void Coalnode::posteriorSimulNode()" << endl;
      exit(1);
    }
  } else { // _contextdep == TRUE
    _mut_class = 0;
    if (_node_status == 0) {
      double bl_a = (*_mut_model)->computeBranchLength(_time, _ancestor->_time);
      (*_mut_model)->contextPosteriorSimulNodeType0(bl_a, _ancestor->_seq, _obs_seq, _seq);
    } else if (_node_status == 1) {
      double bl_a = (*_mut_model)->computeBranchLength(_time, _ancestor->_time);
      double bl_cl = (*_mut_model)->computeBranchLength(_child_left->_time, _time);
      double bl_cr = (*_mut_model)->computeBranchLength(_child_right->_time, _time);
      (*_mut_model)->contextPosteriorSimulNode(bl_a,bl_cl,bl_cr,
					       _ancestor->_seq, 
					       _child_left->_seq,
					       _child_right->_seq, _seq);
    } else if (_node_status == 2) {
      for (int ipos=0 ; ipos<_lseq ; ipos++)
	{
	  _seq[ipos] = 0;
	}
    }
  }
}


void Coalnode::splitTreeWB()
{
  //cout << "splitTreeWB() _child_right " <<  _child_right->node_id() << endl; 
  _ancestor->_child_left = _child_right;
  _child_right->_ancestor = _ancestor;
  //cout << "set _child_right->_ancestor to " << _child_right->_ancestor->node_id() << endl;

  _next->_prev = _prev;
  _prev->_next = _next;
}

void Coalnode::joinTreeWB(Coalnode* target, double coaltime)
{
  //cout << "joinTreeWB this=" << this->node_id() << " target=" << target->node_id() << " target->_ancestor=" << target->_ancestor->node_id() << endl; 
  _ancestor = target->_ancestor;
  _ancestor->_child_left = this;

  _child_right = target;
  _child_right->_ancestor = this;
  
  _time = coaltime;

  Coalnode* current_node = target;
  while (current_node->_time < _time && current_node->_next != NULL)
    {
      current_node = current_node->_next;
    }
  _next = current_node;
  _prev = current_node->_prev;
  _prev->_next = this;
  _next->_prev = this;
}

Coalnode* Coalnode::diffLineageRoot()
{
  if (_time > 0) {
    return NULL;
  }
  if (_time <= 0 && _ancestor->time() > 0) {
    return this;
  } else {
    return _ancestor->diffLineageRoot();
  }
  cerr << "Pb in Coalnode* Coalnode::diffLineageRoot()" << endl;
  return NULL;
}
