/* Copyright (C) 2006 Pierre Nicolas
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "mutmodel.h"
#include "crypt.h"
#include "coalnode.h"

#include<iostream>
#include<map>
#include<set>
#include<gsl/gsl_randist.h>
#include<gsl/gsl_cdf.h>

using namespace std;

int Crypt::_count = 0;

Crypt::Crypt(Mutmodel *mut_model, int N_stem_cells, double lambda,  double g, gsl_rng * rng, 
	     double** stirling2, ifstream& input)
{
  _crypt_id = _count;
  _count++;
  _N_stem_cells = N_stem_cells;
  _rng = rng;
  _lambda = lambda;
  _g = 5;
  _mut_model = mut_model;

  _node_collector = NULL;
  _stirling2 = stirling2;
  _prior_proba_N = new double [100];
  _posterior_proba_N = new double [100];

  int nseq_obs;

  string word;


  input >> word;
  input >> word;
  input >> word;
  input >> _age_crypt;
  input >> word;
  input >> _lseq;
  input >> word;
  input >> nseq_obs;
  input >> word;
  input >> _ncells;
  
  //cout << "_age_crypt " << _age_crypt << " _lseq " << _lseq << " _nseq " << _nseq << " _ncells " << _ncells << endl;


  _cells_obs_seqs = new int* [_ncells];

  _errors_counts = new int* [_ncells];
  for (int i=0 ; i<_ncells ; i++)
    {
      _cells_obs_seqs[i] = new int [_lseq];
      _errors_counts[i] = new int [_lseq];
    }
  _missing = new bool [_lseq];
  for (int i=0; i<_lseq; i++)
    {
      _missing[i] = false;
    }
  
  int icell = 0;
  int times = 0;
  for (int i=0 ; i<nseq_obs ; i++)
    {   
      for (int j=0 ; j<_lseq ; j++)
	{
	  char ic;
	  input >> ic;
	  if (ic == 48 || ic == 49)
	    {
	      _cells_obs_seqs[icell][j] = ic-48; 
	    } else {
	      _cells_obs_seqs[icell][j] = -1; 
	      _missing[j] = true;
	      //cout << "seq " << i << " pos " << j << endl; 
	    }
	}
      icell++;
      input >> times;
      for (int k=1 ; k<times ; k++)
	{	 	  
	  for (int j=0 ; j<_lseq ; j++)
	    {
	      _cells_obs_seqs[icell][j] = _cells_obs_seqs[icell-1][j];
	    }
	  icell++;
	}
    }
  if (icell != _ncells) {
    cerr << "Pb while reading data for crypt " << _count << endl;
    exit(1);
  }
  
  this->drawGenealogyPrior();
  this->drawSeqGenealogyPrior();
  _ptarget_wb = new double [_ncells*3];
  _target_wb = new Coalnode* [_ncells*3];
} 

Crypt::~Crypt()
{
  this->deleteGenealogy();
  delete _ptarget_wb;
  delete _target_wb;
}

void Crypt::printCrypt()
{
  map<vector<int>, int> seqcounts;
  this->countseq(seqcounts, _cells_obs_seqs);
  
  cout << _crypt_id << " X age " << _age_crypt << " lseq " << _lseq << " nseq " << seqcounts.size() << " ncells " << _ncells << endl;
}

void Crypt::printCryptFromObsSeq()
{
  this->printCryptFromObsSeq(cout);
}

void Crypt::printCryptFromObsSeq(ostream& out)
{
  
  map<vector<int>, int> seqcounts;
  this->countseq(seqcounts, _cells_obs_seqs);

  out << _crypt_id << " X age " << _age_crypt << " lseq " << _lseq << " nseq " << seqcounts.size() << " ncells " << _ncells << endl;
  for (map<vector<int>, int>::iterator it = seqcounts.begin();
       it != seqcounts.end(); it++)
    {
      for (int j=0 ; j<_lseq ; j++)
	{
	  if (((*it).first)[j] >=0 && !_missing[j]) 
	    {
	      out << ((*it).first)[j];
	    } else {
	      out << 'N';
	    }
	}
      out << " " << (*it).second << endl;
    }
}

void Crypt::countseq(map<vector<int>, int>& seqcounts, int** sequences)
{
  for (int i=0 ; i<_ncells ; i++)
    {
      vector<int> currseq(_lseq);
      for (int ipos=0; ipos<_lseq ; ipos++)
	{
	  currseq[ipos] = sequences[i][ipos];
	}
      map<vector<int>, int>::iterator it = seqcounts.find(currseq);
      if (it != seqcounts.end()) 
	{
	  (*it).second += 1;
	} else {
	  seqcounts.insert(pair<vector<int>, int>(currseq,1));
	}
    }  
}

void Crypt::resetErrorCounts()
{
  for (int i=0; i<_ncells; i++)
    {
      for (int j=0; j<_lseq; j++)
	{
	  _errors_counts[i][j] = 0;
	}
    }
}

void Crypt::storeErrorCounts()
{
  Coalnode* currentnode = _firstnode;
  while (currentnode->node_status()==0) 
    {
      int* seq = currentnode->seq_ptr();
      int icell = currentnode->iseq();
      for (int ipos=0; ipos<_lseq; ipos++) 
	{
	  if (_cells_obs_seqs[icell][ipos]>=0 
	      && seq[ipos] != _cells_obs_seqs[icell][ipos]) 
	    {
	      _errors_counts[icell][ipos]++;
	    }
	}      
      currentnode = currentnode->next();
    }
}

void Crypt::printErrorCounts()
{
  cout << "crypt " << _crypt_id << endl;
  for (int i=0; i<_ncells; i++)
    {
      cout << "seq" << i << " ";
      for (int j=0; j<_lseq; j++)
	{
	  cout << _errors_counts[i][j] << "\t";
	}
      cout << endl;
    }
}

void Crypt::countErrors(int& sum1, int& sum2)
{
  Coalnode* currentnode = _firstnode;
  while (currentnode->node_status()==0) 
    {
      int* seq = currentnode->seq_ptr();
      int icell = currentnode->iseq();
      for (int ipos=0; ipos<_lseq; ipos++) 
	{
	  if (_cells_obs_seqs[icell][ipos]>=0) 
	    {
	      if (seq[ipos] != _cells_obs_seqs[icell][ipos]) 
		{
		  sum1++;
		} else {
		  sum2++;
		}
	    }      
	}
      currentnode = currentnode->next();
    }
}

void Crypt::drawGenealogyPrior() 
{
  
  // initialize

  _firstnode = NULL;
  _lastnode = _firstnode;
  _nb_internal_nodes = 0;

  for (int k=0 ; k<_ncells ; k++) 
    {
      Coalnode* newnode = new Coalnode(0, _lastnode, NULL, NULL, NULL, NULL, -1, 
				       _lseq, &_mut_model, _rng);
      newnode->setiseq(k, _cells_obs_seqs[k]);
      if (k==0) 
	{
	  _firstnode = newnode;
	}
      _lastnode = newnode;      
    }

  // draw which node come from which stem cell

  int nb_stemcells_sampled = 0;
  set<Coalnode*> *cells_bystemcell = new set<Coalnode*> [_N_stem_cells];
  for (int k=0; k<_N_stem_cells; k++)
    {
      cells_bystemcell[k].clear();
    }
  
  for (Coalnode* currnode = _firstnode; currnode != NULL; currnode=currnode->next()) 
    {
      double proba_incr = (double)(_N_stem_cells - nb_stemcells_sampled)
	/(double)(_N_stem_cells);
      int increase = gsl_ran_bernoulli(_rng, proba_incr);
      nb_stemcells_sampled += increase;
      if (increase == 0)
	{
	  cells_bystemcell[gsl_rng_uniform_int(_rng, nb_stemcells_sampled)].insert(currnode);
	} else {
	  cells_bystemcell[nb_stemcells_sampled-1].insert(currnode);	
	}
    }
  _nstemcell_sampled = nb_stemcells_sampled;

  // draw differentiation genealogy for each stem cell
  set<Coalnode*> stemcellnodes;
  stemcellnodes.clear();
  for (int istem=0; istem<_nstemcell_sampled; istem++) 
    {
      int n = cells_bystemcell[istem].size();
      vector<double> coaltimes;
      coaltimes.clear();
      this->drawDiffCoalTimesFromPrior(coaltimes, n);

      Coalnode* currnode = _firstnode;
      for (int nliving=n; nliving>1; nliving--) 
	{ 
	  int choose_left = gsl_rng_uniform_int(_rng, nliving);
	  int choose_right = gsl_rng_uniform_int(_rng, nliving-1);	
	  if (choose_left <= choose_right) 
	    {
	      // a trick for sampling without replacement
	      choose_right++;
	    }
      
	  Coalnode* childleft;
	  Coalnode* childright;
	  int inode = 0;
	  for (set<Coalnode*>::iterator it = cells_bystemcell[istem].begin(); 
	       it != cells_bystemcell[istem].end(); it++) 
	    {
	      if (inode == choose_left) 
		{
		  childleft = *it;
		}
	      if (inode == choose_right) 
		{
		  childright = *it;
		}
	      inode++;
	    }
	  cells_bystemcell[istem].erase(childright);
	  cells_bystemcell[istem].erase(childleft);

	  double time = coaltimes[n-nliving];
	  while (currnode != NULL && currnode->time() < time) 
	    {
	      currnode = currnode->next();
	    }	  
	  Coalnode* nextnode = currnode;
	  Coalnode* prevnode = _lastnode;
	  if (currnode != NULL) 
	    {
	      prevnode = currnode->prev();
	    }
	  currnode = prevnode;

	  Coalnode* newnode = new Coalnode(1, prevnode, nextnode, 
					   childright, childleft, NULL, 
					   time, 
					   _lseq,
					   &_mut_model, _rng);
	  _nb_internal_nodes++;
	  cells_bystemcell[istem].insert(newnode);
	  if (nextnode == NULL) 
	    {
	      _lastnode = newnode;
	    }
	}
      Coalnode* stemcellnode = *(cells_bystemcell[istem].begin());
      stemcellnodes.insert(stemcellnode);
    }
  delete [] cells_bystemcell;

  // draw stem cell genealogy

  double time = 0;
  int nb_living_cells = stemcellnodes.size();
  while (time < _age_crypt && nb_living_cells > 1)
    {
      //cout << "time " << time << endl;
      double coalrate = this->coalratefunc(nb_living_cells);
      double coaltime = gsl_ran_exponential (_rng, 1/coalrate);
      time += coaltime;
      if (time < _age_crypt) 
	{
	  int choose_left = gsl_rng_uniform_int(_rng, nb_living_cells);
	  int choose_right = gsl_rng_uniform_int(_rng, nb_living_cells-1);
	  if (choose_left <= choose_right) 
	    {
	      choose_right++;
	    }

	  Coalnode* childleft;
	  Coalnode* childright;
	  
	  int inode = 0;
	  for (set<Coalnode*>::iterator it = stemcellnodes.begin(); 
	       it != stemcellnodes.end(); it++) 
	    {
	      if (inode == choose_left) 
		{
		  childleft = *it;
		}
	      if (inode == choose_right) 
		{
		  childright = *it;
		}
	      inode++;
	    }
	  stemcellnodes.erase(childright);
	  stemcellnodes.erase(childleft);

	  nb_living_cells--;
	  
	  Coalnode* newnode = new Coalnode(1, _lastnode, NULL, 
					   childright, childleft, NULL, time, 
					   _lseq,
					   &_mut_model, _rng);
	  _nb_internal_nodes++;
	  stemcellnodes.insert(newnode);
	  _lastnode = newnode;	
	}
    }

  // Add the roots

  for (set<Coalnode*>::iterator it = stemcellnodes.begin() ; 
       it != stemcellnodes.end() ; it++)
    {
      Coalnode* newnode = new Coalnode(2, _lastnode, NULL, 
				       NULL, *it , NULL, _age_crypt, _lseq,
				       &_mut_model, _rng);
      _lastnode = newnode;	        
    }
}

void Crypt::drawSeqErrorPrior() 
{
  Coalnode* currentnode = _firstnode;
  while (currentnode->node_status()==0) 
    {
      vector<int> seq = currentnode->seq();
      int icell = currentnode->iseq();
      for (int ipos=0; ipos<_lseq; ipos++) 
	{
	  _cells_obs_seqs[icell][ipos] = seq[ipos];
	  int rint=gsl_ran_bernoulli(_rng, _mut_model->perror());
	  if (rint==1)
	    {
	      if (_cells_obs_seqs[icell][ipos]==0)
		{
		  _cells_obs_seqs[icell][ipos]=1;
		} else {
		  _cells_obs_seqs[icell][ipos]=0;
		}
	    }
	  if (_missing[ipos]) 
	    {
	      _cells_obs_seqs[icell][ipos]=-1;
	    }
	}      
      currentnode = currentnode->next();
    }
}

void Crypt::simulPrior()
{
  this->drawGenealogyPrior();
  this->drawSeqGenealogyPrior();
  this->drawSeqErrorPrior();
}


double Crypt::coalratefunc(int nbliving)
{
  double rate = _lambda*((double)(nbliving*(nbliving-1)));
  return rate;
}

void Crypt::drawDiffCoalTimesFromPrior(vector<double>& coaltimes, int n)
{
  coaltimes.clear();
  double s = 0;
  for (int k=n; k>1; k--) 
    {
      double coalrate = (double)(k*(k-1)/2);
      double t = gsl_ran_exponential(_rng, 1/coalrate);
      s += t;
      coaltimes.push_back(-1 + this->invIntegratedRateFunc(s));
   }
//   cout << "coaltimes simul ";
//   for (int i=0; i<coaltimes.size();i++) 
//     {
//       cout << coaltimes[i] << " ";
//     }
//   cout << endl;
}

double Crypt::invIntegratedRateFunc(double s)
{
  return (1/(_g*log(2.)))*log(pow(2.,_g)-(pow(2.,_g)-1)*exp(-s));
}

double Crypt::loglikDiffCoal(vector<double>& coaltimes)
{
  // compute the log pdf for the genealogy of a differentiation lineage
  // this computation takes into account both the coal times 
  // and the tree combinatoric
  double ret_val = 0;
  int n = coaltimes.size();
  if (1) {
  for (int i=0; i<n; i++)
    {
      ret_val += log(this->diffRatefunc(1+coaltimes[i]));
      if (i>0) 
	{
	  ret_val += -((n-i+1)*(n-i)/2)*(this->diffIntegratedRatefunc(1+coaltimes[i])
					 -this->diffIntegratedRatefunc(1+coaltimes[i-1]));
	} else {
	  ret_val += -((n-i+1)*(n-i)/2)*this->diffIntegratedRatefunc(1+coaltimes[i]);
	}
      //ret_val += log((n-i+1)*(n-i)/2)-log((n-i+1)*(n-i)/2);
    }
  } else {
    for (int i=0; i<n; i++)
      {
	//cout << coaltimes[i] << " ";
	ret_val += log(i+1.);
	ret_val += -log((n-i+1)*(n-i)/2.);
      }
    //cout << endl;
  }
  return ret_val;
}

void Crypt::computePriorProbaN(int* nb_stemcells_counts, 
			       int Nmin, int Nmax, 
			       int& Npriormin, int& Npriormax,
			       int nstemcells_prior_min, int nstemcells_prior_max)
{
  int n = 0;
  for (int i=Nmin ; i<=Nmax; i++) 
    {
      n += nb_stemcells_counts[i];
    }
  if (n==1) 
    {
      Npriormin=nstemcells_prior_min;
      Npriormax=nstemcells_prior_max;
      for (int i=Npriormin; i<=Npriormax; i++) 
	{
	  _prior_proba_N[i] = 1./(double)(Npriormax-Npriormin+1);
	}
      return;
    } else {
      if (Nmin==Nmax) 
	{
	  Npriormin = Nmin-1;
	  Npriormax = Nmax+1;
	} else if (Nmax-Nmin==1) {
	  if (nb_stemcells_counts[Nmin]>1 && nb_stemcells_counts[Nmax]>1) 
	    {
	      Npriormin=Nmin;
	      Npriormax=Nmax;	      
	    } else if (nb_stemcells_counts[Nmin]==1 && _N_stem_cells==Nmin) {
	      Npriormin=Nmin;
	      Npriormax=Nmax+1;	      	      
	    } else if (nb_stemcells_counts[Nmax]==1 && _N_stem_cells==Nmax) {
	      Npriormin=Nmin-1;
	      Npriormax=Nmax;	      	      
	    } else {
	      Npriormin=Nmin;
	      Npriormax=Nmax;	      	      
	    }
	} else {
	  cerr << "pb 1 in Crypt::computePriorProbaN" << endl;
	  int *k=NULL;
	  *k=1;
	  exit(1);
	}
      if (Npriormin<nstemcells_prior_min) {
	Npriormin = nstemcells_prior_min;
      }
      if (Npriormax>nstemcells_prior_max) {
	Npriormax = nstemcells_prior_max;
      }
      nb_stemcells_counts[_N_stem_cells]--;
      double sum = 0;
      for (int i=Npriormin; i<=Npriormax; i++) 
	{
	  _prior_proba_N[i] = nb_stemcells_counts[i]+1;
	  sum += _prior_proba_N[i];
	}
      for (int i=Npriormin; i<=Npriormax; i++) 
	{
	  _prior_proba_N[i]/=sum;
	}
      nb_stemcells_counts[_N_stem_cells]++;
    }
}


void Crypt::updateNstemcellFancy(int* nb_stemcells_counts,int& Nmin, int& Nmax, 
				 int nstemcells_prior_min, int nstemcells_prior_max)
{
  //cout << "Entering updateNstemcellFancy" << endl;
  //cout << "_nstemcell_sampled " << _nstemcell_sampled << endl;
  

  int n = 0;
  int nstemcells_min = Nmin;
  int nstemcells_max = Nmax;
  for (int i=nstemcells_prior_min ; i<=nstemcells_prior_max; i++) 
    {
      n += nb_stemcells_counts[i];
    }

  // Check
  //cout << "n " << n << " nstemcells_min " << nstemcells_min << " nstemcells_max " << nstemcells_max << " _N_stem_cells " << _N_stem_cells << endl;
  if (nstemcells_max - nstemcells_min > 1) 
    {
      cout << "_N_stem_cells " << _N_stem_cells << endl;
      cout << "nstemcells_max " << nstemcells_max << endl;
      cout << "nstemcells_min " << nstemcells_min << endl;
      cerr << "Pb 1 in void Crypt::updateNstemcellFancy" << endl;
      exit(1);
    }
  if (_N_stem_cells != nstemcells_max && _N_stem_cells != nstemcells_min)
    {
      cout << "_N_stem_cells " << _N_stem_cells << endl;
      cout << "nstemcells_max " << nstemcells_max << endl;
      cout << "nstemcells_min " << nstemcells_min << endl;
      cerr << "Pb 2 in void Crypt::updateNstemcellFancy" << endl;
      exit(1);
   }

  bool increase = false;
  int rint = gsl_ran_bernoulli(_rng,0.5);
  if (rint==0)
    {
      increase = true;
      //cout << "Increase attempt" << endl;
      //cout << _N_stem_cells << " -> " << _N_stem_cells+1 << endl;
    } else {
      //cout << "Decrease attempt" << endl;
      //cout << _N_stem_cells << " -> " << _N_stem_cells-1 << endl;  
    }
  
  if (increase==true && _N_stem_cells==nstemcells_prior_max)
    {
      return;
    }
  if (increase==false && _N_stem_cells==nstemcells_prior_min)
    {
      return;
    }
  double log_prior_ratio = 0;
  if (nstemcells_max != nstemcells_min) 
    {
      if (increase==true && _N_stem_cells==nstemcells_max)
	{
	  return;
	}
      if (increase==false && _N_stem_cells==nstemcells_min)
	{
	  return;
	}

      int k = nb_stemcells_counts[nstemcells_max];
      if (increase==true)
	{
	  log_prior_ratio = log((double)(k+1)/(double)(n-k));
	} else {
	  log_prior_ratio = log((double)(n-k+1)/(double)(k));
	}
    } else {
      log_prior_ratio = -log((double)n);
    }



  double log_lik_num = 0;
  double log_lik_den = 0;

  // log_lik_den += computeLogLikTreeN();
  log_lik_den += computeLogLikN();

  if (increase==true)
    {
      _N_stem_cells++;
    } else {
      _N_stem_cells--;
    }
  //log_lik_num += computeLogLikTreeN();      
  log_lik_num += computeLogLikN();      

  //cout << "log_prior_ratio " << log_prior_ratio << endl;
  //cout << "log_lik_num " << log_lik_num << endl;
  //cout << "log_lik_den " << log_lik_den << endl;

  double logratio = log_prior_ratio + log_lik_num-log_lik_den;

  //cout << "ratio " << exp(logratio) << endl;
  
  double r = log(gsl_ran_flat(_rng, 0, 1));
  if (r < logratio)
    {
      //cout << "updateNstemcell Accepted" << endl;
      if (increase == true) 
	{
	  nb_stemcells_counts[_N_stem_cells-1]--;
	  nb_stemcells_counts[_N_stem_cells]++;
	  if (nb_stemcells_counts[Nmin]==0)
	    {
	      Nmin++;
	    }
	  if (_N_stem_cells>Nmax) 
	    {
	      Nmax = _N_stem_cells;
	    }
	  //cout << "Increasing" << endl;
	} else {
	  nb_stemcells_counts[_N_stem_cells+1]--;
	  nb_stemcells_counts[_N_stem_cells]++;
	  //cout << "Decreasing" << endl;
	  if (nb_stemcells_counts[Nmax]==0)
	    {
	      Nmax--;
	    }
	  if (_N_stem_cells<Nmin) 
	    {
	      Nmin = _N_stem_cells;
	    }
	}
    } else {
      //cout << "updateNstemcell Rejected" << endl;
      if (increase == true) 
	{
	  _N_stem_cells--;
	} else {
	  _N_stem_cells++;
	}      
    }
}

void Crypt::updateNstemcell(double* nb_stemcells_distrib, int nstemcells_min, int nstemcells_max)
{
  //cout << "Entering updateNstemcell" << endl;
  //cout << "_nstemcell_sampled " << _nstemcell_sampled << endl;
  bool increase = false;
  int rint = gsl_ran_bernoulli(_rng,0.5);
  if (rint==0)
    {
      increase = true;
      //cout << "Increase attempt" << endl;
      //cout << _N_stem_cells << " -> " << _N_stem_cells+1 << endl;
    } else {
      //cout << "Decrease attempt" << endl;
      //cout << _N_stem_cells << " -> " << _N_stem_cells-1 << endl;  
    }

  if (increase==true && _N_stem_cells==nstemcells_max)
    {
      return;
    }
  if (increase==false && _N_stem_cells==nstemcells_min)
    {
      return;
    }

  double log_lik_num = 0;
  double log_lik_den = 0;

  log_lik_den += log(nb_stemcells_distrib[_N_stem_cells]);
  log_lik_den += computeLogLikTreeN();

  if (increase==true)
    {
      _N_stem_cells++;
    } else {
      _N_stem_cells--;
    }
  log_lik_num += log(nb_stemcells_distrib[_N_stem_cells]);
  log_lik_num += computeLogLikTreeN();      

  //cout << "log_lik_num " << log_lik_num << endl;
  //cout << "log_lik_den " << log_lik_den << endl;

  double logratio = log_lik_num-log_lik_den;

  //cout << "ratio " << exp(logratio) << endl;
  
  double r = log(gsl_ran_flat(_rng, 0, 1));
  if (r < logratio)
    {
      //cout << "updateNstemcell Accepted" << endl;
      if (increase == true) 
	{
	  //cout << "Increasing" << endl;
	} else {
	  //cout << "Decreasing" << endl;
	}
    } else {
      //cout << "updateNstemcell Rejected" << endl;
      if (increase == true) 
	{
	  _N_stem_cells--;
	} else {
	  _N_stem_cells++;
	}      
    }
 
}

void Crypt::updateInitStemCell()
{
}


void Crypt::updateTermStemCell(bool fake)
{
  //cout << "Entering updateTermStemCell crypt " << _crypt_id << endl;
  bool increase = false;

  int rint = gsl_ran_bernoulli(_rng,0.5);
  if (rint==0)
    {
      increase = true;
      //cout << "Increase attempt" << endl;
    } else {
      //cout << "Decrease attempt" << endl;
    }


  int no = 0;
  double coaltime;

  Coalnode* currentnode = _lastnode;
  while (currentnode->node_status()==2)
    {
      no++;
      currentnode = currentnode->prev();
    }

  if (no == 1 && increase==false) 
    {
      return; // no way of decreasing
    } 
  if (currentnode->node_status()==0 && increase==true) 
    {
      return; // no way of increasing
    }
  Coalnode* last_internal_node = currentnode;
  
  double log_proposal_num = 0;
  double log_proposal_den = 0;
  int nstemcell_new = _nstemcell_sampled;
  int nstemcell_old = _nstemcell_sampled;
  if (increase == true)
    {
      double tmin = last_internal_node->prev()->time();
      double tmax = last_internal_node->next()->time();
      coaltime = last_internal_node->time();
      if (coaltime<0) 
	{
	  nstemcell_new++;
	}
      if (tmin<0) 
	{
	  log_proposal_num += log(0.5);
	  if (coaltime<0) 
	    {
	      log_proposal_num += log(1/(0-tmin));
	    } else {
	      log_proposal_num += log(1/(tmax-0));
	    }
	} else {
	  log_proposal_num += log(1/(tmax-tmin));
	}
      //cout << "log_proposal_num " << log_proposal_num << endl;
      log_proposal_num += log(2./((no+1)*no));
      //cout << "log_proposal_num " << log_proposal_num << endl;
    } else { // increase == false
      double tmin = last_internal_node->time();
      double tmax = last_internal_node->next()->time();
      if (tmin<0) 
	{
	  log_proposal_den += log(0.5);
	  int r_geneal_part = gsl_ran_bernoulli(_rng,0.5);
	  if (r_geneal_part==0) 
	    {
	      coaltime = gsl_ran_flat(_rng,tmin,0);
	      log_proposal_den += log(1/(0-tmin));
	    } else {
	      coaltime = gsl_ran_flat(_rng,0,tmax);	      
	      log_proposal_den += log(1/(tmax-0));
	    }
	} else {
	  coaltime = gsl_ran_flat(_rng,tmin,tmax);
	  log_proposal_den += log(1/(tmax-tmin));
	}
       if (coaltime<0) 
	{
	  nstemcell_new--;
	}
     log_proposal_den += log(2./(no*(no-1)));
    }   


  Coalnode* node1;
  Coalnode* node2;
  
  // Ordering the list to prepare move attempt
  if (increase == true)
    {
      currentnode = last_internal_node;
      while (currentnode->child_left() != last_internal_node) 
	{
	  currentnode = currentnode->next(); 
	}
      if (currentnode != last_internal_node->next())
	{
	  if (_lastnode == currentnode) 
	    {
	      _lastnode = currentnode->prev();
	    }
	  currentnode->moveafter(last_internal_node);
	}
      node1 = last_internal_node;
      node2 = node1->next();
    } else { // increase == false
      int choose_root1 = gsl_rng_uniform_int(_rng, no);
      int choose_root2 = gsl_rng_uniform_int(_rng, no-1);
      if (choose_root2 >= choose_root1)
	{
	  choose_root2++;
	}
      if (choose_root1 > choose_root2) 
	{
	  int tmp = choose_root2;
	  choose_root2 = choose_root1;
	  choose_root1 = tmp;
	}
      currentnode = currentnode->next();
      for (int n=0; n<choose_root1 ; n++)
	{
	  currentnode = currentnode->next();
	}
      Coalnode* tobemoved = currentnode;
      currentnode = currentnode->next();
      tobemoved->moveafter(last_internal_node);
      for (int n=choose_root1+1; n<choose_root2 ; n++)
	{
	  currentnode = currentnode->next();
	}
      if (_lastnode == currentnode) 
	{
	  _lastnode = currentnode->prev();
	}
      currentnode->moveafter(last_internal_node); 
      node1 = last_internal_node->next();
      node2 = node1->next();     
    }

  double log_seqlik_num = 0;
  double log_seqlik_den = 0;
  double log_treelik_num = 0;
  double log_treelik_den = 0;
  
 // Move attempt & likelihood computations
  if (increase == true) {
    log_seqlik_den += log(node1->sumLikelihoodNode());
    log_treelik_den += this->computeLogLikTreeN();
    node1->increaseNbTermStemCell();    
    _nstemcell_sampled = nstemcell_new;
    log_seqlik_num += log(node1->sumLikelihoodNode()) + log(node2->sumLikelihoodNode());
    log_treelik_num += this->computeLogLikTreeN();
  } else { // increase == false
    log_seqlik_den += log(node1->sumLikelihoodNode()) + log(node2->sumLikelihoodNode());
    log_treelik_den += this->computeLogLikTreeN();
    node1->decreaseNbTermStemCell(coaltime);
    _nstemcell_sampled = nstemcell_new;
    log_seqlik_num += log(node1->sumLikelihoodNode());
    log_treelik_num += this->computeLogLikTreeN();
  }

  double logratio = (log_proposal_num + log_seqlik_num + log_treelik_num) 
    - (log_proposal_den + log_seqlik_den + log_treelik_den);
  
  //cout << "no " << no << " increase " << increase << " ratio " << exp(logratio) << endl;
  //cout << "log_proposal_num " << log_proposal_num << " log_seqlik_num " << log_seqlik_num << " log_treelik_num " << log_treelik_num << " log_proposal_den " << log_proposal_den << " log_seqlik_den " << log_seqlik_den << " log_treelik_den " << log_treelik_den << endl; 

  double r = log(gsl_ran_flat(_rng, 0, 1));
  if (r < logratio && !fake)
    {
      //cout << "updateTermStemCell Accepted" << endl;
      if (increase == true) 
	{
	  _nb_internal_nodes--;
	  node1->posteriorSimulNode();
	  //  cout << "Increasing" << endl;
	} else {
	  _nb_internal_nodes++;
	  node1->posteriorSimulNode();
	  // cout << "Decreasing" << endl;
	}
    } else {
      //cout << "updateTermStemCell Rejected" << endl;
      if (increase == true) 
	{
	  //cout << "Increase rejected" << endl;
	  node1->decreaseNbTermStemCell(coaltime);
	} else {
	  //cout << "Decrease rejected" << endl;
	  node1->increaseNbTermStemCell();
	}      
      _nstemcell_sampled = nstemcell_old;
    }
  //cout << "End of updateTermStemCell crypt " << _crypt_id << endl;

}


double Crypt::computeLogLikSeq()
{
  double ret_val = 0;
  Coalnode* currentnode = _firstnode;
  while (currentnode != NULL)
    {
      ret_val += currentnode->loglikSeqChildren();
      currentnode = currentnode->next();
    }
  return ret_val;
}

double Crypt::computeLogLikTree(Coalnode* diffroot1, Coalnode* diffroot2)
{
  vector<double> diff_geneal_coaltimes;
  double ll_diff = 0;
  int n = 0;
  Coalnode* currentnode = _firstnode;
  while (currentnode->node_status() == 0)
    {
      n++;
      currentnode = currentnode->next();
    }
  while (currentnode->time() < 0)
    {
      if (currentnode->ancestor()->time() > 0 
	  && (currentnode==diffroot1 || currentnode==diffroot2)) 
	{
	  diff_geneal_coaltimes.clear();
	  currentnode->getChildCoaltimes(diff_geneal_coaltimes);
	  sort(diff_geneal_coaltimes.begin(),diff_geneal_coaltimes.end());
	  ll_diff += this->loglikDiffCoal(diff_geneal_coaltimes);
	}
      n--;
      currentnode = currentnode->next();      
    }
  
  if (n != _nstemcell_sampled) 
    {
      this->printCrypt();
      this->printInfo();
      cerr << "n=" << n << " _nstemcell_sampled=" << _nstemcell_sampled << endl;
      cerr << "_g=" << _g << " _lambda=" << _lambda << endl;
      cerr << "mutmodel->alpha_diff() " << _mut_model->alpha_diff() << endl;
      //cerr << "mutmodel->printContextDepParams(cerr) " << _mut_model->printContextDepParams(cerr) << endl;
      this->printTree4R(cerr, string("treewithpb"));
      cerr << "pb 1 in Crypt::computeLogLikTree(Coalnode* diffroot1, Coalnode* diffroot2)" << endl;
      int*p =NULL;
      *p=1;
    }

  double ll_stemcells_times = 0;
  double ll_stemcells_combin = 0;
  while (currentnode->node_status() == 1)
    {
      double coaltime = currentnode->time() - currentnode->prev()->time();
      if (currentnode->prev()->time()<0) 
	{
	  coaltime = currentnode->time() - 0;
	}
      double coalrate = this->coalratefunc(n);
      ll_stemcells_times += log(coalrate)-coalrate*coaltime; //log(gsl_ran_exponential_pdf(coaltime, 1/coalrate));
      ll_stemcells_combin += log(2./(n*(n-1)));
      n--; 
      currentnode = currentnode->next();
    }
  if (n>1) 
    {
      double coaltime = currentnode->time() - currentnode->prev()->time();
      if (currentnode->prev()->time()<0) 
	{
	  coaltime = currentnode->time() - 0;
	}
      double coalrate = this->coalratefunc(n);
      ll_stemcells_times += -coalrate*coaltime; //log(gsl_cdf_exponential_Q(coaltime, 1/coalrate));
    }

  //cout << "computeLogLikTree() " << ret_val << endl;
  double ret_val = ll_stemcells_times + ll_stemcells_combin + ll_diff;
  return ret_val;
}

double Crypt::computeLogLikTree()
{
  vector<double> diff_geneal_coaltimes;
  double ll_diff = 0;
  int n = 0;
  Coalnode* currentnode = _firstnode;
  while (currentnode->node_status() == 0)
    {
      n++;
      currentnode = currentnode->next();
    }
  while (currentnode->time() < 0)
    {
      if (currentnode->ancestor()->time() > 0) 
	{
	  diff_geneal_coaltimes.clear();
	  currentnode->getChildCoaltimes(diff_geneal_coaltimes);
	  sort(diff_geneal_coaltimes.begin(),diff_geneal_coaltimes.end());
	  ll_diff += this->loglikDiffCoal(diff_geneal_coaltimes);
	}
      n--;
      currentnode = currentnode->next();      
    }
  
  if (n != _nstemcell_sampled) 
    {
      this->printCrypt();
      this->printInfo();
      cerr << "n=" << n << " _nstemcell_sampled=" << _nstemcell_sampled << endl;
      this->printTree4R(cerr, string("treewithpb"));
      cerr << "pb 1 in Crypt::computeLogLikTree()" << endl;
      int*p =NULL;
      *p=1;
    }

  double ll_stemcells_times = 0;
  double ll_stemcells_combin = 0;
  while (currentnode->node_status() == 1)
    {
      double coaltime = currentnode->time() - currentnode->prev()->time();
      if (currentnode->prev()->time()<0) 
	{
	  coaltime = currentnode->time() - 0;
	}
      double coalrate = this->coalratefunc(n);
      ll_stemcells_times += log(coalrate)-coalrate*coaltime; //log(gsl_ran_exponential_pdf(coaltime, 1/coalrate));
      ll_stemcells_combin += log(2./(n*(n-1)));
      n--; 
      currentnode = currentnode->next();
    }
  if (n>1) 
    {
      double coaltime = currentnode->time() - currentnode->prev()->time();
      if (currentnode->prev()->time()<0) 
	{
	  coaltime = currentnode->time() - 0;
	}
      double coalrate = this->coalratefunc(n);
      ll_stemcells_times += -coalrate*coaltime; //log(gsl_cdf_exponential_Q(coaltime, 1/coalrate));
    }

  //cout << "computeLogLikTree() " << ret_val << endl;
  double ret_val = ll_stemcells_times + ll_stemcells_combin + ll_diff;
  return ret_val;
}

double Crypt::computeLogLikTreeN1()
{
  double ret_val = 0;
  int n = 0;
  Coalnode* currentnode = _firstnode;
  while (currentnode->node_status() != 2)
    {
      if (currentnode->node_status()==0)
	{
	  n++;
	} else if (currentnode->node_status()==1) {
	  double coalrate = this->coalratefunc(n);
	  double coaltime = currentnode->time() - currentnode->prev()->time();
	  ret_val += log(coalrate)-coalrate*coaltime;
	  n--;
	}      
      currentnode = currentnode->next();
    }
  if (n>1) 
    {
      double coaltime = currentnode->time() - currentnode->prev()->time();
      if (currentnode->prev()->time()<0) 
	{
	  coaltime = currentnode->time() - 0;
	}
      double coalrate = this->coalratefunc(n);
      ret_val += -coalrate*coaltime; //log(gsl_cdf_exponential_Q(coaltime, 1/coalrate));
    }
  
  return(ret_val);
}

double Crypt::computeLogLikTreeN()
{
  double ret_val = 0;
  if (_nstemcell_sampled > _N_stem_cells)
    {
      return log((double)0);
    }

  ret_val += this->computeLogLikTree();
  ret_val += -_ncells*log((double)_N_stem_cells); 
  for (int i=1; i<=_nstemcell_sampled; i++) 
    {
      ret_val += log((double)(_N_stem_cells-i+1));
    }  
  return ret_val;
}

double Crypt::computeLogLikN()
{
  // _ncell_byseq and _nstemcell_byseq must be unchanged
  double ret_val = 0;
  if (_nstemcell_sampled > _N_stem_cells)
    {
      return log((double)0);
    }

  ret_val += -_ncells*log((double)_N_stem_cells); 
  for (int i=1; i<=_nstemcell_sampled; i++) 
    {
      ret_val += log((double)(_N_stem_cells-i+1));
    }  
  //cout << "+sampling " << ret_val << endl;
  return ret_val;
}

void Crypt::drawSeqGenealogyPrior()
{
  // Forward sequence simulation

  Coalnode* currentnode = _lastnode;
  while (currentnode != NULL)
    {
      currentnode->drawSeq();
      currentnode = currentnode->prev();
    }   
}

void Crypt::updateSeq()
{
  Coalnode* currentnode = _firstnode;
  while (currentnode != NULL)
    {
      currentnode->posteriorSimulNode();
	
      currentnode = currentnode->next();
    }
 
}

void Crypt::printInfo()
{
  cout << "*********** INFO **********************" << endl;
  cout << "_crypt_id " << _crypt_id << endl;
  cout << "_nstemcell_sampled " << _nstemcell_sampled << endl;
  cout << "_N_stem_cells " << _N_stem_cells << endl;
  //cout << "_nseq " << _nseq << endl;

  cout << "_cells_obs_seqs _cells_ori_seqs" << endl;
  for (int i=0 ; i<_ncells ; i++)
    {
      for (int ipos=0 ; ipos<_lseq ; ipos++)
	{
	  cout << _cells_obs_seqs[i][ipos];
	}
      cout << " (obs)"<< endl;
    }

  cout << "Genealogy " << endl;
  Coalnode* currentnode = _firstnode;
  while (currentnode != NULL)
    {
      cout << "-----" << endl;
      currentnode->printInfo();
      currentnode = currentnode->next();
    }
  cout << "*********** END INFO **********************" << endl;
}

void Crypt::deleteGenealogy()
{
  Coalnode* tmp;
  Coalnode* current_node;
  
  current_node = _firstnode;
  while (current_node != NULL)
    {
      tmp = current_node;
      current_node = current_node->next();
      delete tmp;
    }
  _firstnode = NULL;
  _lastnode = NULL;

  current_node = _node_collector;
  while (current_node != NULL)
    {
      tmp = current_node;
      current_node = current_node->next();
      delete tmp;
    }
  _node_collector = NULL;  
}

StatWithinCrypt* Crypt::summaryStat(bool simulseq)
{
  StatWithinCrypt* ret_val = new StatWithinCrypt;
  
  ret_val->_icrypt = _crypt_id;
  ret_val->_nb_cells = _ncells;
  
  map<vector<int>, int> seq_counts;
  if (!simulseq)
    {
      for (int ic=0; ic<_ncells; ic++)
	{
	  vector<int> currseq(_lseq);
	  for (int i=0; i<_lseq; i++) 
	    {
	      currseq[i] = _cells_obs_seqs[ic][i];
	    }      
	  map<vector<int>, int>::iterator it_seq = seq_counts.find(currseq);
	  if (it_seq == seq_counts.end())
	    {
	      seq_counts.insert(pair<vector<int>, int>(currseq,1));
	    } else {
	      (*it_seq).second++;
	    }
	}
    } else {
      cerr << "Crypt::summaryStat(bool simulseq==true) currently unsupported" << endl;
      exit(1);
    }
 
  map<vector<int>, int> seq_counts2; // without data at position pointed in _missing
  for (map<vector<int>, int>::iterator it_seq = seq_counts.begin(); 
       it_seq != seq_counts.end() ; it_seq++)
    {
      vector<int> newseq;
      newseq.clear();
      for (int i=0; i<_lseq; i++) 
	{
	  if (!_missing[i]) 
	    {
	      newseq.push_back((*it_seq).first[i]);
	    }
	}
      map<vector<int>, int>::iterator it_seq2 = seq_counts2.find(newseq);
      if (it_seq2 == seq_counts2.end())
	{
	  seq_counts2.insert(pair<vector<int>, int>(newseq,(*it_seq).second));
	} else {
	  (*it_seq2).second += (*it_seq).second;
	}	        
    }
  seq_counts = seq_counts2;



  ret_val->_nb_tags = seq_counts.size();
  int effective_lseq = seq_counts.begin()->first.size();

  int max_methyl = 0;
  int min_methyl = effective_lseq; 
  int sum_methyl = 0;
  int n_methyl = 0;
  int nb_occ_max = 0;
  int nb_singletons = 0;
  int max_dist = 0;
  double sum_dist = 0;
  double sum_sqdist = 0;
  int nb_unmeth = 0;

  vector<bool> polym(effective_lseq,false);
  vector<int> nbmeth_bypos(effective_lseq,0);

  for (map<vector<int>, int>::iterator it_seq = seq_counts.begin(); 
       it_seq != seq_counts.end() ; it_seq++)
    {
      if ((*it_seq).second > nb_occ_max)
	{
	  nb_occ_max = (*it_seq).second;
	}
      if ((*it_seq).second == 1)
	{
	  nb_singletons++;
	}
      int n=0;
      int m=0;
      for (int i=0; i<effective_lseq;i++)
	{
	  if ((*it_seq).first[i]==1)
	    {
	      if ((*it_seq).first[i] != _cells_obs_seqs[0][i])
		{
		  polym[i] = true;
		}
	      nbmeth_bypos[i] += (*it_seq).second;
	      n++;
	      m++;
	    } else if ((*it_seq).first[i]==0) {
	      if ((*it_seq).first[i] != _cells_obs_seqs[0][i])
		{
		  polym[i] = true;
		}
	      n++;
	    }
	}
      if (m > max_methyl)
	{
	  max_methyl = m;
	}
      if (m < min_methyl)
	{
	  min_methyl = m;
	}
      if (m == 0)
	{
	  nb_unmeth = (*it_seq).second;
	}
      sum_methyl += m*(*it_seq).second;
      n_methyl += n*(*it_seq).second;
      map<vector<int>, int>::iterator it_seq2 = it_seq;
      it_seq2++;
      while (it_seq2 != seq_counts.end()) 
	{
	  int dist = 0;
	  for (int i=0; i<effective_lseq; i++)
	    {
	      if ((*it_seq).first[i]>=0 && (*it_seq2).first[i]>=0 
		  && (*it_seq).first[i] != (*it_seq2).first[i])
		{
		  dist++;
		}
	    }
	  if (dist>max_dist)
	    {
	      max_dist = dist;
	    }
	  sum_dist += (*it_seq).second*(*it_seq2).second*dist;
	  sum_sqdist += (*it_seq).second*(*it_seq2).second*pow((double)dist,2);
	  it_seq2++;
	}
    }
  ret_val->_max_methyl = max_methyl;
  ret_val->_min_methyl = min_methyl;
  ret_val->_nb_unmeth = nb_unmeth; 
  ret_val->_ave_methyl = (double)sum_methyl/(double)n_methyl;
  ret_val->_nb_occ_max = nb_occ_max;
  ret_val->_nb_singletons = nb_singletons;
  ret_val->_max_dist = max_dist;
  ret_val->_ave_dist = (2*sum_dist)/(_ncells*(_ncells-1));
  ret_val->_ave_sqdist = sqrt((2*sum_sqdist)/(_ncells*(_ncells-1)));
  int nb_polym = 0;
  for (int i=0; i<effective_lseq; i++)
    {
      if (polym[i])
	{
	  nb_polym++;
	}
    }
  ret_val->_nb_polym_sites = nb_polym;

  int nb_100_meth = 0;
  int nb_80_meth = 0;
  for (int i=0; i<effective_lseq; i++)
    {
      if (nbmeth_bypos[i]==_ncells) 
	{
	  nb_100_meth++;
	}
      if (nbmeth_bypos[i]>= (double)_ncells*0.8)
	{
	  nb_80_meth++;
	}
    }
  ret_val->_nb_100_meth = nb_100_meth;
  ret_val->_nb_80_meth = nb_80_meth;
  
  return ret_val;
  
}

void Crypt::printTree4R(ostream& out, string tagtree)
{
  map<Coalnode*, int> indexes;
  Coalnode* current_node = _firstnode;
  int inode = -1;

  vector<vector<int> > labels;
  vector<double> order;
  while (current_node->node_status()==0)
    {
      labels.push_back(current_node->seq());
      order.push_back(-inode);
      indexes.insert(pair<Coalnode*, int>(current_node, inode));
      inode--;
      current_node = current_node->next();
    }
  inode = 1;
  
  vector<double> heights;
  vector<int> merge_left;
  vector<int> merge_right;


  
  while (current_node != NULL)
    {
      //      if (current_node->node_status() == 2 && current_node->child_left() != current_node->prev())
      if (current_node->node_status() == 2 && current_node->prev()->node_status() == 2)
	{
	  indexes.insert(pair<Coalnode*, int>(current_node, inode));
	  map<Coalnode*, int>::iterator it_left = indexes.find(current_node->child_left());
	  merge_left.push_back((*it_left).second);
	  
	  if (current_node->prev()->prev()->node_status() != 2)
	    {
	      map<Coalnode*, int>::iterator it_right = indexes.find(current_node->prev()->child_left());
	      merge_right.push_back((*it_right).second);	      
	    } else {
	      merge_right.push_back(inode-1);
	    } 

	  heights.push_back(current_node->time());
	  inode++;
	}
      if (current_node->node_status() != 2)
	{
	  indexes.insert(pair<Coalnode*, int>(current_node, inode));
	  if (current_node != _lastnode) 
	    {	  
	      map<Coalnode*, int>::iterator it_left = indexes.find(current_node->child_left());
	      map<Coalnode*, int>::iterator it_right = indexes.find(current_node->child_right());
	      merge_right.push_back((*it_right).second);
	      merge_left.push_back((*it_left).second);
	      heights.push_back(current_node->time());
	    }
	  inode++;
	}
      current_node = current_node->next();
    }
 
  out << tagtree << " <- list(";
  out << "merge=rbind(";
  for (unsigned int i=0; i<merge_left.size(); i++)
    {
      if (i>0)
	{
	  out << ",";
	}
      out << "c(" << merge_left[i] << "," << merge_right[i] << ")";
    }
  out << "), ";
  out << "height=c(";
  for (unsigned int i=0; i<merge_left.size(); i++)
    {
      if (i>0)
	{
	  out << ",";
	}
      out << heights[i];
    }
  out << "), ";
  out << "order=c(";
  for (unsigned int i=0; i<order.size(); i++)
    {
      if (i>0)
	{
	  out << ",";
	}
      out << order[i];
    }
  out << "), ";
  out << "labels=c(";
  for (unsigned int i=0; i<labels.size(); i++)
    {
      if (i>0)
	{
	  out << ",";
	}
      out << "\"";
      for (int j=0; j<_lseq; j++)
	{
	  out << labels[i][j];
	}
      out << "\"";
    }
  out << "), ";
  out << "method = \"mymethod\", call = \"mycall\")" << endl;
  
}

void Crypt::printCoaltimes()
{
   Coalnode* currentnode = _firstnode;
   cout << "coaltimes ";
   while (currentnode != _lastnode) 
    {
      if (currentnode->node_status()==1) 
	{
	  cout << currentnode->time() << " ";
	}
      currentnode = currentnode->next();
    }
  cout << endl;
 
}

void Crypt::updateTreeWB(bool updateN, int* nb_stemcells_counts,
			 int& Nmin, int& Nmax, 
			 int nstemcells_prior_min, int nstemcells_prior_max)
{
  // inplement Wilson Balding MH step to update the genealogical tree

  int Npriormin;
  int Npriormax;
  this->computePriorProbaN(nb_stemcells_counts, Nmin, Nmax, Npriormin, Npriormax, nstemcells_prior_min, nstemcells_prior_max);
  
  if (0 && Nmin < 5)
    {
      cout << "*****************" << endl;
      cout << "nb_stemcells_counts Nmin " << Nmin << " Nmax " << Nmax << " ";
      for (int i=nstemcells_prior_min; i<nstemcells_prior_max; i++) 
	{
	  cout << i << ":" << nb_stemcells_counts[i] << " ";
	}
      cout << endl;
      cout << "prior_proba_N with N = " << _N_stem_cells << " ";
      for (int i=Npriormin; i<=Npriormax; i++) 
	{
	  cout << i << ":" << _prior_proba_N[i] << " ";
	}
      cout << endl;
    }

  Coalnode* currentnode = _firstnode;
  Coalnode* chosen_node = NULL;

  if (_nb_internal_nodes + _ncells <= 2)
    {
      return;
    }
  
  int choose_node = gsl_rng_uniform_int(_rng, _ncells 
					+ _nb_internal_nodes);
  int rbidon = gsl_ran_bernoulli(_rng,0.5);
  if (0 && rbidon==0) 
    {
      int nmax = 5;
      if (nmax > _ncells+_nb_internal_nodes)
	{
	  nmax = _ncells+_nb_internal_nodes;
	}
      choose_node = _ncells+_nb_internal_nodes-1
	- gsl_rng_uniform_int(_rng, nmax);
    }


  int nbliving = _ncells;
  int inode = -1;
  currentnode = _firstnode;
  while (currentnode != NULL && chosen_node == NULL)
    {
      inode++;
      if (currentnode->node_status() == 1) 
	{
	  nbliving--;
	}	     
      if (inode == choose_node)
	{
	  chosen_node = currentnode;
	}
      currentnode = currentnode->next();
    }

  Coalnode* to_be_moved = chosen_node->ancestor();
  if (to_be_moved->node_status()==2)
    {
      return;
    }
  if (to_be_moved->child_left() != chosen_node)
    {
      to_be_moved->swap_childs();
    }
  if (to_be_moved->ancestor()->child_left() != to_be_moved)
    {
      to_be_moved->ancestor()->swap_childs();
    }

  double tmin = chosen_node->time();
  if (tmin < to_be_moved->child_right()->time())
    {
      tmin = to_be_moved->child_right()->time();
    }

  _target_wb[0] = to_be_moved->child_right();
  int ntargets = 1;
  currentnode = chosen_node->next();
  while (currentnode != NULL)
    {
      if (currentnode->node_status() != 0 && currentnode != to_be_moved)
	{
	  if (currentnode->node_status() != 2) 
	    {
	      if (currentnode->child_left() != to_be_moved)
		{
		  _target_wb[ntargets] = currentnode->child_left();
		  ntargets++;
		}
	      _target_wb[ntargets] = currentnode->child_right();
	      ntargets++;
	    } else {
	      if (currentnode->child_left() != to_be_moved)
		{
		  _target_wb[ntargets] = currentnode->child_left();
		  ntargets++;
		}
	    }	  
	}
      currentnode = currentnode->next();
    }
  double sum = 0;
  for (int itarget=0; itarget<ntargets; itarget++)
    {
      _ptarget_wb[itarget] = 1/(1+chosen_node->distseq(_target_wb[itarget]));
      sum += _ptarget_wb[itarget];
    }
  double rchoose = gsl_ran_flat(_rng,0,1);
  double sumpar = 0;
  int ichoose = -1;
  while(sumpar < rchoose && ichoose < ntargets-1) 
    {
      ichoose++;
      _ptarget_wb[ichoose] /= sum;
      sumpar += _ptarget_wb[ichoose];
    }
  if (ichoose == -1) 
    {
      ichoose = 0;
    }
  if (sumpar < rchoose) 
    {
      cout << "ichoose would have been == ntarget!" << endl;
    }

  Coalnode* target = _target_wb[ichoose];

  double lik_ratio_proposal = log(_ptarget_wb[0])-log(_ptarget_wb[ichoose]);
  double t1 = to_be_moved->time();
  double t1min = chosen_node->time();
  if (t1min < to_be_moved->child_right()->time())
    {
      t1min = to_be_moved->child_right()->time();
    }
  double t1max = to_be_moved->ancestor()->time();
  if (1) {
  if (t1min < 0 && t1max > 0) 
    {
      lik_ratio_proposal += log(0.5);
      if (t1<0) 
	{
	  lik_ratio_proposal += log(1/(0-t1min));
	} else if (t1>0 && to_be_moved->ancestor()->node_status()==2 
		   && to_be_moved->ancestor()->next()==NULL
		   && to_be_moved->ancestor()->prev()->node_status()!=2) {
	  
	  double coalrate = this->coalratefunc(2);
	  lik_ratio_proposal += log(gsl_ran_exponential_pdf(t1-0, 1/coalrate));
	} else {
	  lik_ratio_proposal += log(1/(t1max-0));      	  
	}
    } else {
      if (t1min > 0 && to_be_moved->ancestor()->node_status()==2 
	  && to_be_moved->ancestor()->next()==NULL
	  && to_be_moved->ancestor()->prev()->node_status()!=2) {	  
	  double coalrate = this->coalratefunc(2);
	  lik_ratio_proposal += log(gsl_ran_exponential_pdf(t1-t1min, 1/coalrate));
      } else {
	  lik_ratio_proposal += log(1/(t1max-t1min));      	  	
      }
    }
  } else {
    if (to_be_moved->ancestor()->node_status()==2 
	&& to_be_moved->ancestor()->next()==NULL
	&& to_be_moved->ancestor()->prev()->node_status()!=2) {	  
      double coalrate = this->coalratefunc(2);
      lik_ratio_proposal += log(gsl_ran_exponential_pdf(t1-t1min, 1/coalrate));
    } else {
      lik_ratio_proposal += log(1/(t1max-t1min));      	  	
    }    
  }
  int nbstemcell_sampled1 = _nstemcell_sampled;  

  double t2 = -1;
  double t2min = chosen_node->time();  
  if (t2min < target->time())
    {
      t2min = target->time();
    }
  double t2max = target->ancestor()->time();
  bool special_case = (target->ancestor()->node_status()==2 
		       && target->ancestor()->next()==NULL
		       && target->ancestor()->prev()->node_status()!=2);
  if (ichoose == 0) 
    {
      t2max = target->ancestor()->ancestor()->time();
      special_case = (target->ancestor()->ancestor()->node_status()==2 
		      && target->ancestor()->ancestor()->next()==NULL
		      && target->ancestor()->ancestor()->prev()->node_status()!=2);
    }

  if (t2min > t2max) 
    {
      cout << "pb: t2min > t2max" << endl;
      this->printInfo();
    }

  if (1) {
  if (t2min<0 && t2max>0) 
    {
      lik_ratio_proposal -= log(0.5);
      int choose_genealogy_part = gsl_rng_uniform_int(_rng, 2);
      if (choose_genealogy_part==0)
	{
	  t2 = gsl_ran_flat(_rng,t2min,0);
	  lik_ratio_proposal -= log(1/(0-t2min));
	  if (t2==t2min) { // gsl_ran_flat draw a number >= t2min and < t2max
	    return;
	  }
	} else if (choose_genealogy_part==1) {
	  if (special_case) 
	    {
	      double coalrate = this->coalratefunc(2);
	      t2 = 0 + gsl_ran_exponential(_rng, 1/coalrate);
	      lik_ratio_proposal -= log(gsl_ran_exponential_pdf(t2-0, 1/coalrate));
	      if (t2==0 || t2>t2max) 
		{
		  return;
		}
	    } else {
	      t2 = gsl_ran_flat(_rng,0,t2max);
	      lik_ratio_proposal -= log(1/(t2max-0));
	      if (t2==0) {
		return;
	      }
	    }
	}
    } else {
      if (t2min > 0 && special_case) 
	{
	  double coalrate = this->coalratefunc(2);
	  t2 = t2min + gsl_ran_exponential(_rng, 1/coalrate);	  
	  lik_ratio_proposal -= log(gsl_ran_exponential_pdf(t2-t2min, 1/coalrate));
	  if (t2==t2min || t2>t2max) 
	    {
	      return;
	    }
	} else {
	  t2 = gsl_ran_flat(_rng,t2min,t2max);
	  lik_ratio_proposal -= log(1/(t2max-t2min));
	  if (t2==t2min) {
	    return;
	  }
	}
    }
  } else {
    if (special_case) 
      {
	double coalrate = this->coalratefunc(2);
	t2 = t2min + gsl_ran_exponential(_rng, 1/coalrate);	  
	lik_ratio_proposal -= log(gsl_ran_exponential_pdf(t2-t2min, 1/coalrate));
	if (t2==t2min || t2>t2max) 
	  {
	    return;
	  }
      } else {
	t2 = gsl_ran_flat(_rng,t2min,t2max);
	lik_ratio_proposal -= log(1/(t2max-t2min));
	if (t2==t2min) {
	  return;
	}
      }    
  }


  int nbstemcell_sampled2 = _nstemcell_sampled;
  if (t1<0 && t2>0) 
    {
      nbstemcell_sampled2++;
    } else if (t2<0 && t1>0) {
      nbstemcell_sampled2--;
    }

  if (0 && Nmin < 5) 
    {
      cout << "nbstemcell_sampled1=" << nbstemcell_sampled1 
	   << " nbstemcell_sampled2=" << nbstemcell_sampled2 << endl;
    }

  if (!this->checkNodeOrder())
    {
      cerr << "checkNodeOrder() failed before WB move" << endl;
      this->printInfo();
    }
  //cout << "**************" << endl;
  //this->printInfo();

  double lik_ratio_tree = 0;
  //  double lik_ratio_tree_test = 0;
  //double lik_ratio_tree1 = this->computeLogLikTreeN();
  //cout << "lik_ratio_tree1 " << lik_ratio_tree1 << endl;
  double lik_ratio_tree_den = 0;
  if (!updateN) 
    {
      Coalnode* diffroot1 = chosen_node->diffLineageRoot();
      Coalnode* diffroot2 = target->diffLineageRoot();
      if (t2>0) {
	diffroot2 = NULL;
      }
      lik_ratio_tree_den = this->computeLogLikTree(diffroot1,diffroot2)+this->computeLogLikN();
      //      lik_ratio_tree_den = this->computeLogLikTreeN();
    } else {
      double sum = 0;
      int N1 = _N_stem_cells;
      for (int i=Npriormin; i<=Npriormax; i++) 
	{
	  _N_stem_cells = i;
	  sum += _prior_proba_N[i]*exp(this->computeLogLikN());
	}
      _N_stem_cells = N1;
      //lik_ratio_tree_den = this->computeLogLikTree()+log(sum); 
      Coalnode* diffroot1 = chosen_node->diffLineageRoot();
      Coalnode* diffroot2 = target->diffLineageRoot();
      if (t2>0) {
	diffroot2 = NULL;
      }
      lik_ratio_tree_den = this->computeLogLikTree(diffroot1,diffroot2)+log(sum);             
    }
  //cout << "--------------" << endl;
  double lik_ratio_seq = 0;
  lik_ratio_seq -= log(to_be_moved->sumLikelihoodNode());
  if (ichoose != 0) 
    {
      lik_ratio_seq -= target->loglikAncestor();
    }
  to_be_moved->splitTreeWB();
  if (target->ancestor()->child_left() != target)
    {
      target->ancestor()->swap_childs(); 
    }
  to_be_moved->joinTreeWB(target, t2);
  _nstemcell_sampled = nbstemcell_sampled2;  


  double lik_ratio_tree_num = 0;
  if (!updateN) 
    {
      Coalnode* diffroot1 = chosen_node->diffLineageRoot();
      Coalnode* diffroot2 = _target_wb[0]->diffLineageRoot();
      if (t1>0) {
	diffroot2 = NULL;
      }
      lik_ratio_tree_num = this->computeLogLikTree(diffroot1,diffroot2)+this->computeLogLikN();
      //lik_ratio_tree_num = this->computeLogLikTreeN();
    } else {
      double sum = 0;
      int N1 = _N_stem_cells;
      for (int i=Npriormin; i<=Npriormax; i++) 
	{
	  _N_stem_cells = i;
	  _posterior_proba_N[i] = _prior_proba_N[i]*exp(this->computeLogLikN());
	  sum += _posterior_proba_N[i];
	}
      _N_stem_cells = N1;
      //lik_ratio_tree_num = this->computeLogLikTree()+log(sum);      
      Coalnode* diffroot1 = chosen_node->diffLineageRoot();
      Coalnode* diffroot2 = _target_wb[0]->diffLineageRoot();
      if (t1>0) {
	diffroot2 = NULL;
      }
      lik_ratio_tree_num = this->computeLogLikTree(diffroot1,diffroot2)+log(sum);
      for (int i=Npriormin; i<=Npriormax; i++) 
	{
	  _posterior_proba_N[i]/=sum;
	}
      if (0 && Nmin < 5) 
	{
	  cout << "posterior_proba_N " ;
	  for (int i=Npriormin; i<=Npriormax; i++) 
	    {
	      cout << i << ":" <<_posterior_proba_N[i] << " ";
	    }	  
	  cout << endl;
	}
    }

  //double lik_ratio_tree2 = this->computeLogLikTreeN();
  //cout << "lik_ratio_tree2 " << lik_ratio_tree2 << endl;
  lik_ratio_tree = lik_ratio_tree_num-lik_ratio_tree_den;
//   if (updateN && (lik_ratio_tree-lik_ratio_tree_test>0.0001 || lik_ratio_tree-lik_ratio_tree_test<-0.0001)) {
//     cout << "lik_ratio_tree " << lik_ratio_tree << " " << lik_ratio_tree_test << endl;
//   }
  lik_ratio_seq += log(to_be_moved->sumLikelihoodNode());
  if (ichoose != 0) 
    {
      lik_ratio_seq += _target_wb[0]->loglikAncestor();
    }

  double r = gsl_ran_flat(_rng,0,1);
  
  //cout << "lik_ratio_proposal=" << lik_ratio_proposal << " lik_ratio_seq=" << lik_ratio_seq << " lik_ratio_tree=" << lik_ratio_tree << " total p=" << exp(lik_ratio_proposal + lik_ratio_seq + lik_ratio_tree) << endl; 
  //lik_ratio_seq = 0; //just for testing
  //lik_ratio_tree = 0; //just for testing
  //cout << "nbstemcell_sampled " << nbstemcell_sampled1 << " -> " 
  //   << nbstemcell_sampled2 << endl;
  //cout << "log(alpha) " << lik_ratio_proposal + lik_ratio_seq + lik_ratio_tree << endl;
  //cout << "lik_ratio_proposal " << lik_ratio_proposal << endl;
  //cout << "lik_ratio_tree " << lik_ratio_tree << endl;
  if (log(r) < lik_ratio_proposal + lik_ratio_seq + lik_ratio_tree)
    {
      //cout << "Accepted" << endl;
      to_be_moved->posteriorSimulNode();
      if (updateN)   // draw new _N_stem_cells
	{
	  int N1 = _N_stem_cells;
	  double s=gsl_ran_flat(_rng,0,1);
	  double sum = _posterior_proba_N[Npriormin];
	  _N_stem_cells = Npriormin;
	  while (s>sum && _N_stem_cells<Npriormax) 
	    {
	      _N_stem_cells++;
	      sum += _posterior_proba_N[_N_stem_cells];
	    }
	  nb_stemcells_counts[N1]--;
	  nb_stemcells_counts[_N_stem_cells]++;

	  Nmin=_N_stem_cells;
	  Nmax=_N_stem_cells;
	  if (_N_stem_cells>nstemcells_prior_min 
	      && nb_stemcells_counts[_N_stem_cells-1]>0)
	    {
	      Nmin = _N_stem_cells-1;
	    }
	  if (_N_stem_cells<nstemcells_prior_max
	      && nb_stemcells_counts[_N_stem_cells+1]>0)
	    {
	      Nmax = _N_stem_cells+1;
	    }
	}
      //cout << "new _N_stem_cells " << _N_stem_cells << " Nmin " << Nmin << " Nmax " << Nmax << endl;
    } else {
      //cout << "Rejected" << endl;
      to_be_moved->splitTreeWB();
      if (_target_wb[0]->ancestor()->child_left() != _target_wb[0])
	{
	  _target_wb[0]->ancestor()->swap_childs(); 
	}
      to_be_moved->joinTreeWB(_target_wb[0], t1);
      _nstemcell_sampled = nbstemcell_sampled1;
    }  
  if (!this->checkNodeOrder())
    {
      cerr << "checkNodeOrder() failed" << endl;
      this->printInfo();
      double *x=NULL;
      *x=1;
      exit(1);
    }
}


void Crypt::checkSeqLikComputations()
{
  cout << "--------" << endl;
  cout << "Crypt::checkSeqLikComputations()" << endl;
  Coalnode* currentnode = _firstnode;
  while (currentnode != NULL) 
    {
      cout << "_node_id " << currentnode->node_id() << endl;
      cout << "likelihoodNode() " << currentnode->likelihoodNode() <<  endl;
      cout << "sumLikelihoodNode() " << currentnode->sumLikelihoodNode() <<  endl;
      cout << "loglikSeqChildren() " << currentnode->loglikSeqChildren() <<  endl;
      cout << "loglikAncestor() " << currentnode->loglikAncestor() <<  endl;
      currentnode = currentnode->next();      
    }
  cout << "--------" << endl;

}

bool Crypt::checkNodeOrder()
{
  Coalnode* currentnode = _firstnode;
  while (currentnode->next() != NULL) 
    {
      if (currentnode->time() > currentnode->next()->time()) 
	{
	  return false;
	}
      currentnode = currentnode->next();
    }
  return true;
}
