/* Copyright (C) 2006 Pierre Nicolas
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <iostream>
#include <fstream>
#include <cstdlib>

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_sf_gamma.h>

#include "params.h"
#include "mcmc.h"

using namespace std;


int 
main(int argc, char** argv) 
{
  string datafilename;
  string outfilename;
  string paramfilename;

  if (argc>3)
    {
      int iarg = 1;
      while (iarg < argc)
	{
	  if (strcmp(argv[iarg], "-data") == 0) 
	    {
	      datafilename = string(argv[iarg+1]);
	      cout << "datafilename: " << datafilename << endl;	
	      iarg+=2;
	    } else if (strcmp(argv[iarg], "-params") == 0) {
	      paramfilename = string(argv[iarg+1]);	      	      
	      cout << "paramfilename: " << paramfilename << endl;	
	      iarg+=2;
	    } else if (strcmp(argv[iarg], "-out") == 0) {
	      outfilename = string(argv[iarg+1]);	      	      
	      cout << "outfilename: " << outfilename << endl;	
	      iarg+=2;
	    }
	}
      if (datafilename.size() == 0 || paramfilename.size() == 0 || outfilename.size() == 0) 
	{
	  cerr << "Usage: " << argv[0] 
	       << " -data <filename1> -params <fn2> -out <fn3>" 
	       << endl; 
	  exit(1);	  
	}
    } else {
    cerr << "Usage: " << argv[0] 
	 << " -data <filename1> -params <fn2> -out <fn3>" 
	 << endl; 
    exit(1);	  
  }
  Params params(paramfilename);
  
  mcmc sampler(params, datafilename);
  sampler.setPrior(params);
  sampler.samplingFancy(params, outfilename);
  return 1;
}

