/* Copyright (C) 2006 Pierre Nicolas
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <iostream>
#include <fstream>
#include <cstdlib>

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_sf_gamma.h>

#include "mcmc.h"
#include "params.h"

using namespace std;


int 
main(int argc, char** argv) 
{
  string datafilename;
  string paramsfilename;
  string outstatfilename;
  string outseqfilename;
  bool printseq = false;

  datafilename.clear();
  paramsfilename.clear();
  outstatfilename.clear();
  outseqfilename.clear();

  if (argc>=7)
    {
      int iarg = 1;
      while (iarg < argc)
	{
	  if (strcmp(argv[iarg], "-data") == 0) 
	    {
	      datafilename = string(argv[iarg+1]);
	      cout << "datafilename: " << datafilename << endl;	
	      iarg+=2;
	    } else if (strcmp(argv[iarg], "-params") == 0) {
	      paramsfilename = string(argv[iarg+1]);	      	      
	      cout << "paramsfilename: " << paramsfilename << endl;	
	      iarg+=2;
	    } else if (strcmp(argv[iarg], "-outstat") == 0) {
	      outstatfilename = string(argv[iarg+1]);	      	      
	      cout << "outstatfilename: " << outstatfilename << endl;	
	      iarg+=2;
	    } else if (strcmp(argv[iarg], "-outseq") == 0) {
	      outseqfilename = string(argv[iarg+1]);	      	      
	      cout << "outseqfilename: " << outseqfilename << endl;
	      printseq = true;	
	      iarg+=2;
	    } else {
	      iarg++;
	    }
	}
      if (datafilename.size() == 0 || paramsfilename.size() == 0 || outstatfilename.size() == 0) 
	{
	  cerr << "Usage: " << argv[0] 
	       << " -data <filename1> -params <fn2> -outstat <fn3> [-outseq <fn4>]" 
	       << endl; 
	  exit(1);	  
	}
    } else {
      cerr << "Usage: " << argv[0] 
	   << " -data <filename1> -params <fn2> -outstat <fn3> [-outseq <fn4>]" 
	   << endl; 
      exit(1);
    }

  Params params(paramsfilename);
  mcmc sampler(params, datafilename);
  
  sampler.simul(params._nsimul, outstatfilename, outseqfilename, printseq);
	
  return 0;
}

