/* Copyright (C) 2006 Pierre Nicolas
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <iostream>
#include <fstream>
#include <cstdlib>
#include <string>

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_sf_gamma.h>

#include "mutmodel.h"
#include "crypt.h"
#include "mcmc.h"
#include "stat_withincrypt.h"
#include "params.h"


using namespace std;



mcmc::mcmc(Params& params, string datafilename)
{

  _unique_Nlambda = params._unique_Nlambda;
  _unique_mutmodel = params._unique_mutmodel;
  _no_clusterupdate = params._no_clusterupdate;
  _unique_perror = params._unique_perror;

  this->stirlingAlloc();
  _nb_stemcells_dirichletprior = NULL;
  _nb_stemcells_pseudocounts = NULL;
  _nb_stemcells_distrib = NULL;
  _contextdep_mutmodel = params._contextdep_mutmodel;
  _rng = gsl_rng_alloc(gsl_rng_taus);
  _seed = params._seed;
  if (_seed < 0)
    {
      _seed = time(NULL); 
    }
  gsl_rng_set(_rng, _seed);
  cout << "seed " << _seed << endl;

  _sym_mutrate = params._sym_mutrate;
  
  _nclusters = params._nclusters;
  _nb_mutclass = params._nb_mutrate;
  _pclusters = new double [_nclusters];
  _pclusters_dirpseudocounts = new double [_nclusters];
  _mutrate_m = new double* [_nclusters];
  _mutrate_u = new double* [_nclusters];
  _lambda = new double [_nclusters];
  _g = new double [_nclusters];
  _perror = new double [_nclusters];
  _clusters_counts = new int [_nclusters];
  _switch_rates_inc = new double* [_nclusters];
  _switch_rates_dec = new double* [_nclusters];

  for (int ic=0; ic<_nclusters; ic++) 
    {
      _pclusters[ic] = 1./(double)_nclusters;
      _mutrate_m[ic] = new double [_nb_mutclass];
      _mutrate_u[ic] = new double [_nb_mutclass];
      _switch_rates_inc[ic] = new double [_nb_mutclass];
      _switch_rates_dec[ic] = new double [_nb_mutclass];
      for (int imut=0; imut<_nb_mutclass; imut++) 
	{
	  _mutrate_m[ic][imut] = params._mu_m;
	  _mutrate_u[ic][imut] = params._mu_u;
	  cout << "_mutrate_m["<<ic<<"]["<<imut<<"] " << _mutrate_m[ic][imut] << endl;
	}
      for (int imut=0; imut<_nb_mutclass-1; imut++) 
	{
	  _switch_rates_inc[ic][imut] = params._switch_rates_inc[imut];
	  _switch_rates_dec[ic][imut] = params._switch_rates_dec[imut];
	}
      _lambda[ic] = params._lambda;
      _perror[ic] = params._p_error;
      _g[ic] = params._g;
    }

	  
  string word;
  ifstream input (datafilename.c_str(), ios::in);
  
  input >> word;
  input >> _nbsamples;
  _crypt_sample = new Crypt* [_nbsamples];
  _mut_model = new Mutmodel* [_nclusters];
  if (!_unique_mutmodel) 
    {
      for (int ic=0; ic<_nclusters; ic++) 
	{
	  if (!_contextdep_mutmodel)
	    {
	      if (_sym_mutrate)
		{
		  _mut_model[ic] = new Mutmodel(_nb_mutclass, params._mu_m_vect, params._switch_rates_inc, 
						params._switch_rates_dec, params._p_error);
		} else {
		  _mut_model[ic] = new Mutmodel(_nb_mutclass, params._mu_m_vect, params._mu_u_vect, params._switch_rates_inc, 
						params._switch_rates_dec, params._p_error,_rng);
		}
	    } else {
	      _mut_model[ic] = new Mutmodel(params._lseq, params._nb_meth_range, 
					    params._mu_m_vect, params._mu_u_vect, params._p_error,_rng);
	    }
	  cout << "set alpha_diff to " << params._alpha/params._lambda << endl; 
	  _mut_model[ic]->set_alpha_diff(params._alpha/params._lambda);
	}
    } else {
      if (!_contextdep_mutmodel)
	{
	  if (_sym_mutrate)
	    {
	      _mut_model[0] = new Mutmodel(_nb_mutclass, params._mu_m_vect, params._switch_rates_inc, 
					    params._switch_rates_dec, params._p_error);
	    } else {
	      _mut_model[0] = new Mutmodel(_nb_mutclass, params._mu_m_vect, params._mu_u_vect, params._switch_rates_inc, 
					    params._switch_rates_dec, params._p_error,_rng);
	    }
	} else {
	  _mut_model[0] = new Mutmodel(params._lseq, params._nb_meth_range, params._mu_m_vect, params._mu_u_vect, params._p_error,_rng);
	}
      _mut_model[0]->set_alpha_diff(params._alpha/params._lambda);
      for (int ic=1; ic<_nclusters; ic++) 
	{
	  _mut_model[ic] = _mut_model[0];
	  _mut_model[ic]->set_alpha_diff(params._alpha/params._lambda);
	}
    }
  _clust_ids = new int [_nbsamples];
//   for (int is=0; is<_nbsamples; is++) 
//     {
//       _clust_ids[is] = gsl_rng_uniform_int(_rng, _nclusters);
//     }
  this->initClustAlloc();
  this->initPatRateFact();

  cout << "-------------------" << endl;
  cout << "ncrypts " << _nbsamples << endl;
  if (params._simul) 
    {
      cerr << "simul option is not currently supported" << endl;
      exit(1);
    } else {
      for (int is=0; is<_nbsamples; is++) 
	{
	  _crypt_sample[is] = new Crypt(_mut_model[_clust_ids[is]], params._N, 
					params._lambda, params._g, _rng, _stirling2, 
					input); 
	  //_crypt_sample[is]->printCrypt();
	  //_crypt_sample[is]->printInfo();
	}
    }
  cout << "-------------------" << endl;
  input.close();
}


mcmc::~mcmc()
{
  
}

void mcmc::initClustAlloc()
{
  if (_nclusters==1) {
    for (int is=0; is <_nbsamples; is++) {
      _clust_ids[is] = 0;
    }
  } else if (_nclusters==2 && _nbsamples==57) {
    for (int is=0; is<=56; is++) {
      _clust_ids[is] = 0; // all 
    }
    for (int is=37; is<=43; is++) {
      _clust_ids[is] = 1; // but pat X
    }
  } else if (_nclusters==2 && _nbsamples==2) {
    _clust_ids[0] = 0;
    _clust_ids[1] = 1;
  }  else {
    cerr << "pb in void mcmc::initClustAlloc()"<< endl;
    exit(1);
  }
  
}

void mcmc::initPatRateFact()
{
  bool several_rate = true;
  _pat_rate_hyperp = 1;
  _crypt2pat = new int [_nbsamples];
  
  if (0 && several_rate && _nbsamples==57) {
    _nb_patients = 7;
    for (int is=0; is<=6; is++) {
      _crypt2pat[is] = 0; // pat I
    }
    for (int is=7; is<=13; is++) {
      _crypt2pat[is] = 1; // pat D
    }
    for (int is=14; is<=22; is++) {
      _crypt2pat[is] = 2; // pat F
    }
    for (int is=23; is<=28; is++) {
      _crypt2pat[is] = 3; // pat H
    }
    for (int is=29; is<=36; is++) {
      _crypt2pat[is] = 4; // pat M
    }
    for (int is=37; is<=43; is++) {
      _crypt2pat[is] = 5; // pat X
    }
    for (int is=44; is<=50; is++) {
      _crypt2pat[is] = 6; // pat Y
    }
    for (int is=51; is<=56; is++) {
      _crypt2pat[is] = 4; // pat M
    }
  } else if (several_rate && _nbsamples==57) {
    _nb_patients = 2;
    for (int is=0; is<=56; is++) {
      _crypt2pat[is] = 0; // all 
    }
    for (int is=37; is<=43; is++) {
      _crypt2pat[is] = 1; // but pat X
    }
    
  } else if (several_rate && _nbsamples==50) {
    _nb_patients = 7;
    for (int is=0; is<=6; is++) {
      _crypt2pat[is] = 0; // pat I
    }
    for (int is=7; is<=13; is++) {
      _crypt2pat[is] = 1; // pat D
    }
    for (int is=14; is<=22; is++) {
      _crypt2pat[is] = 2; // pat F
    }
    for (int is=23; is<=28; is++) {
      _crypt2pat[is] = 3; // pat H
    }
    for (int is=29; is<=36; is++) {
      _crypt2pat[is] = 4; // pat M
    }
    for (int is=37; is<=44; is++) {
      _crypt2pat[is] = 5; // pat Y
    }
    for (int is=45; is<=50; is++) {
      _crypt2pat[is] = 4; // pat M
    }
  } else if (several_rate && _nbsamples==2) {
    _nb_patients = 2;
    _crypt2pat[0] = 0; 
    _crypt2pat[1] = 1; 
  } else {
    _nb_patients = 1;
    for (int is=0; is<_nbsamples; is++) {
      _crypt2pat[is] = 0;
    }
  }

  _pat_rate_fact = new double [_nb_patients];
  for (int ip=0; ip<_nb_patients; ip++) {
    _pat_rate_fact[ip] = 1;
  }
}

void mcmc::simul(int nsimul, string outstatfile, string outseqfile, bool printseq)
{
  vector<StatWithinCrypt*> withinstat;

  for (int is=0; is<_nbsamples; is++) 
    {
      StatWithinCrypt* stat = _crypt_sample[is]->summaryStat(false);
      stat->_isimul = -1;
      withinstat.push_back(stat);
    }

  if (_nclusters>1) 
    {
      cerr << "void mcmc::simul(int nsimul, string outstatfile, string outseqfile) does not support nclusters>1" << endl;
    }
  
  if (printseq) 
    {
      ofstream outseq(outseqfile.c_str(),ios::out);
      for (int i=0; i<nsimul; i++)
	{
	  if (i%10==0) {
	    cout << "isimul " << i << endl;
	  }
	  for (int is=0; is<_nbsamples; is++) 
	    {
	      _mut_model[_clust_ids[is]]->set_pat_rate_factor(_pat_rate_fact[_crypt2pat[is]]);
	      _mut_model[_clust_ids[is]]->changeperror(_perror[_clust_ids[is]]);
	      _crypt_sample[is]->deleteGenealogy();
	      _crypt_sample[is]->simulPrior();
	      StatWithinCrypt* stat = _crypt_sample[is]->summaryStat(false);
	      stat->_isimul = i;
	      withinstat.push_back(stat);	  
	      //	  _crypt_sample[is]->printTree4R(otree, string("crypt")+string(tmpis)+string("sim")+string(tmpi));
	    }      
	  outseq << "# Simulated data " << i << endl;
	  outseq << "ncrypts " << _nbsamples << endl;
	  for (int is=0; is<_nbsamples; is++) 
	    {
	      _crypt_sample[is]->printCryptFromObsSeq(outseq);
	    }
	}
      outseq.close();
    } else {
      for (int i=0; i<nsimul; i++)
	{
	  if (i%10==0) {
	    cout << "isimul " << i << endl;
	  }
	  for (int is=0; is<_nbsamples; is++) 
	    {
	      _mut_model[_clust_ids[is]]->set_pat_rate_factor(_pat_rate_fact[_crypt2pat[is]]);
	      _mut_model[_clust_ids[is]]->changeperror(_perror[_clust_ids[is]]);
	      _crypt_sample[is]->deleteGenealogy();
	      _crypt_sample[is]->simulPrior();
	      StatWithinCrypt* stat = _crypt_sample[is]->summaryStat(false);
	      stat->_isimul = i;
	      withinstat.push_back(stat);	  
	    }      
	}      
    }

  ofstream outstat (outstatfile.c_str(), ios::out);
  withinstat[0]->printHeader(outstat);
  for (unsigned int i=0; i<withinstat.size() ; i++)
    {
      withinstat[i]->print(outstat);
    }
  outstat.close();
}


void mcmc::predictiveSimulFromFile(string paramfile, string allocfile, 
				   string patratesfile, string outstatfile)
{
  vector<StatWithinCrypt*> withinstat;

  for (int is=0; is<_nbsamples; is++) 
    {
      StatWithinCrypt* stat = _crypt_sample[is]->summaryStat(false);
      stat->_isimul = -1;
      withinstat.push_back(stat);
    }

  ifstream inparams(paramfile.c_str(), ios::in);
  ifstream inalloc(allocfile.c_str(), ios::in);
  ifstream inpatrates(patratesfile.c_str(), ios::in);

  string treefile = outstatfile + string(".trees");
  ofstream otree(treefile.c_str(), ios::out);
  char* tmpi = new char [100];
  char* tmpis = new char [100];


  string word;
  int nsimul;
  inparams >> word;
  inparams >> nsimul;
  for (int i=0; i<nsimul; i++)
    {
      for (int ic=0 ; ic<_nclusters ; ic++)
	{
	  inparams >> word;
	  inparams >> _pclusters[ic];
	  if (0 && !_contextdep_mutmodel)
	    {
	      for (int imut=0; imut<_nb_mutclass; imut++)
		{
		  inparams >> _mutrate_m[ic][imut];
		  inparams >> _mutrate_u[ic][imut];
		  _mut_model[ic]->changemutrate_m(imut, _mutrate_m[ic][imut]);
		  _mut_model[ic]->changemutrate_u(imut, _mutrate_u[ic][imut]);
		}
	      for (int imut=0; imut<_nb_mutclass-1; imut++)
		{
		  inparams >>  _switch_rates_inc[ic][imut];
		  _mut_model[ic]->changeswitchrate(_switch_rates_inc[ic][imut], imut, imut+1);
		}
	      for (int imut=0; imut<_nb_mutclass-1; imut++)
		{
		  inparams >>  _switch_rates_dec[ic][imut];
		  _mut_model[ic]->changeswitchrate(_switch_rates_dec[ic][imut], imut+1, imut);
		}
	    } else {
	      cout << "reading" << endl;
	      _mut_model[ic]->readContextDepParams(inparams);
	    }
	  inparams >> _lambda[ic];
	  inparams >> _g[ic];
	  double alpha;
	  inparams >> alpha;
	  _mut_model[ic]->set_alpha_diff(alpha/_lambda[ic]);
	  inparams >> _perror[ic];
	  _mut_model[ic]->changeperror(_perror[ic]);	  
	}
      for (int is=0; is<_nbsamples; is++) 
	{
	  int cluster;
	  int N;
	  double patrate;
	  inalloc >> cluster;
	  inalloc >> N;
	  inpatrates >> patrate;
	  cout << "i " << i << " is " << is << " c " << cluster << " N " << N << " _mutrate_m " << _mutrate_m[cluster][0] <<" _mutrate_u " << _mutrate_u[cluster][0] << " _perror " << _perror[cluster] << " _lambda " << _lambda[cluster] << " patrate " << patrate << endl;
	  _mut_model[cluster]->set_pat_rate_factor(patrate);
	  _mut_model[cluster]->changeperror(_perror[cluster]);	  
	  _crypt_sample[is]->changeMutmodel(_mut_model[cluster]);
	  _crypt_sample[is]->setlambda(_lambda[cluster]);
	  _crypt_sample[is]->setg(_g[cluster]);
	  _crypt_sample[is]->setNstemcell(N);
	  _crypt_sample[is]->deleteGenealogy();
	  _crypt_sample[is]->simulPrior();
	  StatWithinCrypt* stat = _crypt_sample[is]->summaryStat(false);
	  stat->_isimul = i;
	  withinstat.push_back(stat);	  
	  sprintf(tmpi,"%d",i);
	  sprintf(tmpis,"%d",is);
	  _crypt_sample[is]->printTree4R(otree, string("crypt")+string(tmpis)+string("sim")+string(tmpi));
	}
      cout << endl;
      if (i==0)
	{
	  cout << "----------From Obs For sim = 0---------" << endl;
	  cout << "ncrypts " << _nbsamples << endl;
	  for (int is=0; is<_nbsamples; is++) 
	    {
	      _crypt_sample[is]->printCryptFromObsSeq();
	    }
	  cout << "-------------------" << endl;
	
	  cout << "----------From Ori For sim = 0---------" << endl;
	  cout << "ncrypts " << _nbsamples << endl;
	  for (int is=0; is<_nbsamples; is++) 
	    {
	      _crypt_sample[is]->printCrypt();
	    }
	  cout << "-------------------" << endl;
	}
    }
  inparams.close();
  inalloc.close();
  inpatrates.close();
  delete tmpi;
  delete tmpis;
  otree.close();

  ofstream out (outstatfile.c_str(), ios::out);
  withinstat[0]->printHeader(out);
  for (unsigned int i=0; i<withinstat.size() ; i++)
    {
      withinstat[i]->print(out);
    }
  out.close();
}

void mcmc::stirlingAlloc()
{
  _nstirling2 = 30;

  _stirling2 = new double* [_nstirling2];
  for (int i=0; i<_nstirling2; i++)
    {
      _stirling2[i] = new double [_nstirling2];
      if (i==0) 
	{
	  for (int j=0 ; j<_nstirling2; j++) 
	    {
	      _stirling2[i][j] = 0;
	    }
	} else {
	  _stirling2[i][0] = 0;
	  _stirling2[i][1] = 1;
	  for (int j=2 ; j<_nstirling2; j++) 
	    {
	      _stirling2[i][j] = _stirling2[i-1][j-1] + j*_stirling2[i-1][j];
	    }
	}      
    }

//   cout << "Stirling number of second kind" << endl;
//   for (int i=0; i<_nstirling2; i++)
//     {
//       for (int j=0 ; j<_nstirling2; j++) 
// 	{
// 	  cout << _stirling2[i][j] << "\t";
// 	}      
//       cout << endl;
//     }

   int N = 40;
   int ncells = 24;
   cout << "pdf nb stem cell sampled" << endl;
   double sum = 0;
   int nmax = N;
   if (ncells<nmax) 
     {
       nmax = ncells;
     }
   for (int i=1; i<=nmax; i++)
     {
       double p = 1/pow((double)N,ncells);
       for (int j=1; j<=i; j++)
 	{
 	  p *= (N-j+1);
 	}
       p *= _stirling2[ncells][i];
       sum += p;
       cout << " nstemsampled " << i << " p=" << p << endl;
     }
   cout << "sum " << sum << endl;

}

void  mcmc::setPrior(Params& params)
{
  _mutratemin = params._mutratemin;
  _mutratemax = params._mutratemax;
  _lambdamin = params._lambdamin;
  _lambdamax = params._lambdamax;
  _gmin = params._gmin;
  _gmax = params._gmax;
  _nstemcells_prior_min = params._nstemcells_prior_min;
  _nstemcells_prior_max = params._nstemcells_prior_max;
  _nstemcells_diff_max = params._nstemcells_diff_max;

  _nb_stemcells_dirichletprior = new double [_nstemcells_prior_max+1];
  _nb_stemcells_pseudocounts = new double [_nstemcells_prior_max+1];
  _nb_stemcells_distrib = new double [_nstemcells_prior_max+1];
  _nb_stemcells_counts = new int [_nstemcells_prior_max+1];

  cout << "_nb_stemcells_pseudocounts " << _nb_stemcells_pseudocounts << endl;
  cout << "_nstemcells_prior_max " << _nstemcells_prior_max << endl;

  for (int i=0 ; i<= _nstemcells_prior_max ; i++) 
    {
      _nb_stemcells_dirichletprior[i] = 1;
      _nb_stemcells_counts[i] = 0;
    }
}

void mcmc::samplingFancy(Params& params, string outfilename)//int niter, string outfilename, int update_coalrate, bool update_perror, bool inverse_prior, bool update_N)//, char* mutmodelparamsfile)
{
  cout << "entering samplingFancy outfilename " << outfilename << endl;
  string predfilename = outfilename + string(".pred");
  string treesfilename = outfilename + string(".trees");
  string allocfilename = outfilename + string(".alloc");
  string patratefilename = outfilename + string(".patrates");

  ofstream out (outfilename.c_str(), ios::out);
  ofstream outpred (predfilename.c_str(), ios::out);
  ofstream outtrees (treesfilename.c_str(), ios::out);
  ofstream outalloc (allocfilename.c_str(), ios::out);
  ofstream outpatrate (patratefilename.c_str(), ios::out);

  // cout << "gmin " << _gmin << " gmax " << _gmax << endl;

  this->initNcounts();
  _nb_accept_lambdaup = 0;
  for (int is=0; is<_nbsamples; is++) 
    {
      //   _crypt_sample[is]->printInfo();
      _crypt_sample[is]->resetErrorCounts();
    }
  
  int nb_Ncounts = 1;
  if (_nclusters>1 && !_unique_Nlambda) {
    nb_Ncounts = _nclusters;
  }
  int* Nmin_byclust = new int [nb_Ncounts];
  int* Nmax_byclust = new int [nb_Ncounts];
  int** Ncounts_byclust = new int* [nb_Ncounts];
  for (int ic=0 ; ic<nb_Ncounts ; ic++)
    {
      Nmin_byclust[ic] = _nstemcells_prior_max;
      Nmax_byclust[ic] = _nstemcells_prior_min;
      Ncounts_byclust[ic] = new int [_nstemcells_prior_max+1];
      for (int j=0; j<=_nstemcells_prior_max ; j++)
	{
	  Ncounts_byclust[ic][j] = 0;
	}
    }
  for (int is=0; is<_nbsamples; is++)
    {
      int ic = 0;
      if (_nclusters>1 && !_unique_Nlambda) {
	ic = _clust_ids[is];
      }      
      int n = _crypt_sample[is]->Nstemcell();
      Ncounts_byclust[ic][n]++;
      if (Nmin_byclust[ic] > n) 
	{
	  Nmin_byclust[ic] = n;
	}
      if (Nmax_byclust[ic] < n) 
	{
	  Nmax_byclust[ic] = n;
	}
    }

//   if (_contextdep_mutmodel) 
//     {
//       for (int ic=0 ; ic<_nclusters ; ic++)
// 	{
// 	  ifstream inparams(mutmodelparamsfile, ios::in);  
// 	  _mut_model[ic]->readContextDepParams(inparams);
// 	  inparams.close();
// 	}
//     }

  StatWithinCrypt statheader;
  statheader.printHeader(outpred);
  int nbrecorded = 0;
  for (int is=0; is<_nbsamples; is++) 
    {
      _crypt_sample[is]->firstnode()->checkContextDep();
    }
//   for (int is=0; is<_nbsamples; is++) 
//     {
//       _mut_model[_clust_ids[is]]->set_pat_rate_factor(_pat_rate_fact[_crypt2pat[is]]);
//       _mut_model[_clust_ids[is]]->changeperror(_perror[_clust_ids[is]]);
//       _crypt_sample[is]->checkSeqLikComputations();
//     }

//   for (int iter=0; iter<niter; iter++) 
//     {
//       for (int is=0; is<_nbsamples; is++) 
// 	{
// 	  _mut_model[_clust_ids[is]]->set_pat_rate_factor(_pat_rate_fact[_crypt2pat[is]]);
// 	  _mut_model[_clust_ids[is]]->changeperror(_perror[_clust_ids[is]]);
// 	  _crypt_sample[is]->updateSeq();
// 	  _crypt_sample[is]->updateTreeWB();
// 	  _crypt_sample[is]->updateTermStemCell(false);
// 	  if (iter%100==0) 
// 	    {
// 	      cout << "iter " << iter << " " << is << " " 
// 		   << _crypt_sample[is]->nstemcellsampled() << endl;
// 	      _crypt_sample[is]->printCoaltimes();
// 	      cout << "termstemcell " << _crypt_sample[is]->ntermstemcell() << endl;
// 	      //_crypt_sample[is]->printInfo();
// 	    }
// 	}
//     }
//   exit(1);

  bool update_N = false;
  bool update_perror = false;
  for (int iter=0; iter<params._niter; iter++) 
    {
      if (iter==params._update_N) 
	{
	  update_N = true;
	}
      if (iter==params._update_perror) 
	{
	  update_perror = true;
	}
      //cout << _mut_model[0]->alpha_diff() << endl;
      for (int is=0; is<_nbsamples; is++) 
	{
	  _mut_model[_clust_ids[is]]->set_pat_rate_factor(_pat_rate_fact[_crypt2pat[is]]);
	  _mut_model[_clust_ids[is]]->changeperror(_perror[_clust_ids[is]]);
	   _crypt_sample[is]->updateSeq();
	   int ncells = _crypt_sample[is]->ncells();
	   for (int ik=0; ik<ncells; ik++)
	     {
	       _crypt_sample[is]->updateTreeWB(update_N, 
					       Ncounts_byclust[_clust_ids[is]], 
					       Nmin_byclust[_clust_ids[is]], 
					       Nmax_byclust[_clust_ids[is]],
					       _nstemcells_prior_min, 
					       _nstemcells_prior_max);
	     }
	   _crypt_sample[is]->updateTermStemCell(false); 
	   _crypt_sample[is]->updateInitStemCell();
	   if (update_N) {	   
	     if (_nclusters>1 && _unique_Nlambda) {
	       _crypt_sample[is]->updateNstemcellFancy(Ncounts_byclust[0], 
						       Nmin_byclust[0], 
						       Nmax_byclust[0],
						       _nstemcells_prior_min, 
						       _nstemcells_prior_max);
	     } else {	     
	       _crypt_sample[is]->updateNstemcellFancy(Ncounts_byclust[_clust_ids[is]], 
						       Nmin_byclust[_clust_ids[is]], 
						       Nmax_byclust[_clust_ids[is]],
						       _nstemcells_prior_min, 
						       _nstemcells_prior_max);
	     }
	   }
	   _crypt_sample[is]->storeErrorCounts();
 	}     

      if (params._update_mutrate) {
	this->updateMutrateNew();
      }
      if (params._update_alpha) 
	{
	  this->updateAlpha(params._alphamin, params._alphamax);
	}
      if (params._update_g) 
	{
	  this->updateg();
	}
      if (params._update_coalrate==1 || (params._update_coalrate==2 && iter > 20000))// || iter > 20000) 
	{
	  this->updateLambda(params._inverse_prior);
	} else {
	  //	  cout << "no updateLambda()" << endl;
	}
      if (update_perror)
	{
	  this->updatePerror();
	}
       //cout << _mutrate_m[0][0] << endl;
      this->updateSwitchRate();
      if (update_N) {	   
	if (_nclusters>1 && _unique_Nlambda) {
	  this->updateAllN(Nmin_byclust[0], 
			   Nmax_byclust[0], Ncounts_byclust[0], -1);
	} else {	     
	  for (int ic=0; ic<_nclusters; ic++)
	    {
	      this->updateAllN(Nmin_byclust[ic], Nmax_byclust[ic], 
			       Ncounts_byclust[ic],ic);
	    }
	}
      }
      
      if (!_no_clusterupdate) 
	{
	  this->updatePcluster();
	  //cout << "1 _mutrate_m[1][0] " << _mutrate_m[1][0] << endl;
	  this->updateClusterAlloc(Nmin_byclust, Nmax_byclust, Ncounts_byclust);
	}
      if (iter%100==0)
	{
	  //cout << "_nb_accept_lambdaup " << _nb_accept_lambdaup << endl;
	  //out << iter << " ";
	  //out << _crypt_sample[0]->nstemcellsampled() << " " << _crypt_sample[0]->ntermstemcell() << " ";
	  this->printStateFancySampling(Nmin_byclust, Nmax_byclust, Ncounts_byclust,out, outalloc);
	  //this->printStateFancySampling(out);
	  //cout << "current _pat_rate_fact ";
	  for (int ip=0; ip<_nb_patients; ip++) 
	    {
	      //  cout << _pat_rate_fact[ip] << " ";
	    }
	  //	  cout << endl;
	  // cout << "_pat_rate_hyperp " << _pat_rate_hyperp << endl;
	  for (int is=0; is<_nbsamples; is++) 
	    {
	      outpatrate << _pat_rate_fact[_crypt2pat[is]] << " ";
	    }
	  outpatrate << endl;
	  
	  double loglikseq = 0;
	  double logliktree = 0;
	  for (int is=0; is<_nbsamples; is++) 
	    {      
	      _mut_model[_clust_ids[is]]->set_pat_rate_factor(_pat_rate_fact[_crypt2pat[is]]);
	      _mut_model[_clust_ids[is]]->changeperror(_perror[_clust_ids[is]]);
	      double ls = _crypt_sample[is]->computeLogLikSeq();
	      loglikseq += ls;
	      double lt = _crypt_sample[is]->computeLogLikTreeN();
	      logliktree += lt;
 	      //double lt1 = _crypt_sample[is]->computeLogLikTree();
 	      //double lt2 = _crypt_sample[is]->computeLogLikN();
	      // cout << "(" << is <<","<< ls <<"," << lt << "," << lt1 << "," << lt2 << ")";
	    }
	  //cout << endl;
	  //cout << loglikseq << endl;
	  if (nbrecorded%10==0)
	    {
	      //cout << "************ iter " << iter << endl;
	      for (int is=0; is<_nbsamples; is++) 
		{
		  //_crypt_sample[is]->printInfo();
		  //StatWithinCrypt* stat = _crypt_sample[is]->summaryStat(true);
		  //stat->_isimul = nbrecorded;
		  //stat->print(outpred);
		}
	      //cout << "******************" << endl;
	    }
	  if (nbrecorded%1000==0)
	    {
	      char* tmpi = new char [100];
	      char* tmpj = new char [100];
	      for (int is=0; is<_nbsamples; is++) 
		{
		  sprintf(tmpi,"%d",is);
		  sprintf(tmpj,"%d",nbrecorded);
		  _crypt_sample[is]->printTree4R(outtrees, 
						 string("crypt")+string(tmpi)+string("iter")+string(tmpj));
		}
	      delete [] tmpi;
	      delete [] tmpj;
	    }
	  
	  nbrecorded++;
	}
    }

  for (int is=0; is<_nbsamples; is++) 
    {
      _crypt_sample[is]->printErrorCounts();
    }
  out.close();
  outpred.close();
  outtrees.close();
  outalloc.close();
  outpatrate.close();
}

void mcmc::testPeeling(int niter, string outfilename)
{
  cout << "entering testpeeling " << endl;
  string allocfilename = outfilename + string(".alloc");
  ofstream outalloc (allocfilename.c_str(), ios::out);
  ofstream out (outfilename.c_str(), ios::out);

  this->initNcounts();
  for (int is=0; is<_nbsamples; is++) 
    {
      //   _crypt_sample[is]->printInfo();
      _crypt_sample[is]->resetErrorCounts();
    }
  
  int* Nmin_byclust = new int [_nclusters];
  int* Nmax_byclust = new int [_nclusters];
  int** Ncounts_byclust = new int* [_nclusters];
  for (int ic=0 ; ic<_nclusters ; ic++)
    {
      Nmin_byclust[ic] = _nstemcells_prior_max;
      Nmax_byclust[ic] = _nstemcells_prior_min;
      Ncounts_byclust[ic] = new int [_nstemcells_prior_max+1];
      for (int j=0; j<=_nstemcells_prior_max ; j++)
	{
	  Ncounts_byclust[ic][j] = 0;
	}
    }
  for (int is=0; is<_nbsamples; is++)
    {
      int ic = _clust_ids[is];
      int n = _crypt_sample[is]->Nstemcell();
      Ncounts_byclust[ic][n]++;
      if (Nmin_byclust[ic] > n) 
	{
	  Nmin_byclust[ic] = n;
	}
      if (Nmax_byclust[ic] < n) 
	{
	  Nmax_byclust[ic] = n;
	}
    }

  for (int is=0; is<_nbsamples; is++) 
    {
      _crypt_sample[is]->firstnode()->resetCountSeq();
      _crypt_sample[is]->firstnode()->checkContextDep();
    }
  int nbrecorded = 0;
  for (int iter=0; iter<niter; iter++) 
    {
      for (int is=0; is<_nbsamples; is++) 
	{
	  _mut_model[_clust_ids[is]]->set_pat_rate_factor(_pat_rate_fact[_crypt2pat[is]]);
	  _crypt_sample[is]->updateSeq();
	  //_crypt_sample[is]->updateSeqForwBackw();
	  _crypt_sample[is]->firstnode()->countSeq();
 	}
      //this->updateMutrate();
      if (iter%100==0)
	{
	  this->printStateFancySampling(Nmin_byclust, Nmax_byclust, Ncounts_byclust,out, outalloc);
	  if (nbrecorded%1000==0)
	    {
	      cout << "------- ITER " << iter << endl; 
	      for (int is=0; is<_nbsamples; is++) 
		{
		  _crypt_sample[is]->firstnode()->printCountSeq();
		}
	    }	  
	  nbrecorded++;
	}
    }

  for (int is=0; is<_nbsamples; is++) 
    {
      _crypt_sample[is]->printErrorCounts();
    }
  out.close();
  outalloc.close();
}


bool mcmc::printStateFancySampling(int* Nmin_byclust, int* Nmax_byclust, 
				   int** Ncounts_byclust, ofstream& out, ofstream& out_alloc)
{
  //  int unique_N = -1;

//   for (int i=_nstemcells_prior_min; i<=_nstemcells_prior_max; i++)
//     {
//       if (_nb_stemcells_counts[i] == _nbsamples) 
// 	{
// 	  unique_N = i;
// 	}
//     }

//   if (unique_N<0)
//     {
//       return false;
//     }
  
//   out << unique_N;
  
  for (int ic=0; ic<_nclusters; ic++)
    {
      double N;
      if (!_unique_Nlambda) {
	N = (double)(Nmin_byclust[ic]*Ncounts_byclust[ic][Nmin_byclust[ic]] + Nmax_byclust[ic]*Ncounts_byclust[ic][Nmax_byclust[ic]])/(double)(Ncounts_byclust[ic][Nmin_byclust[ic]]+Ncounts_byclust[ic][Nmax_byclust[ic]]);
      } else {
	N = (double)(Nmin_byclust[0]*Ncounts_byclust[0][Nmin_byclust[0]] + Nmax_byclust[0]*Ncounts_byclust[0][Nmax_byclust[0]])/(double)(Ncounts_byclust[0][Nmin_byclust[0]]+Ncounts_byclust[0][Nmax_byclust[0]]);	
      }

      out << N << " " << _pclusters[ic] << " ";
      if (0 && !_contextdep_mutmodel)
	{
	  for (int imut=0; imut<_nb_mutclass; imut++)
	    {
	      out << _mutrate_m[ic][imut] << " " << _mutrate_u[ic][imut] << " ";
	    }
	  for (int imut=0; imut<_nb_mutclass-1; imut++)
	    {
	      out << _switch_rates_inc[ic][imut] << " ";
	    }
	  for (int imut=0; imut<_nb_mutclass-1; imut++)
	    {
	      out << _switch_rates_dec[ic][imut] << " ";
	    }
	} else {
	  _mut_model[ic]->printContextDepParams(out);	  
	}
      out << _lambda[ic] << " " << _g[ic] << " " << _mut_model[ic]->alpha_diff()*_lambda[ic] << " " << _perror[ic] << " ";
    }
  out << endl;
    
  for (int is=0; is<_nbsamples; is++)
    {
      out_alloc << _clust_ids[is] << " " << _crypt_sample[is]->Nstemcell() << " ";
    }
  out_alloc << endl;

  return true;
}

void mcmc::updateSupportN()
{
  cout << "entering void mcmc::updateSupportN()" << endl;
  double log_lik_den;
  double log_lik_num;
  
  bool increase;
  int rint;
  int rint2;
  double sum1;
  double sum2;
  double logratio;
  double r;
  
  rint = gsl_ran_bernoulli(_rng,0.5);
  if (rint == 0)
    {
      // on the right
      log_lik_den = 0;
      log_lik_num = 0;
      
      increase = false;
      rint2 = gsl_ran_bernoulli(_rng,0.5);
      if (rint2==0)
	{
	  increase = true;
	  cout << "Right support increase attempt" << endl;
	  cout << _nstemcells_max << " -> " << _nstemcells_max+1 << endl;
	} else {
	  cout << "Right support decrease attempt" << endl;
	  cout << _nstemcells_max << " -> " << _nstemcells_max-1 << endl;  
	  if (_nstemcells_min==1 && _nstemcells_max==2)
	    {
	      cout << "crypt Nstem "; 
	      for (int is=0; is<_nbsamples; is++) 
		{
		  cout << _crypt_sample[is]->Nstemcell() << " ";
		}	      
	      cout << "bingo" << endl;
	    }
	  for (int is=0; is<_nbsamples; is++) 
	    {
	      if (_crypt_sample[is]->Nstemcell() >= _nstemcells_max)
		{
		  return;
		}
	    }
	}
      
      if (increase==true && _nstemcells_max==_nstemcells_prior_max)
	{
	  return;
	}
      if (increase==false && _nstemcells_min==_nstemcells_max)
	{
	  return;
	}
      if (increase==true && _nstemcells_max-_nstemcells_min+1>=_nstemcells_diff_max)
	{
	  return;
	}
      
      sum1 = 0;
      sum2 = 0;
      for (int in=_nstemcells_min ; in<=_nstemcells_max; in++)
	{
	  cout << in << endl;
	  sum1 += _nb_stemcells_dirichletprior[in];
	  sum2 += _nb_stemcells_pseudocounts[in];
	  cout << sum1 << " " << sum2 << " " 
	       << _nb_stemcells_dirichletprior[in] << " " << _nb_stemcells_pseudocounts[in] << endl;
	  log_lik_den += gsl_sf_lngamma(_nb_stemcells_pseudocounts[in]);
	  log_lik_den -= gsl_sf_lngamma(_nb_stemcells_dirichletprior[in]);
	}
      log_lik_den += gsl_sf_lngamma(sum1);
      log_lik_den -= gsl_sf_lngamma(sum2);
      
      if (increase==true)
	{
	  _nstemcells_max++;
	} else {
	  _nstemcells_max--;
	}
      
      sum1 = 0;
      sum2 = 0;
      for (int in=_nstemcells_min ; in<=_nstemcells_max; in++)
	{
	  sum1 += _nb_stemcells_dirichletprior[in];
	  sum2 += _nb_stemcells_pseudocounts[in];
	  log_lik_num += gsl_sf_lngamma(_nb_stemcells_pseudocounts[in]);
	  log_lik_num -= gsl_sf_lngamma(_nb_stemcells_dirichletprior[in]);
	}
      log_lik_num += gsl_sf_lngamma(sum1);
      log_lik_num -= gsl_sf_lngamma(sum2);

      cout << "log_lik_num " << log_lik_num << endl;
      cout << "log_lik_den " << log_lik_den << endl;
      
      logratio = log_lik_num-log_lik_den;
      
      cout << "ratio " << exp(logratio) << endl;
      
      r = log(gsl_ran_flat(_rng, 0, 1));
      if (r < logratio)
	{
	  cout << "updateSupportN right Accepted" << endl;
	  this->updateNprior();
	  if (increase == true) 
	    {
	      cout << "Increasing" << endl;
	    } else {
	      cout << "Decreasing" << endl;
	    }
	} else {
	  cout << "updateSupportN right Rejected" << endl;
	  if (increase == true) 
	    {
	      _nstemcells_max--;
	    } else {
	      _nstemcells_max++;
	    }      
	}
    } else {
      // on the left
      log_lik_den = 0;
      log_lik_num = 0;
      
      increase = false;
      rint2 = gsl_ran_bernoulli(_rng,0.5);
      if (rint2==0)
	{
	  increase = true;
	  cout << "Left support increase attempt" << endl;
	  cout << _nstemcells_min << " -> " << _nstemcells_min-1 << endl;
	} else {
	  cout << "Left support decrease attempt" << endl;
	  cout << _nstemcells_min << " -> " << _nstemcells_min+1 << endl;  
	  for (int is=0; is<_nbsamples; is++) 
	    {
	      if (_crypt_sample[is]->Nstemcell() <= _nstemcells_min)
		{
		  return;
		}
	    }
	}
      
      if (increase==true && _nstemcells_min==_nstemcells_prior_min)
	{
	  return;
	}
      if (increase==true && _nstemcells_max-_nstemcells_min+1>=_nstemcells_diff_max)
	{
	  return;
	}
      if (increase==false && _nstemcells_min==_nstemcells_max)
	{
	  return;
	}
      
      sum1 = 0;
      sum2 = 0;
      for (int in=_nstemcells_min ; in<=_nstemcells_max; in++)
	{
	  cout << in << endl;
	  sum1 += _nb_stemcells_dirichletprior[in];
	  sum2 += _nb_stemcells_pseudocounts[in];
	  cout << sum1 << " " << sum2 << " " 
	       << _nb_stemcells_dirichletprior[in] << " " << _nb_stemcells_pseudocounts[in] << endl;
	  log_lik_den += gsl_sf_lngamma(_nb_stemcells_pseudocounts[in]);
	  log_lik_den -= gsl_sf_lngamma(_nb_stemcells_dirichletprior[in]);
	}
      log_lik_den += gsl_sf_lngamma(sum1);
      log_lik_den -= gsl_sf_lngamma(sum2);
      
      if (increase==true)
	{
	  _nstemcells_min--;
	} else {
	  _nstemcells_min++;
	}
      
      sum1 = 0;
      sum2 = 0;
      for (int in=_nstemcells_min ; in<=_nstemcells_max; in++)
	{
	  sum1 += _nb_stemcells_dirichletprior[in];
	  sum2 += _nb_stemcells_pseudocounts[in];
	  log_lik_num += gsl_sf_lngamma(_nb_stemcells_pseudocounts[in]);
	  log_lik_num -= gsl_sf_lngamma(_nb_stemcells_dirichletprior[in]);
	}
      log_lik_num += gsl_sf_lngamma(sum1);
      log_lik_num -= gsl_sf_lngamma(sum2);
      
      cout << "log_lik_num " << log_lik_num << endl;
      cout << "log_lik_den " << log_lik_den << endl;
      
      logratio = log_lik_num-log_lik_den;
      
      cout << "ratio " << exp(logratio) << endl;
      
      r = log(gsl_ran_flat(_rng, 0, 1));
      if (r < logratio)
	{
	  cout << "updateSupportN left Accepted" << endl;
	  this->updateNprior();
	  if (increase == true) 
	    {
	      cout << "Increasing" << endl;
	    } else {
	      cout << "Decreasing" << endl;
	    }
	} else {
	  cout << "updateSupportN left Rejected" << endl;
	  if (increase == true) 
	    {
	      _nstemcells_min++;
	    } else {
	      _nstemcells_min--;
	    }      
	}
    }


  cout << "sample SupportN " << _nstemcells_min << " " << _nstemcells_max << endl;
}

void mcmc::initNcounts()
{
  for (int i=0 ; i<= _nstemcells_prior_max ; i++) 
    {
      _nb_stemcells_counts[i] = 0;
    }
  for (int is=0; is<_nbsamples; is++) 
    {
      _nb_stemcells_counts[_crypt_sample[is]->Nstemcell()]++;
    }
 
}

void mcmc::updateNprior()
{
  cout << "void mcmc::updateNprior()" << endl;
  for (int i=0 ; i<= _nstemcells_prior_max ; i++) 
    {
      _nb_stemcells_pseudocounts[i] = _nb_stemcells_dirichletprior[i];
    }
  for (int is=0; is<_nbsamples; is++) 
    {
      _nb_stemcells_pseudocounts[_crypt_sample[is]->Nstemcell()]++;
    }
  gsl_ran_dirichlet(_rng, _nstemcells_max-_nstemcells_min+1, 
  		    _nb_stemcells_pseudocounts+_nstemcells_min, 
  		    _nb_stemcells_distrib+_nstemcells_min);
  
  cout << "sample nb_stemcells_counts ";
  for (int i=_nstemcells_min ; i<=_nstemcells_max ; i++)
    {
      cout << _nb_stemcells_pseudocounts[i]-1 << " ";
    }
  cout << endl;
  cout << "sample nb_stemcells_distrib ";
  for (int i=_nstemcells_prior_min ; i<_nstemcells_min ; i++)
    {
      cout << 0 << " ";
    }
  for (int i=_nstemcells_min ; i<=_nstemcells_max ; i++)
    {
      cout << _nb_stemcells_distrib[i] << " ";
    }
  for (int i=_nstemcells_max ; i<=_nstemcells_prior_max ; i++)
    {
      cout << 0 << " ";
    }
  cout << endl;  
}

void mcmc::updateMutrate()
{
  if (_contextdep_mutmodel)
    {
      return;
    }
  int rmut_class;
  if (_nb_mutclass>1)
    {
      rmut_class = gsl_rng_uniform_int(_rng, _nb_mutclass);
    } else {
      rmut_class = 0;
    }
  if (_sym_mutrate)
    {
      for (int ic=0; ic<_nclusters; ic++)
	{
	  double loglikseq1 = 0;
	  for (int is=0; is<_nbsamples; is++) 
	    {      
	      if (_clust_ids[is] == ic)
		{
		  _mut_model[_clust_ids[is]]->set_pat_rate_factor(_pat_rate_fact[_crypt2pat[is]]);
		  _mut_model[_clust_ids[is]]->changeperror(_perror[_clust_ids[is]]);
		  loglikseq1 += _crypt_sample[is]->computeLogLikSeq();
		}
	    }
	  
	  double mutrate1 = _mutrate_m[ic][rmut_class];
	  //cout << "ic " << ic << " " << mutrate1 << mutrate1 <<" loglikseq " << loglikseq1 << endl;
	  double inverse1 = 1/mutrate1;
	  double inverse2 = inverse1 + gsl_ran_gaussian(_rng,inverse1/5);
	  double mutrate2 = 1/inverse2;
	  double log_proposal = log(gsl_ran_gaussian_pdf(inverse2-inverse1,inverse2/5))-log(gsl_ran_gaussian_pdf(inverse2-inverse1,inverse1/5));
	  
// 	  double mutrate1 = _mutrate_m[ic];
// 	  double mutrate2 = mutrate1 + gsl_ran_gaussian(_rng,mutrate1/5); 
// 	  double log_proposal = log(gsl_ran_gaussian_pdf(mutrate2-mutrate1,mutrate2/5))-log(gsl_ran_gaussian_pdf(mutrate2-mutrate1,mutrate1/5));

	  if (mutrate2 > _mutratemin && mutrate2 < _mutratemax)
	    {
	      _mut_model[ic]->changemutrate(rmut_class, mutrate2);
	      double loglikseq2 = 0;
	      for (int is=0; is<_nbsamples; is++) 
		{	  
		  if (_clust_ids[is] == ic)
		    {
		      //_crypt_sample[is]->mutmodel()->changemutrate(rmut_class, mutrate2);
		      _mut_model[_clust_ids[is]]->set_pat_rate_factor(_pat_rate_fact[_crypt2pat[is]]);
		      _mut_model[_clust_ids[is]]->changeperror(_perror[_clust_ids[is]]);
		      loglikseq2 += _crypt_sample[is]->computeLogLikSeq();
		    }
		}
	      //cout << "mutrate2 "<< mutrate2 <<" loglikseq " << loglikseq2 << endl;
	      double r = gsl_ran_flat(_rng, 0, 1);
	      if (log(r) < loglikseq2-loglikseq1+log_proposal)
		{
		  _mutrate_m[ic][rmut_class] = mutrate2;
		  _mutrate_u[ic][rmut_class] = mutrate2;
		  //cout << "Accept change mutrate p=" << exp(loglikseq2-loglikseq1) 
		  //   << " " << mutrate1 << " -> " << mutrate2 << endl;
		} else {
		  _mut_model[ic]->changemutrate(rmut_class, mutrate1);
		}
	      
	    }
	  //      cout << "sample mutrate " << _mut_model->mutrate() << endl;  
	}
    } else {
      int rint = gsl_ran_bernoulli(_rng,0.5);
      for (int ic=0; ic<_nclusters; ic++)
	{
	  double mutrate1m = _mutrate_m[ic][rmut_class];
	  double mutrate1u = _mutrate_u[ic][rmut_class];

	  double loglikseq1 = 0;
	  for (int is=0; is<_nbsamples; is++) 
	    {      
	      if (_clust_ids[is] == ic)
		{
		  _mut_model[_clust_ids[is]]->set_pat_rate_factor(_pat_rate_fact[_crypt2pat[is]]);
		  _mut_model[_clust_ids[is]]->changeperror(_perror[_clust_ids[is]]);
		  loglikseq1 += _crypt_sample[is]->computeLogLikSeq();
		}
	    }
	  double mutrate2m;
	  double mutrate2u;
	  double log_proposal;
// 	  if (rint==0)
// 	    {
// 	      mutrate2m = mutrate1m + gsl_ran_gaussian(_rng,mutrate1m/5);
// 	      mutrate2u = mutrate1u*(mutrate2m/mutrate1m);
// 	      log_proposal = log(gsl_ran_gaussian_pdf(mutrate2m-mutrate1m,mutrate2m/5))-log(gsl_ran_gaussian_pdf(mutrate2m-mutrate1m,mutrate1m/5));
// 	    } else {
// 	      mutrate2m = mutrate1m;
// 	      double ratio1 = mutrate1u/mutrate1m;
// 	      double ratio2 = ratio1 + gsl_ran_gaussian(_rng,ratio1/5);
// 	      mutrate2u = ratio2*mutrate2m;
// 	      log_proposal = log(gsl_ran_gaussian_pdf(ratio2-ratio1,ratio2/5))-log(gsl_ran_gaussian_pdf(ratio2-ratio1,ratio1/5));	      
// 	    }

	  if (rint==0)
	    {
	      double inverse1 = 1/mutrate1m;
	      double inverse2 = inverse1 + gsl_ran_gaussian(_rng,inverse1/5);
	      mutrate2m = 1/inverse2;
	      mutrate2u = mutrate1u*(mutrate2m/mutrate1m);
	      log_proposal = log(gsl_ran_gaussian_pdf(inverse2-inverse1,inverse2/5))-log(gsl_ran_gaussian_pdf(inverse2-inverse1,inverse1/5));

// 	      mutrate2m = mutrate1m + gsl_ran_gaussian(_rng,mutrate1m/5); 
// 	      mutrate2u = mutrate1u*(mutrate2m/mutrate1m);
// 	      log_proposal = log(gsl_ran_gaussian_pdf(mutrate2m-mutrate1m,mutrate2m/5))-log(gsl_ran_gaussian_pdf(mutrate2m-mutrate1m,mutrate1m/5));
	    } else {
 	      mutrate2m = mutrate1m;
 	      double ratio1 = mutrate1u/mutrate1m;
 	      double ratio2 = ratio1 + gsl_ran_gaussian(_rng,ratio1/5);
 	      mutrate2u = ratio2*mutrate2m;
 	      log_proposal = log(gsl_ran_gaussian_pdf(ratio2-ratio1,ratio2/5))-log(gsl_ran_gaussian_pdf(ratio2-ratio1,ratio1/5)) + log(gsl_ran_lognormal_pdf(ratio2, 0, 1.15)) - log(gsl_ran_lognormal_pdf(ratio1, 0, 1.15));	      
// 	      mutrate2m = mutrate1m;
// 	      double inverse1 = 1/mutrate1u;
// 	      double inverse2 = inverse1 + gsl_ran_gaussian(_rng,inverse1/5);
// 	      mutrate2u = 1/inverse2;
// 	      log_proposal = log(gsl_ran_gaussian_pdf(inverse2-inverse1,inverse2/5))-log(gsl_ran_gaussian_pdf(inverse2-inverse1,inverse1/5));
	    }
	  //	  cout << mutrate1m << "->" << mutrate2m << " " << mutrate1u << "->" << mutrate2u << endl;
	  if (mutrate2m > _mutratemin && mutrate2m < _mutratemax && mutrate2u > _mutratemin && mutrate2u < _mutratemax )
	    {
	      _mut_model[ic]->changemutrate_m(rmut_class, mutrate2m);
	      _mut_model[ic]->changemutrate_u(rmut_class, mutrate2u);

	      double loglikseq2 = 0;
	      for (int is=0; is<_nbsamples; is++) 
		{	  
		  if (_clust_ids[is] == ic)
		    {
		      _mut_model[_clust_ids[is]]->set_pat_rate_factor(_pat_rate_fact[_crypt2pat[is]]);
		      _mut_model[_clust_ids[is]]->changeperror(_perror[_clust_ids[is]]);
		      loglikseq2 += _crypt_sample[is]->computeLogLikSeq();
		    }
		}
	      //cout << "mutrate2 "<< mutrate2 <<" loglikseq " << loglikseq2 << endl;
	      double r = gsl_ran_flat(_rng, 0, 1);
	      //  cout << r << " " << loglikseq2 << " " << loglikseq1 << " " << log_proposal << endl; 
	      if (log(r) < loglikseq2-loglikseq1+log_proposal)
		{
		  _mutrate_m[ic][rmut_class] = mutrate2m;
		  _mutrate_u[ic][rmut_class] = mutrate2u;
		} else {
		  _mut_model[ic]->changemutrate_m(rmut_class, mutrate1m);
		  _mut_model[ic]->changemutrate_u(rmut_class, mutrate1u);
		}	      
	    }
	}
    }
}

void mcmc::updateMutrateNew()
{
  if (_nclusters > 1 && !_unique_mutmodel) 
    {
      //    cout << "updating the right thing" << endl;
      for (int ic=0; ic<_nclusters; ic++)
	{
	  //  cout << "_mut_model[ic] " << _mut_model[ic] << endl;
	  _mut_model[ic]->updateHyperParamSigmao(_lambda[ic]);
	  double loglikseq1 = 0;
	  for (int is=0; is<_nbsamples; is++) 
	    {      
	      if (_clust_ids[is] == ic)
		{
		  _mut_model[_clust_ids[is]]->set_pat_rate_factor(_pat_rate_fact[_crypt2pat[is]]);
		  _mut_model[_clust_ids[is]]->changeperror(_perror[_clust_ids[is]]);
		  loglikseq1 += _crypt_sample[is]->computeLogLikSeq();
		}
	    }
	  bool ok;
	  double log_proposal;
	  if (!_contextdep_mutmodel)
	    {
	      log_proposal = 
		_mut_model[ic]->proposeModelUpdate(_lambda[ic],ok);
	    } else {
	      log_proposal = 
		_mut_model[ic]->proposeContextDepModelUpdate(_lambda[ic],ok);
	    }
	  if (ok)
	    {
	      double loglikseq2 = 0;
	      for (int is=0; is<_nbsamples; is++) 
		{      
		  if (_clust_ids[is] == ic)
		    {
		      _mut_model[_clust_ids[is]]->set_pat_rate_factor(_pat_rate_fact[_crypt2pat[is]]);
		      _mut_model[_clust_ids[is]]->changeperror(_perror[_clust_ids[is]]);
		      loglikseq2 += _crypt_sample[is]->computeLogLikSeq();
		    }
		}
	      double r = gsl_ran_flat(_rng, 0, 1);
	      // cout << "updateContextDepMutrate() " << r << " " << loglikseq2 << " " << loglikseq1 << " " << log_proposal << endl; 
	      if (log(r) < loglikseq2-loglikseq1+log_proposal)
		{
		  //cout << "accepted" << endl;
		} else {
		  //cout << "rejected" << endl;
		  _mut_model[ic]->cancelContextDepModelUpdate();
		}	      	  
	    }
	}
    } else {
      _mut_model[0]->updateHyperParamSigmao(_lambda[0]);
      double loglikseq1 = 0;
      for (int is=0; is<_nbsamples; is++) 
	{      
	  _mut_model[0]->set_pat_rate_factor(_pat_rate_fact[_crypt2pat[is]]);
	  _mut_model[0]->changeperror(_perror[_clust_ids[is]]);
	  loglikseq1 += _crypt_sample[is]->computeLogLikSeq();	  
	}
      bool ok;
      double log_proposal;
      if (!_contextdep_mutmodel)
	{
	  log_proposal = 
	    _mut_model[0]->proposeModelUpdate(_lambda[0],ok);
	} else {
	  log_proposal = 
	    _mut_model[0]->proposeContextDepModelUpdate(_lambda[0],ok);
	}
      if (ok)
	{
	  double loglikseq2 = 0;
	  for (int is=0; is<_nbsamples; is++) 
	    {      
	      _mut_model[0]->set_pat_rate_factor(_pat_rate_fact[_crypt2pat[is]]);
	      _mut_model[0]->changeperror(_perror[_clust_ids[is]]);
	      loglikseq2 += _crypt_sample[is]->computeLogLikSeq();
	    }
	  
	  double r = gsl_ran_flat(_rng, 0, 1);
	  // cout << "updateContextDepMutrate() " << r << " " << loglikseq2 << " " << loglikseq1 << " " << log_proposal << endl; 
	  if (log(r) < loglikseq2-loglikseq1+log_proposal)
	    {
	      //cout << "accepted" << endl;
	    } else {
	      //cout << "rejected" << endl;
	      _mut_model[0]->cancelContextDepModelUpdate();
	    }	      	  
	}
    }

}

void mcmc::updateAlpha(double alphamin, double alphamax)
{
  for (int ic=0; ic<_nclusters; ic++)
    {
      double lambda = _lambda[ic];
      double alpha_diff1 = _mut_model[ic]->alpha_diff();
      double alpha1 = lambda*alpha_diff1;
      double alpha2 = alpha1 + gsl_ran_gaussian(_rng,alpha1/5);
      double alpha_diff2 = alpha2/lambda;
      double log_proposal = log(gsl_ran_gaussian_pdf(alpha2-alpha1,alpha2/5))
	- log(gsl_ran_gaussian_pdf(alpha2-alpha1,alpha1/5));
      double log_prior = log(gsl_ran_exponential_pdf(alpha2,1)) 
	- log(gsl_ran_exponential_pdf(alpha1,1));
      double loglikseq1 = 0;
      double loglikseq2 = 0;
      if (alpha2>0 && alpha2>alphamin && alpha2<alphamax) 
	{
	  for (int is=0; is<_nbsamples; is++) 
	    {      
	      if (_clust_ids[is] == ic)
		{
		  _mut_model[_clust_ids[is]]->set_pat_rate_factor(_pat_rate_fact[_crypt2pat[is]]);
		  _mut_model[_clust_ids[is]]->changeperror(_perror[_clust_ids[is]]);
		  loglikseq1 += _crypt_sample[is]->computeLogLikSeq();
		}
	    }
	  _mut_model[ic]->set_alpha_diff(alpha_diff2);
	  for (int is=0; is<_nbsamples; is++) 
	    {      
	      if (_clust_ids[is] == ic)
		{
		  _mut_model[_clust_ids[is]]->set_pat_rate_factor(_pat_rate_fact[_crypt2pat[is]]);
		  _mut_model[_clust_ids[is]]->changeperror(_perror[_clust_ids[is]]);
		  loglikseq2 += _crypt_sample[is]->computeLogLikSeq();
		}
	    }
	  double r = gsl_ran_flat(_rng, 0, 1);
	  //cout << "updateAlpha() " << alpha1 << "->" << alpha2 << " paccept " << exp(loglikseq2-loglikseq1+log_proposal+log_prior) << " " << loglikseq2 << " " << loglikseq1 << " " << log_proposal << " " << log_prior << endl; 
	  if (log(r) < loglikseq2-loglikseq1+log_proposal+log_prior)
	    {
	      //  cout << "accepted" << endl;
	    } else {
	      // cout << "rejected" << endl;	      
	      _mut_model[ic]->set_alpha_diff(alpha_diff1);
	    }	      	  
	}
    }
}

void mcmc::updateSwitchRate()
{
  if (_nb_mutclass==1) 
    {
      return;
    }
  int rmut_class = gsl_rng_uniform_int(_rng, _nb_mutclass-1);
  int rate_inc = gsl_ran_bernoulli(_rng, 0.1);

  for (int ic=0; ic<_nclusters; ic++)
    {
      double loglikseq1 = 0;
      for (int is=0; is<_nbsamples; is++) 
	{      
	  if (_clust_ids[is] == ic)
	    {
	      _mut_model[_clust_ids[is]]->set_pat_rate_factor(_pat_rate_fact[_crypt2pat[is]]);
	      _mut_model[_clust_ids[is]]->changeperror(_perror[_clust_ids[is]]);
	      loglikseq1 += _crypt_sample[is]->computeLogLikSeq();
	    }
	}
	  
      double rate1;
      if (rate_inc==1)
	{
	  rate1 = _switch_rates_inc[ic][rmut_class];
	} else {
	  rate1 = _switch_rates_dec[ic][rmut_class];
	}
      //cout << "ic " << ic << " " << mutrate1 << mutrate1 <<" loglikseq " << loglikseq1 << endl;
      double inverse1 = 1/rate1;
      double inverse2 = inverse1 + gsl_ran_gaussian(_rng,inverse1/5);
      double rate2 = 1/inverse2;
      double log_proposal = log(gsl_ran_gaussian_pdf(inverse2-inverse1,inverse2/5))-log(gsl_ran_gaussian_pdf(inverse2-inverse1,inverse1/5));
      
      if (rate2 > _mutratemin && rate2 < _mutratemax)
	{
	  if (rate_inc==1)
	    {	  
	      _mut_model[ic]->changeswitchrate(rate2, rmut_class, rmut_class+1);
	    } else {
	      _mut_model[ic]->changeswitchrate(rate2, rmut_class+1, rmut_class);
	    }

	  double loglikseq2 = 0;
	  for (int is=0; is<_nbsamples; is++) 
	    {	  
	      if (_clust_ids[is] == ic)
		{
		  _mut_model[_clust_ids[is]]->set_pat_rate_factor(_pat_rate_fact[_crypt2pat[is]]);
		  _mut_model[_clust_ids[is]]->changeperror(_perror[_clust_ids[is]]);
		  loglikseq2 += _crypt_sample[is]->computeLogLikSeq();
		}
	    }
	  //cout << "mutrate2 "<< mutrate2 <<" loglikseq " << loglikseq2 << endl;
	  double r = gsl_ran_flat(_rng, 0, 1);
	  if (log(r) < loglikseq2-loglikseq1+log_proposal)
	    {
	      if (rate_inc==1)
		{	  
		  _switch_rates_inc[ic][rmut_class] = rate2;
		} else {
		  _switch_rates_dec[ic][rmut_class] = rate2;
		}
	      //cout << "Accept change mutrate p=" << exp(loglikseq2-loglikseq1) 
	      //   << " " << mutrate1 << " -> " << mutrate2 << endl;
	    } else {
	      if (rate_inc==1)
		{	  
		  _mut_model[ic]->changeswitchrate(rate1, rmut_class, rmut_class+1);
		} else {
		  _mut_model[ic]->changeswitchrate(rate1, rmut_class+1, rmut_class);
		}
	    }
	  
	}
      //      cout << "sample mutrate " << _mut_model->mutrate() << endl;  
    }
}


void mcmc::updatePatRateFact() 
{
  this->updatePatRateFactHyperP();
  int ip = gsl_rng_uniform_int(_rng, _nb_patients);
  double pat_rate_fact1 = _pat_rate_fact[ip];
  double pat_rate_fact2 = pat_rate_fact1 + gsl_ran_gaussian(_rng,pat_rate_fact1/5);
  double log_proposal = log(gsl_ran_gaussian_pdf(pat_rate_fact2-pat_rate_fact1, pat_rate_fact2/5))
    -log(gsl_ran_gaussian_pdf(pat_rate_fact2-pat_rate_fact1, pat_rate_fact1/5));
  //double logpriorratio = log(gsl_ran_exponential_pdf(pat_rate_fact2,1)) - log(gsl_ran_exponential_pdf(pat_rate_fact1,1));
  double logpriorratio = log(gsl_ran_gamma_pdf(pat_rate_fact2,_pat_rate_hyperp,1./_pat_rate_hyperp)) - log(gsl_ran_gamma_pdf(pat_rate_fact1,_pat_rate_hyperp,1./_pat_rate_hyperp));
  double loglikseq1 = 0;
  double loglikseq2 = 0;
  for (int is=0; is<_nbsamples; is++) 
    {
      if (_crypt2pat[is]==ip) 
	{
	  _mut_model[_clust_ids[is]]->set_pat_rate_factor(pat_rate_fact1);
	  _mut_model[_clust_ids[is]]->changeperror(_perror[_clust_ids[is]]);
	  loglikseq1 += _crypt_sample[is]->computeLogLikSeq();
	  _mut_model[_clust_ids[is]]->set_pat_rate_factor(pat_rate_fact2);
	  _mut_model[_clust_ids[is]]->changeperror(_perror[_clust_ids[is]]);
	  loglikseq2 += _crypt_sample[is]->computeLogLikSeq();
	}
    }
  double r = gsl_ran_flat(_rng, 0, 1);
  //  cout << "ip " << ip << " pat_rate_fact1 " << pat_rate_fact1 << " pat_rate_fact2 " << pat_rate_fact2 << "\n";
  //cout << "loglikseq1 " << loglikseq1 << " loglikseq2 " << loglikseq2 << endl;
  //cout << "log_proposal " << log_proposal << " logpriorratio " << logpriorratio << endl;

  if (log(r) < logpriorratio+loglikseq2-loglikseq1+log_proposal)
    {
      _pat_rate_fact[ip] = pat_rate_fact2;
      //  cout << "Accepted" << endl;
    } else {
      // noting
    }
}

void mcmc::updatePatRateFactHyperP() 
{
  double hp1 = _pat_rate_hyperp;
  double hp2 = hp1 + gsl_ran_gaussian(_rng,hp1/5);
  if (hp2>0 && hp2<100) 
    {
      double log_proposal = log(gsl_ran_gaussian_pdf(hp2-hp1, hp2/5))
	-log(gsl_ran_gaussian_pdf(hp2-hp1, hp1/5));
      double logpriorratio = 0;
      double loglik1 = 0;
      double loglik2 = 0;
      for (int ip=0; ip<_nb_patients; ip++) 
	{
	  loglik1 += log(gsl_ran_gamma_pdf(_pat_rate_fact[ip],hp1,1./hp1));
	  loglik2 += log(gsl_ran_gamma_pdf(_pat_rate_fact[ip],hp2,1./hp2));
	}
      double r = gsl_ran_flat(_rng, 0, 1);

      if (log(r) < logpriorratio+loglik2-loglik1+log_proposal)
	{
	  _pat_rate_hyperp = hp2;
	  //  cout << "Accepted" << endl;
	} else {
	  // noting
	}      
    }
}

void mcmc::updateLambda(bool inverse_prior)
{
  //bool inverse_prior = false;//true;
  if (!_unique_Nlambda || _nclusters==1) {
    for (int ic=0; ic<_nclusters; ic++)
      {
	double lambda1; 
	double lambda2; 
	double log_proposal; 
	if (!inverse_prior)
	  {
	    lambda1 = _lambda[ic];
	    lambda2 = lambda1 + gsl_ran_gaussian(_rng,lambda1/5);
	    log_proposal = 
	      log(gsl_ran_gaussian_pdf(lambda2-lambda1, lambda2/5))
	      -log(gsl_ran_gaussian_pdf(lambda2-lambda1, lambda1/5));
	  } else {
	    lambda1 = _lambda[ic];
	    double inverse1 = 1/lambda1;
	    double inverse2 = inverse1 + gsl_ran_gaussian(_rng,inverse1/5);
	    lambda2 = 1/inverse2;
	    log_proposal = 
	      log(gsl_ran_gaussian_pdf(inverse2-inverse1, inverse2/5))
	      -log(gsl_ran_gaussian_pdf(inverse2-inverse1, inverse1/5));
	  }

	if (lambda2 > _lambdamin && lambda2 < _lambdamax)
	  {
	    //cout << "ic " << ic << " lambda1 " << lambda1 << " lambda2 " << lambda2 << endl;
	    double logliktree1 = 0;
	    double logliktree2 = 0;
	    
	    // theta unchanged, mutmodel changed
	    double loglikseq1 = 0;
	    double loglikseq2 = 0;
      
	    for (int is=0; is<_nbsamples; is++) 
	      {
		if (_clust_ids[is] == ic)
		  {		  
		    _mut_model[_clust_ids[is]]->set_pat_rate_factor(_pat_rate_fact[_crypt2pat[is]]);
		    _mut_model[_clust_ids[is]]->changeperror(_perror[_clust_ids[is]]);
		    logliktree1 += _crypt_sample[is]->computeLogLikTree();
		    loglikseq1 += _crypt_sample[is]->computeLogLikSeq();
		  }
	      }
	    _mut_model[ic]->changeLambdaKeepTheta(lambda1,lambda2);
	    for (int is=0; is<_nbsamples; is++) 
	      {
		if (_clust_ids[is] == ic)
		  {		  
		    _mut_model[_clust_ids[is]]->set_pat_rate_factor(_pat_rate_fact[_crypt2pat[is]]);
		    _mut_model[_clust_ids[is]]->changeperror(_perror[_clust_ids[is]]);
		    _crypt_sample[is]->setlambda(lambda2);
		    logliktree2 += _crypt_sample[is]->computeLogLikTree();
		    loglikseq2 += _crypt_sample[is]->computeLogLikSeq();
		  }
	      }
	    double r = gsl_ran_flat(_rng, 0, 1);
	    // 	  cout << " logliktree2 " << logliktree2 << " logliktree1 " << logliktree1 
	    // 	       << " loglikseq2 " << loglikseq2
	    // 	       << " loglikseq1 " << loglikseq1 << " log_proposal " << log_proposal << endl;
	    if (log(r) < logliktree2-logliktree1+loglikseq2-loglikseq1+log_proposal)
	      {
		_nb_accept_lambdaup++;
		_lambda[ic] = lambda2;
		//cout << "Accept change lambda p=" << exp(logliktree2-logliktree1) 
		//   << " " << lambda1 << " -> " << lambda2 << endl;
	      } else {
		_mut_model[ic]->cancelContextDepModelUpdate();	      
		for (int is=0; is<_nbsamples; is++) 
		  {
		    if (_clust_ids[is] == ic)
		      {		  
			_crypt_sample[is]->setlambda(lambda1);
		      }
		  }
		//cout << "Reject change mutrate p=" << exp(logliktree2-logliktree1) 
		//   << " " << lambda1 << " -> " << lambda2 << endl;
	      }
	    
	  }
	//cout << "sample lambda " << _crypt_sample[0]->lambda() << endl;  
      }
  } else { //_nclusters > 1 && _unique_Nlambda
    double lambda1; 
    double lambda2; 
    double log_proposal; 
    if (!inverse_prior)
      {
	lambda1 = _lambda[0];
	lambda2 = lambda1 + gsl_ran_gaussian(_rng,lambda1/5);
	log_proposal = 
	  log(gsl_ran_gaussian_pdf(lambda2-lambda1, lambda2/5))
	  -log(gsl_ran_gaussian_pdf(lambda2-lambda1, lambda1/5));
      } else {
	lambda1 = _lambda[0];
	double inverse1 = 1/lambda1;
	double inverse2 = inverse1 + gsl_ran_gaussian(_rng,inverse1/5);
	lambda2 = 1/inverse2;
	log_proposal = 
	  log(gsl_ran_gaussian_pdf(inverse2-inverse1, inverse2/5))
	  -log(gsl_ran_gaussian_pdf(inverse2-inverse1, inverse1/5));
      }

    if (lambda2 > _lambdamin && lambda2 < _lambdamax)
      {
	//cout << "ic " << ic << " lambda1 " << lambda1 << " lambda2 " << lambda2 << endl;
	double logliktree1 = 0;
	double logliktree2 = 0;
	
	// theta unchanged, mutmodel changed
	double loglikseq1 = 0;
	double loglikseq2 = 0;
      
	for (int is=0; is<_nbsamples; is++) 
	  {
	    _mut_model[_clust_ids[is]]->set_pat_rate_factor(_pat_rate_fact[_crypt2pat[is]]);
	    _mut_model[_clust_ids[is]]->changeperror(_perror[_clust_ids[is]]);
	    logliktree1 += _crypt_sample[is]->computeLogLikTree();
	    loglikseq1 += _crypt_sample[is]->computeLogLikSeq();	  
	  }
	if (!_unique_mutmodel) 
	  {
	    for (int ic=0; ic<_nclusters; ic++) 
	      {
		_mut_model[ic]->changeLambdaKeepTheta(lambda1,lambda2);
	      }
	  } else {
		_mut_model[0]->changeLambdaKeepTheta(lambda1,lambda2);	    
	  }
	for (int is=0; is<_nbsamples; is++) 
	  {
	    _mut_model[_clust_ids[is]]->set_pat_rate_factor(_pat_rate_fact[_crypt2pat[is]]);
	    _mut_model[_clust_ids[is]]->changeperror(_perror[_clust_ids[is]]);
	    _crypt_sample[is]->setlambda(lambda2);
	    logliktree2 += _crypt_sample[is]->computeLogLikTree();
	    loglikseq2 += _crypt_sample[is]->computeLogLikSeq();
	  }
	double r = gsl_ran_flat(_rng, 0, 1);
	// 	  cout << " logliktree2 " << logliktree2 << " logliktree1 " << logliktree1 
	// 	       << " loglikseq2 " << loglikseq2
	// 	       << " loglikseq1 " << loglikseq1 << " log_proposal " << log_proposal << endl;
	if (log(r) < logliktree2-logliktree1+loglikseq2-loglikseq1+log_proposal)
	  {
	    _nb_accept_lambdaup++;
	    
	    for (int ic=0; ic<_nclusters; ic++) 
	      {
		_lambda[ic] = lambda2;
	      }
	    //cout << "Accept change lambda p=" << exp(logliktree2-logliktree1) 
	    //   << " " << lambda1 << " -> " << lambda2 << endl;
	  } else {
	    if (!_unique_mutmodel) 
	      {
		for (int ic=0; ic<_nclusters; ic++) 
		  {
		    _mut_model[ic]->cancelContextDepModelUpdate();
		  }
	      } else {
		_mut_model[0]->cancelContextDepModelUpdate();
	      }
	    	      
	    for (int is=0; is<_nbsamples; is++) 
	      {
		_crypt_sample[is]->setlambda(lambda1);	      
	      }
	  }
	//cout << "sample lambda " << _crypt_sample[0]->lambda() << endl;  
      }
  }
}

void mcmc::updateg()
{
  if (!_unique_Nlambda || _nclusters==1) {
    for (int ic=0; ic<_nclusters; ic++)
      {
	double g1; 
	double g2; 
	double log_proposal; 

	g1 = _g[ic];
	g2 = g1 + gsl_ran_gaussian(_rng,g1/5);
	log_proposal = 
	  log(gsl_ran_gaussian_pdf(g2-g1, g2/5))
	  -log(gsl_ran_gaussian_pdf(g2-g1, g1/5));

	if (g2 > _gmin && g2 < _gmax)
	  {
	    double logliktree1 = 0;
	    double logliktree2 = 0;
      
	    for (int is=0; is<_nbsamples; is++) 
	      {
		if (_clust_ids[is] == ic)
		  {		  
		    _mut_model[_clust_ids[is]]->set_pat_rate_factor(_pat_rate_fact[_crypt2pat[is]]);
		    _mut_model[_clust_ids[is]]->changeperror(_perror[_clust_ids[is]]);
		    logliktree1 += _crypt_sample[is]->computeLogLikTree();
		  }
	      }
	    for (int is=0; is<_nbsamples; is++) 
	      {
		if (_clust_ids[is] == ic)
		  {		  
		    _mut_model[_clust_ids[is]]->set_pat_rate_factor(_pat_rate_fact[_crypt2pat[is]]);
		    _mut_model[_clust_ids[is]]->changeperror(_perror[_clust_ids[is]]);
		    _crypt_sample[is]->setg(g2);
		    logliktree2 += _crypt_sample[is]->computeLogLikTree();
		  }
	      }
	    double r = gsl_ran_flat(_rng, 0, 1);
	    if (log(r) < logliktree2-logliktree1+log_proposal)
	      {
		_g[ic] = g2;
	      } else {
		for (int is=0; is<_nbsamples; is++) 
		  {
		    if (_clust_ids[is] == ic)
		      {		  
			_crypt_sample[is]->setg(g1);
		      }
		  }
	      }
	    
	  }
      }
  } else { //_nclusters > 1 && _unique_Nlambda
    double g1; 
    double g2; 
    double log_proposal; 

    g1 = _g[0];
    g2 = g1 + gsl_ran_gaussian(_rng,g1/5);
    log_proposal = 
      log(gsl_ran_gaussian_pdf(g2-g1, g2/5))
      -log(gsl_ran_gaussian_pdf(g2-g1, g1/5));

    if (g2 > _gmin && g2 < _gmax)
      {
	double logliktree1 = 0;
	double logliktree2 = 0;
	
	for (int is=0; is<_nbsamples; is++) 
	  {
	    _mut_model[_clust_ids[is]]->set_pat_rate_factor(_pat_rate_fact[_crypt2pat[is]]);
	    _mut_model[_clust_ids[is]]->changeperror(_perror[_clust_ids[is]]);
	    logliktree1 += _crypt_sample[is]->computeLogLikTree();
	  }
	for (int is=0; is<_nbsamples; is++) 
	  {
	    _mut_model[_clust_ids[is]]->set_pat_rate_factor(_pat_rate_fact[_crypt2pat[is]]);
	    _mut_model[_clust_ids[is]]->changeperror(_perror[_clust_ids[is]]);
	    _crypt_sample[is]->setg(g2);
	    logliktree2 += _crypt_sample[is]->computeLogLikTree();
	  }
	double r = gsl_ran_flat(_rng, 0, 1);
	if (log(r) < logliktree2-logliktree1+log_proposal)
	  {
	    for (int ic=0; ic<_nclusters; ic++) 
	      {
		_g[ic] = g2;
	      }
	  } else {
	    for (int is=0; is<_nbsamples; is++) 
	      {
		_crypt_sample[is]->setg(g1);	      
	      }
	  }
      }
  }
}


void mcmc::updateLambda2()
{
  for (int ic=0; ic<_nclusters; ic++)
    {
      double lambda1 = _lambda[ic];
      double alpha = gsl_ran_gamma(_rng, 25,1./25.);
      double alpha_tilde = 1/alpha;
      double lambda2 = alpha*lambda1;
      double log_proposal = 
	log(gsl_ran_gamma_pdf(alpha_tilde, 25, 1./25.))
	-log(gsl_ran_gamma_pdf(alpha, 25., 1./25.));

      double log_jacobian = (_mut_model[ic]->dimModel()+1)*log(1/alpha);

      double log_ratio_theta = 
	_mut_model[ic]->computeLogRatioLambdaChange(lambda1, lambda2);

      
      if (lambda2 > _lambdamin && lambda2 < _lambdamax)
	{
	  //cout << "ic " << ic << " lambda1 " << lambda1 << " lambda2 " << lambda2 << endl;
	  double logliktree1 = 0;
	  double logliktree2 = 0;
	  for (int is=0; is<_nbsamples; is++) 
	    {
	      if (_clust_ids[is] == ic)
		{		  
		  logliktree1 += _crypt_sample[is]->computeLogLikTree();
		  _crypt_sample[is]->setlambda(lambda2);
		  logliktree2 += _crypt_sample[is]->computeLogLikTree();
		}
	    }

	  double r = gsl_ran_flat(_rng, 0, 1);
	  if (log(r) < logliktree2-logliktree1+log_proposal+log_jacobian+log_ratio_theta)
	    {
	      _lambda[ic] = lambda2;
	      //cout << "Accept change lambda p=" << exp(logliktree2-logliktree1) 
	      //   << " " << lambda1 << " -> " << lambda2 << endl;
	    } else {	      
	      for (int is=0; is<_nbsamples; is++) 
		{
		  if (_clust_ids[is] == ic)
		    {		  
		      _crypt_sample[is]->setlambda(lambda1);
		    }
		}
	      //cout << "Reject change mutrate p=" << exp(logliktree2-logliktree1) 
	      //   << " " << lambda1 << " -> " << lambda2 << endl;
	    }
	  
	}
      //cout << "sample lambda " << _crypt_sample[0]->lambda() << endl;  
    }
}


void mcmc::updateLambda1()
{
  for (int ic=0; ic<_nclusters; ic++)
    {
//       double lambda1 = _lambda[ic];
//       double inverse1 = 1/lambda1;
//       double inverse2 = inverse1 + gsl_ran_gaussian (_rng, inverse1/5);
//       double lambda2 = 1/inverse2;
//       double log_proposal = log(gsl_ran_gaussian_pdf(inverse2-inverse1, inverse2/5))
// 	-log(gsl_ran_gaussian_pdf(inverse2-inverse1, inverse1/5));

      double lambda1 = _lambda[ic];
      double lambda2 = lambda1 + gsl_ran_gaussian (_rng, lambda1/5);
      double log_proposal = log(gsl_ran_gaussian_pdf(lambda2-lambda1, lambda2/5))
	-log(gsl_ran_gaussian_pdf(lambda2-lambda1, lambda1/5));


      if (lambda2 > _lambdamin && lambda2 < _lambdamax)
	{
	  //cout << "ic " << ic << " lambda1 " << lambda1 << " lambda2 " << lambda2 << endl;
	  double logliktree1 = 0;
	  double logliktree2 = 0;
	  for (int is=0; is<_nbsamples; is++) 
	    {
	      if (_clust_ids[is] == ic)
		{		  
		  logliktree1 += _crypt_sample[is]->computeLogLikTree();
		  _crypt_sample[is]->setlambda(lambda2);
		  logliktree2 += _crypt_sample[is]->computeLogLikTree();
		  //cout << is << " " << logliktree1 << " " << logliktree2 <<  endl;
		}
	    }
	  //cout << "logliktree1 " << logliktree1 << " logliktree2 " << logliktree2 << endl;
	  double r = gsl_ran_flat(_rng, 0, 1);
	  if (log(r) < logliktree2-logliktree1+log_proposal)
	    {
	      _lambda[ic] = lambda2;
	      //cout << "Accept change lambda p=" << exp(logliktree2-logliktree1) 
	      //   << " " << lambda1 << " -> " << lambda2 << endl;
	    } else {
	      for (int is=0; is<_nbsamples; is++) 
		{
		  if (_clust_ids[is] == ic)
		    {
		      _crypt_sample[is]->setlambda(lambda1);
		    }
		}
	      //cout << "Reject change mutrate p=" << exp(logliktree2-logliktree1) 
	      //   << " " << lambda1 << " -> " << lambda2 << endl;
	    }
	  
	}
      //cout << "sample lambda " << _crypt_sample[0]->lambda() << endl;  
    }
}

void mcmc::updatePerror()
{
  if (_unique_perror) 
    {
      int sum1 = 0;
      int sum2 = 0;
      for (int is=0; is<_nbsamples; is++) 
	{      
	  _crypt_sample[is]->countErrors(sum1, sum2);
	}
      //cout << "sum1 " << sum1 << " sum2 " << sum2 << endl;
      _perror[0] = gsl_ran_beta(_rng,sum1+1,sum2+1);
      _mut_model[0]->changeperror(_perror[0]);	  
      for (int ic=1 ; ic<_nclusters ; ic++)
	{
	  _perror[ic] = _perror[0];
	  _mut_model[ic]->changeperror(_perror[0]);	  
	}
    } else {
      for (int ic=0; ic<_nclusters; ic++)
	{
	  int sum1 = 0;
	  int sum2 = 0;
	  for (int is=0; is<_nbsamples; is++) 
	    {      
	      if (_clust_ids[is] == ic)
		{
		  _crypt_sample[is]->countErrors(sum1, sum2);
		}
	    }
	  _perror[ic] = gsl_ran_beta(_rng,sum1+1,sum2+1);
	  _mut_model[ic]->changeperror(_perror[ic]);	  
	  //cout << "ic " << ic << " perror " << _perror[ic] << endl;
	}
    }
}

void mcmc::updateAllN(int& Nmin, int& Nmax, int* Ncounts, int clust)
{
  int rint = gsl_ran_bernoulli(_rng,0.5);
  bool ok = true;
  double log_lik_den = 0;
  double log_lik_num = 0;
  if (rint == 0)
    {
      // decrease
      //cout << "updateAllN() decrease attempt" << endl;
      for (int is=0; is<_nbsamples; is++)
	{
	  if (_clust_ids[is] == clust || clust==-1)
	    {
	      if (_crypt_sample[is]->Nstemcell()==_nstemcells_prior_min) 
		{
		  return;
		}
	    }
	}
      for (int is=0; is<_nbsamples; is++)
	{	      
	  if (_clust_ids[is] == clust || clust==-1)
	    {
	      log_lik_den += _crypt_sample[is]->computeLogLikN(); // WARNING MAKE SENSE IF COAL RATE DOESN'T DEPEND ON N
	      bool ok1 = _crypt_sample[is]->setNstemcell(_crypt_sample[is]->Nstemcell()-1);
	      ok = ok && ok1;
	    }
	}    
    } else {
      // increase
      //cout << "updateAllN() increase attempt" << endl;
      for (int is=0; is<_nbsamples; is++)
	{
	  if (_clust_ids[is] == clust || clust==-1)
	    {
	      if (_crypt_sample[is]->Nstemcell()==_nstemcells_prior_max) 
		{
		  return;
		}
	    }
	}
      for (int is=0; is<_nbsamples; is++)
	{
	  if (_clust_ids[is] == clust || clust==-1)
	    {
	      log_lik_den += _crypt_sample[is]->computeLogLikN(); // WARNING MAKE SENSE IF COAL RATE DOESN'T DEPEND ON LAMBDA
	      bool ok1 = _crypt_sample[is]->setNstemcell(_crypt_sample[is]->Nstemcell()+1);
	      ok = ok && ok1;
	    }    
	}
    }
      
  if (ok)
    {
      for (int is=0; is<_nbsamples; is++)
	{
	  if (_clust_ids[is] == clust || clust==-1)
	    {
	      log_lik_num += _crypt_sample[is]->computeLogLikN(); // WARNING MAKE SENSE IF COAL RATE DOESN'T DEPEND ON LAMBDA
	    }
	}
    }
  //cout << "log_lik_den " << log_lik_den  << " log_lik_num  " << log_lik_num << endl;
  double r = gsl_ran_flat(_rng, 0, 1);
  if (ok && log(r) < log_lik_num-log_lik_den)
    {
      //cout << "Accepted AllN update p=" << exp(log_lik_num-log_lik_den)  << endl;
      if (rint == 0)
	{
	  for (int is=0; is<_nbsamples; is++)
	    {
	      if (_clust_ids[is] == clust || clust==-1)
		{
		  int N = _crypt_sample[is]->Nstemcell();
		  Ncounts[N+1]--;
		  Ncounts[N]++;
		}
	    }
	  Nmin--;
	  Nmax--;
	} else {
	  for (int is=0; is<_nbsamples; is++)
	    {
	      if (_clust_ids[is] == clust || clust==-1)
		{
		  int N = _crypt_sample[is]->Nstemcell();
		  Ncounts[N-1]--;
		  Ncounts[N]++;
		}
	    }
	  Nmin++;
	  Nmax++;
	}
    } else {
      //cout << "Rejected AllN update p=" << exp(log_lik_num-log_lik_den)  << endl;
      if (rint == 0)
	{
	  for (int is=0; is<_nbsamples; is++)
	    {
	      if (_clust_ids[is] == clust || clust==-1)
		{
		  _crypt_sample[is]->setNstemcell(_crypt_sample[is]->Nstemcell()+1);
		} 
	    }
	} else {
	  for (int is=0; is<_nbsamples; is++)
	    {
	      if (_clust_ids[is] == clust || clust==-1)
		{
		  _crypt_sample[is]->setNstemcell(_crypt_sample[is]->Nstemcell()-1);
		} 
	    }
	}	  
    }
    
}

void mcmc::updatePcluster()
{
  for (int ic=0; ic<_nclusters ; ic++)
    {
      _pclusters_dirpseudocounts[ic] = 1;
    }
  for (int is=0 ; is<_nbsamples ; is++)
    {
      _pclusters_dirpseudocounts[_clust_ids[is]]++;
    }  
  gsl_ran_dirichlet(_rng, _nclusters, _pclusters_dirpseudocounts, _pclusters);
  //cout << "pcluster ";
  for (int ic=0; ic<_nclusters ; ic++)
    {
      //cout << _pclusters[ic] << " ";
    }  
  //cout << endl;
}

void mcmc::updateClusterAlloc(int* Nmin_byclust,int* Nmax_byclust,int** Ncounts_byclust)
{
  for (int ic=0 ; ic<_nclusters ; ic++)
    {
      _clusters_counts[ic]=0;
      //cout << ic << " "<< _switch_rates_inc[0] << endl;
    }
  for (int is=0 ; is<_nbsamples ; is++)
    {
      _clusters_counts[_clust_ids[is]]++;
    }

  for (int i=0 ; i<_nbsamples ; i++)
    {
//       cout << "3.0 _mutrate_m[1][0] " << _mutrate_m[1][0] << endl;
//       cout << "3.0 _mutrate_m[0][0] " << _mutrate_m[0][0] << endl;
      int is = gsl_rng_uniform_int(_rng, _nbsamples);
      int c1=_clust_ids[is];
      int c2= gsl_rng_uniform_int(_rng, _nclusters);
      if (c1 != c2 && _clusters_counts[c1]>1)
	{
	  int N1=_crypt_sample[is]->Nstemcell();
	  int countN1a = Ncounts_byclust[c1][N1];
	  int countN1b;
	  if (Nmin_byclust[c1]==Nmax_byclust[c1])
	    {
	      countN1b = 0;
	    } else if (Nmin_byclust[c1]==N1) {
	      countN1b = Ncounts_byclust[c1][Nmax_byclust[c1]];
	    } else {
	      countN1b = Ncounts_byclust[c1][Nmin_byclust[c1]];
	    }

	  //cout << "N1 " << N1 << " countN1a " << countN1a << " countN1b " << countN1b << endl;
	  double log_ratio_proposal = 0;
	  double log_ratio_prior_N = 0;

	  log_ratio_prior_N += log(likratioN1N2(countN1a, countN1b, 1, false));
	  if (Nmin_byclust[c1]==Nmax_byclust[c1] || Ncounts_byclust[c1][N1]==1) 
	    {
	      log_ratio_proposal += log(1./3.);
	    } else {
	      log_ratio_proposal += log(1./2.);
	    }

	  int N2 = -1;
	  if (Nmin_byclust[c2]==Nmax_byclust[c2]) 
	    {
	      log_ratio_proposal -= log(1./3.);
	      N2 = Nmin_byclust[c2]+gsl_rng_uniform_int(_rng, 3)-1;
	    } else {
	      log_ratio_proposal -= log(1./2.);
	      N2 = Nmin_byclust[c2]+gsl_rng_uniform_int(_rng, 2);
	    }
	  int countN2a = Ncounts_byclust[c2][N2];
	  int countN2b = -1;
	  if (Nmin_byclust[c2]==Nmax_byclust[c2] && N2 == Nmin_byclust[c2]) 
	    {
	      countN2b = 0;
	    } else if (Nmin_byclust[c2]==Nmax_byclust[c2] && N2 != Nmin_byclust[c2]) {
	      countN2b = Ncounts_byclust[c2][Nmin_byclust[c2]];
	    } else if (N2 == Nmin_byclust[c2]) {
	      countN2b = Ncounts_byclust[c2][Nmax_byclust[c2]];
	    } else if (N2 == Nmax_byclust[c2]) {
	      countN2b = Ncounts_byclust[c2][Nmin_byclust[c2]];
	    }
	  //cout << "N2 " << N2 << " countN2a " << countN2a << " countN2b " << countN2b << endl;
	  log_ratio_prior_N += log(likratioN1N2(countN2a, countN2b, 1, true));	  
	  
	  //cout << "c1 " << Nmin_byclust[c1] << " " << Nmax_byclust[c1] << " c2 " << Nmin_byclust[c2] << " " << Nmax_byclust[c2] << endl;
	  //cout << "c1:" << endl;
	  for (int i=_nstemcells_prior_min; i<=_nstemcells_prior_max; i++)
	    {
	      //cout << Ncounts_byclust[c1][i] << " ";
	    }
	  //cout << endl;
	  //cout << "c2:" << endl;
	  for (int i=_nstemcells_prior_min; i<=_nstemcells_prior_max; i++)
	    {
	      //cout << Ncounts_byclust[c2][i] << " ";
	    }
	  //cout << endl;

	  int sum1 = 0;
	  int sum2 = 0;
	  _crypt_sample[is]->countErrors(sum1,sum2);

	  double log_dens_den = 0;
	  log_dens_den += log(_pclusters[c1]);
	  _mut_model[c1]->set_pat_rate_factor(_pat_rate_fact[_crypt2pat[is]]);
	  _mut_model[c1]->changeperror(_perror[c1]);
	  log_dens_den += _crypt_sample[is]->computeLogLikTreeN();
	  log_dens_den += _crypt_sample[is]->computeLogLikSeq();
	  log_dens_den += sum1*log(_perror[c1]) + sum2*log(1-_perror[c1]);

	  double log_dens_num = 0;
	  _crypt_sample[is]->changeMutmodel(_mut_model[c2]);
	  _mut_model[c2]->set_pat_rate_factor(_pat_rate_fact[_crypt2pat[is]]);
	  _mut_model[c2]->changeperror(_perror[c2]);
	  _crypt_sample[is]->setlambda(_lambda[c2]);
	  _crypt_sample[is]->setNstemcell(N2);
	  //cout << "1 log_dens_num " << log_dens_num << endl;
	  log_dens_num += log(_pclusters[c2]);
	  //cout << "2 log_dens_num " << log_dens_num << endl;
	  log_dens_num += _crypt_sample[is]->computeLogLikTreeN();
	  //cout << "3 log_dens_num " << log_dens_num << endl;
	  log_dens_num += _crypt_sample[is]->computeLogLikSeq();
	  //cout << "4 log_dens_num " << log_dens_num << endl;
	  log_dens_num += sum1*log(_perror[c2]) + sum2*log(1-_perror[c2]);

	  double r = gsl_ran_flat(_rng, 0, 1);
	  //cout << "log_dens_num " << log_dens_num << " log_dens_den " << log_dens_den << " log_ratio_proposal " << log_ratio_proposal << " log_ratio_prior_N " << log_ratio_prior_N << endl;
	  double paccept = exp(log_dens_num - log_dens_den +  log_ratio_proposal + log_ratio_prior_N);
	  //cout << "paccept " << paccept << endl;
	  if (r<paccept && N2>=_nstemcells_prior_min && N2<=_nstemcells_prior_max) 
	    {
	      //cout << "Accepted p=" << paccept << " is " << c1 << " -> " << c2 << endl;
	      _clust_ids[is] = c2;
	      if (Ncounts_byclust[c1][N1]==1) 
		{
		  if (N1 == Nmin_byclust[c1])
		    {
		      Nmin_byclust[c1] = Nmax_byclust[c1];
		    } else {
		      Nmax_byclust[c1] = Nmin_byclust[c1];		      
		    }
		}
	      Ncounts_byclust[c1][N1]--;
	      _clusters_counts[c1]--;
	      Ncounts_byclust[c2][N2]++;	      
	      _clusters_counts[c2]++;
	      if (Ncounts_byclust[c2][N2]==1) 
		{
		  if (N2 < Nmin_byclust[c2])
		    {
		      Nmin_byclust[c2] = N2;
		    } else {
		      Nmax_byclust[c2] = N2;
		    }
		}
	    } else {
	      //cout << "Rejected p=" << paccept << " is " << c1 << " -> " << c2 << endl;	      
	      _crypt_sample[is]->changeMutmodel(_mut_model[c1]);

	      _crypt_sample[is]->setlambda(_lambda[c1]);	      
	      _crypt_sample[is]->setNstemcell(N1);	      
	    }
	} else {
	  //cout << "rejected cond. no cluster empty" << endl;
	}
    }

  //cout << "*******************" << endl;
}

double mcmc::likratioN1N2(int N1, int N2, int modif, bool increase)
{
  //cout << "double mcmc::likratioN1N2(" << N1 << ", "<< N2<< ", "<< modif <<", " << increase << ") " << endl;
  double likratio = 1;
  int N1n, N2n;
  if (modif==1)
    {
      if (increase)
	{
	  N1n = N1+1;
	} else {
	  N1n = N1-1;
	}
      N2n = N2;
    } else {
      if (increase)
	{
	  N2n = N2+1;
	} else {
	  N2n = N2-1;
	}
      N1n = N1;
    }

  //cout << "N1n " << N1n << " N2n " << N2n << endl;
  //cout << "_nstemcells_prior_max " << _nstemcells_prior_max << " _nstemcells_prior_min " << _nstemcells_prior_min << endl;
  likratio *= 1+(_nstemcells_prior_max-_nstemcells_prior_min)*(N1+N2);

  //cout << "likratio0 " << likratio << endl;
  likratio /= 1+(_nstemcells_prior_max-_nstemcells_prior_min)*(N1n+N2n);

  //cout << "likratio1 " << likratio << endl;
  if (increase)
    {
      if (modif == 1)
	{
	  likratio *= (double)N1n/(double)(N1n+N2n);
	} else {
	  likratio *= (double)N2n/(double)(N1n+N2n);	  
	}
    } else {
      if (modif == 1)
	{
	  likratio *= (double)(N1+N2)/(double)N1;
	} else {
	  likratio *= (double)(N1+N2)/(double)N2;	  
	}
    }
  //cout << "likratio2 " << likratio << endl;
  return likratio;
}

//       if (_clusters_counts[c]>1)
// 	{
// 	  double sumtot = 0;
// 	  int sum1 = 0;
// 	  int sum2 = 0;
// 	  _crypt_sample[is]->countErrors(sum1,sum2);
// 	  for (int ic=0; ic<_nclusters; ic++)
// 	    {
// 	      _crypt_sample[is]->mutmodel()->changeperror(_perror[ic]);
// 	      _crypt_sample[is]->mutmodel()->changemutrate(_mutrate[ic]);
// 	      _crypt_sample[is]->setlambda(_lambda[ic]);
// 	      _pclusters_dirpseudocounts[ic] = log(_pclusters[ic]);
// 	      _pclusters_dirpseudocounts[ic] += _crypt_sample[is]->computeLogLikTreeN();
// 	      _pclusters_dirpseudocounts[ic] += _crypt_sample[is]->computeLogLikSeq();
// 	      _pclusters_dirpseudocounts[ic] += sum1*log(_perror[ic]) + sum2*log(1-_perror[ic]);
// 	      _pclusters_dirpseudocounts[ic] = exp(_pclusters_dirpseudocounts[ic]);
// 	      sumtot += _pclusters_dirpseudocounts[ic];
// 	    }
// 	  double r = gsl_ran_flat(_rng, 0, sumtot);
// 	  double sum = _pclusters_dirpseudocounts[0];
// 	  int ri = 0;
// 	  while (r>sum) 
// 	    {
// 	      ri++;
// 	      sum += _pclusters_dirpseudocounts[ri];
// 	    }
// 	  _clust_ids[is] = ri;
// 	  _crypt_sample[is]->mutmodel()->changeperror(_perror[ri]);
// 	  _crypt_sample[is]->mutmodel()->changemutrate(_mutrate[ri]);
// 	  _crypt_sample[is]->setlambda(_lambda[ri]); 
// 	  _clusters_counts[c]--;
// 	  _clusters_counts[ri]++;
// 	} else {
// 	  //cout << "rejected cond. no cluster empty" << endl;
// 	}
