/* Copyright (C) 2006 Pierre Nicolas
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef _MCMC_H_
#define _MCMC_H_

#include <string>
#include <vector>

#include <gsl/gsl_rng.h>

#include "mutmodel.h"
#include "crypt.h"
#include "params.h"

using namespace std;

class mcmc
{

 protected:

  Mutmodel **_mut_model;
  Mutmodel *_mut_model_test;
  Crypt** _crypt_sample;
  gsl_rng* _rng;
  time_t _seed;

  int _nb_mutclass;
  double** _switch_rates_inc;
  double** _switch_rates_dec;

  bool _sym_mutrate;
  bool _unique_perror;
  bool _unique_Nlambda;
  bool _unique_mutmodel;
  bool _no_clusterupdate;
  bool _contextdep_mutmodel;

  double **_mutrate_m;
  double **_mutrate_u;

  double *_lambda;
  double *_g;
  double *_perror;
  int _nbsamples;

  int _nclusters;
  int *_clust_ids;
  double *_pclusters;
  double *_pclusters_dirpseudocounts;
  int *_clusters_counts;

  double _mutratemin;
  double _mutratemax;
  double _lambdamin;
  double _lambdamax;
  double _gmin;
  double _gmax;

  int _nstirling2;
  double** _stirling2;

  int _nstemcells_prior_max;
  int _nstemcells_prior_min;
  int _nstemcells_diff_max;
  double* _nb_stemcells_dirichletprior;
  double* _nb_stemcells_pseudocounts;
  int* _nb_stemcells_counts;
  double* _nb_stemcells_distrib;
  int _nstemcells_max;
  int _nstemcells_min;

  int _nb_accept_lambdaup;

  int _nb_patients;
  int* _crypt2pat;
  double* _pat_rate_fact;
  double _pat_rate_hyperp;
  
  void stirlingAlloc();
 public:
  mcmc(Params& params, string datafilename);
  ~mcmc();
  void setPrior(Params& params);
  void initPatRateFact();
  void initClustAlloc();
  void samplingFancy(Params& params, string outfilename);
  void testPeeling(int niter, string outfilename);
  void updateMutrate();
  //void updateContextDepMutrate();
  void updatePclusters();
  void updateLambda(bool inverse_prior); 
  void updateg();
  void updateLambda1(); 
  void updateLambda2(); 
  void updateNprior();
  void updateSupportN();
  void updateSwitchRate();
  void initNcounts();
  void updatePerror();
  void updateAlpha(double alphamin, double alphamax);
  void updatePatRateFact();
  void updatePatRateFactHyperP();

  bool printStateFancySampling(int* Nmin_byclust, int* Nmax_byclust, 
			       int** Ncounts_byclust, ofstream& out, 
			       ofstream& out_alloc);
  void updateAllN(int& Nmin, int& Nmax, int* Ncounts, int clust);
  void predictiveSimulFromFile(string paramfile, string allocfile, string patratesfile, 
			       string outstatfile);
  void simul(int nsimul, string outstatfile, string outseqfile, bool printseq);
  void updatePcluster();
  void updateClusterAlloc(int* Nmin_byclust,int* Nmax_byclust,int** Ncounts_byclust);
  double likratioN1N2(int N1, int N2, int modif, bool increase);
  void updateMutrateNew();
};

#endif
