/* Copyright (C) 2006 Pierre Nicolas
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "mutmodel.h"

#include <iostream>
#include <gsl/gsl_randist.h>

#include "HTNT/tnt_array2d.h"
#include "HTNT/jama_eig.h"
#include "HTNT/jama_lu.h"

using namespace std;
using namespace TNT;
using namespace JAMA;


Mutmodel::Mutmodel(int nb_mutclass, vector<double> mutrate, 
		   vector<double> switch_rates_inc, vector<double> switch_rates_dec,
		   double p_error)
{
  _p_error = p_error;
  _unique_rate = true;
  _nb_mutclass = nb_mutclass;
  _pat_rate_fact = 1;  
  _alpha_diff = 0.5;

  _switch_rate = new double* [_nb_mutclass];
  _mutrate = new double [_nb_mutclass];
  _mutrate_old = new double [_nb_mutclass];
  for (int imut=0; imut<_nb_mutclass; imut++)
    {
      _switch_rate[imut] = new double [_nb_mutclass];
      _mutrate[imut] = mutrate[imut];
      for (int jmut=0; jmut<_nb_mutclass; jmut++)
	{
	  _switch_rate[imut][jmut] = 0;
	}
    }
   for (int imut=0; imut<_nb_mutclass; imut++)
    {
      if (imut+1<_nb_mutclass)
	{
	  _switch_rate[imut][imut+1] = switch_rates_inc[imut];
	} 
      if (imut-1>=0)
	{
	  _switch_rate[imut][imut-1] = switch_rates_dec[imut-1];
	}       
    } 
   for (int imut=0; imut<_nb_mutclass; imut++)
    {
      double sum = 0;
      for (int jmut=0; jmut<_nb_mutclass; jmut++)
	{
	  if (imut != jmut)
	    {
	      sum += _switch_rate[imut][jmut];
	    }
	}      
      _switch_rate[imut][imut] = -sum;
    }
   _contextdep = false;
}

Mutmodel::Mutmodel(int nb_mutclass, vector<double> mutrate_m, 
		   vector<double> mutrate_u, 
		   vector<double> switch_rates_inc, vector<double> switch_rates_dec, 
		   double p_error, gsl_rng * rng)
{
  _rng = rng;
  _p_error = p_error;
  _unique_rate = false;
  _nb_mutclass = nb_mutclass;
  _pat_rate_fact = 1;  
  _alpha_diff = 0.5;

  _sigmao = 1;
  _muo = 0;
  _switch_rate = new double* [_nb_mutclass];
  _mutrate_m = new double [_nb_mutclass];
  _mutrate_u = new double [_nb_mutclass];
  _mutrate_m_old = new double [_nb_mutclass];
  _mutrate_u_old = new double [_nb_mutclass];
  for (int imut=0; imut<_nb_mutclass; imut++)
    {
      _switch_rate[imut] = new double [_nb_mutclass];
      _mutrate_m[imut] = mutrate_m[imut];
      _mutrate_u[imut] = mutrate_u[imut];
      for (int jmut=0; jmut<_nb_mutclass; jmut++)
	{
	  _switch_rate[imut][jmut] = 0;
	}
    }
   for (int imut=0; imut<_nb_mutclass; imut++)
    {
      if (imut+1<_nb_mutclass)
	{
	  _switch_rate[imut][imut+1] = switch_rates_inc[imut];
	} 
      if (imut-1>=0)
	{
	  _switch_rate[imut][imut-1] = switch_rates_dec[imut-1];
	}       
    } 
   for (int imut=0; imut<_nb_mutclass; imut++)
    {
      double sum = 0;
      for (int jmut=0; jmut<_nb_mutclass; jmut++)
	{
	  if (imut != jmut)
	    {
	      sum += _switch_rate[imut][jmut];
	    }
	}      
      _switch_rate[imut][imut] = -sum;
    }
   _contextdep = false;
}


Mutmodel::Mutmodel(int ltag, int nb_contextdep_mutrate, vector<double> mutrate_m, 
		   vector<double> mutrate_u, 
		   double p_error, gsl_rng * rng)
{
  _ltag = ltag;
  _pat_rate_fact = 1; 
  _alpha_diff = 0.5;
 

//   _nbcontextdep_mutrate = 1;
//   _limit_contextdep_mutrate = new int [_nbcontextdep_mutrate+1];
//   _limit_contextdep_mutrate[0] = 0;
//   _limit_contextdep_mutrate[1] = _ltag+1;
//   _nbcontextdep_mutrate_old = 1;

  _nbcontextdep_mutrate =  nb_contextdep_mutrate;
  if (_nbcontextdep_mutrate > _ltag+1) 
    {
      cerr << "Error: _nbcontextdep_mutrate cannot be greater than the length of the patterns" << endl;
      cerr << "_nbcontextdep_mutrate changed from " << _nbcontextdep_mutrate << " to " << _ltag+1 << endl;
      _nbcontextdep_mutrate = _ltag+1;
    }
  _limit_contextdep_mutrate = new int [_nbcontextdep_mutrate+1];
  for (int i=0; i<_nbcontextdep_mutrate; i++) 
    {
      _limit_contextdep_mutrate[i] = i;
    }
  _limit_contextdep_mutrate[_nbcontextdep_mutrate] = _ltag+1;
  _nbcontextdep_mutrate_old = _nbcontextdep_mutrate;

  _limit_contextdep_mutrate_old = new int [_nbcontextdep_mutrate+1];

   _contextdep = true;
  _nb_mutclass = 1;
  _mutrate_context_m = new double [ltag+1];
  _mutrate_context_u = new double [ltag+1];
  _mutrate_context_m_old = new double [ltag+1];
  _mutrate_context_u_old = new double [ltag+1];
  _p_error = p_error;
  _unique_rate = false;
  _rng = rng;

  for (int i=0; i<=_ltag; i++)
    {
      _mutrate_context_m[i] = mutrate_m[i];
      _mutrate_context_u[i] = mutrate_u[i];
      //cout << "_mutrate_context_m["<< i<< "]" << _mutrate_context_m[i] << endl;
      //cout << "_mutrate_context_m " << _mutrate_context_m << endl;
    }
  double sum_mu = 0;
  for (int i=0; i<_nbcontextdep_mutrate; i++)
    {      
      sum_mu += _mutrate_context_m[_limit_contextdep_mutrate[i]];
      sum_mu += _mutrate_context_u[_limit_contextdep_mutrate[i]];
    }
  _muo = sum_mu/(2*_nbcontextdep_mutrate);
  _sigmao = 1;

  _ntag=1;
  _codfact = new int [_ltag];
  for (int i=0; i<_ltag; i++)
    {
      _codfact[_ltag-i-1]=_ntag;
      _ntag*=2;
    }
  
  _cod2tag = new int* [_ntag];
  for (int itag=0; itag<_ntag; itag++)
    {
      _cod2tag[itag] = new int [_ltag];
      int cod=itag;      
      //cout << itag << " ";
      for (int ipos=0; ipos<_ltag; ipos++)
	{
	  int x=cod/_codfact[ipos];
	  _cod2tag[itag][ipos] = x;
	  //cout << x;
	  cod = cod%_codfact[ipos];
	}
      //cout << endl;
    }

  _int_ptrans1 = new double [(_ltag/2+2)*(_ltag/2+2)];
  _int_ptrans2 = new double [(_ltag/2+2)*(_ltag/2+2)];
  _int_ptrans3 = new double [(_ltag/2+2)*(_ltag/2+2)];
  _int_postpr = new double [_ntag];
  _int_exp = new double [(_ltag/2+2)*(_ltag/2+2)];

  //cout << "_ntag " << _ntag << endl;
  _equivtag = new int* [_ntag];
  _nbmethtag = new int [_ntag];
  _nbeqclass_bytag = new int [_ntag];
  _nbtag_byeqclass_bytag = new int* [_ntag];
  _tag_byeqclass_bytag = new int** [_ntag];

  //cout << "_equivtag" << endl;
  for (int itag=0; itag<_ntag; itag++)
    {
      _equivtag[itag] = new int [_ntag];
      
      int n0=0;
      for (int ipos=0; ipos<_ltag; ipos++)
	{
	  if (_cod2tag[itag][ipos]==0)
	    {
	      n0++;
	    }
	}
      int n1=_ltag-n0;
      _nbmethtag[itag] = n1;
      _nbeqclass_bytag[itag] = (n0+1)*(n1+1);
      _nbtag_byeqclass_bytag[itag] = new int [_nbeqclass_bytag[itag]];
      _tag_byeqclass_bytag[itag] =  new int* [_nbeqclass_bytag[itag]];
      for (int i=0;i<_nbeqclass_bytag[itag]; i++)
	{
	  _nbtag_byeqclass_bytag[itag][i]=0;
	}
      for (int jtag=0; jtag<_ntag; jtag++)
	{
	  int d0=0;
	  int d1=0;
	  for (int ipos=0; ipos<_ltag; ipos++)
	    {
	      if (_cod2tag[itag][ipos]==0 && _cod2tag[jtag][ipos]==1)
		{
		  d0++;
		} else if (_cod2tag[itag][ipos]==1 && _cod2tag[jtag][ipos]==0) {
		  d1++;
		} 
	    }
	  _equivtag[itag][jtag] = d0*(n1+1)+d1;
	  _nbtag_byeqclass_bytag[itag][_equivtag[itag][jtag]]++;
	  //cout << _equivtag[itag][jtag] << "\t";
	}
      for (int w=0; w<_nbeqclass_bytag[itag]; w++)
	{	  
	  _tag_byeqclass_bytag[itag][w] =  new int [_nbtag_byeqclass_bytag[itag][w]];
	  int n=0;
	  for (int jtag=0; jtag<_ntag; jtag++) 
	    {
	      if (_equivtag[itag][jtag] == w)
		{
		  _tag_byeqclass_bytag[itag][w][n] = jtag;
		  n++;
		}
	    }
	}
      //cout << endl;
    }

  //  cout << "1_mutrate_context_m["<< 0<< "]" << _mutrate_context_m[0] << endl;
  //cout << "_mutrate_context_m " << _mutrate_context_m << endl;
  //cout << "_ltag " << _ltag << endl;
  _dim_evd = new int [_ltag+1];
  _eigenvalues = new double* [_ltag+1];
  _P = new double** [_ltag+1];
  _Pinv = new double** [_ltag+1];
  _eigenvalues_old = new double* [_ltag+1];
  _P_old = new double** [_ltag+1];
  _Pinv_old = new double** [_ltag+1];
  //cout << "_mutrate_context_m " << _mutrate_context_m << endl;
  //cout << "2_mutrate_context_m["<< 0<< "]" << _mutrate_context_m[0] << endl;
  for (int n1=0; n1<=_ltag; n1++)
    {
      //cout << "2.1 " << n1 << "_mutrate_context_m["<< 0<< "]" << _mutrate_context_m[0] << endl;
      int n0=_ltag-n1;
      int dim = (n0+1)*(n1+1);
      _dim_evd[n1] = dim;
      _eigenvalues[n1] = new double [dim];
      _P[n1] = new double* [dim];
      _Pinv[n1] = new double* [dim];
      _eigenvalues_old[n1] = new double [dim];
      _P_old[n1] = new double* [dim];
      _Pinv_old[n1] = new double* [dim];
      for (int i=0; i<dim; i++)
	{
	  _P[n1][i] = new double [dim];
	  _Pinv[n1][i] = new double [dim];
	  _P_old[n1][i] = new double [dim];
	  _Pinv_old[n1][i] = new double [dim];
	}
    }
  //cout << "3_mutrate_context_m["<< 0<< "]" << _mutrate_context_m[0] << endl;
  this->computeContextTrans();
//   cout << "0 P 50 100 300 " << this->probatranscontext(50,100,300) << endl;
//   cout << "1 P 50 100 300 " << this->probatranscontext1(50,100,300) << endl;
//   this->prepProbaTransContext(_int_ptrans1, 50, 100, true);
//   cout << "3 P 50 100 300 " << this->probatranscontext(300) << endl;
//   cout << "3 P 50 100 100 " << this->probatranscontext(100) << endl;
//   this->prepProbaTransContext(_int_ptrans1, 50, 300, false);
//   cout << "4 P 50 100 300 " << this->probatranscontext(100) << endl;
}


Mutmodel::~Mutmodel()
{
  if (!_contextdep)
    {
      for (int imut=0; imut<_nb_mutclass; imut++)
	{
	  delete _switch_rate[imut];
	} 
      delete _switch_rate;
      if (_unique_rate)
	{
	  delete _mutrate;
	} else {
	  delete _mutrate_m;
	  delete _mutrate_u;
	}
    } else {
      delete _mutrate_m;
      delete _mutrate_u;      
    }
}

void Mutmodel::computeContextTrans()
{
  for (int n1=0; n1<=_ltag; n1++)
    {
      int n0 = _ltag - n1;
      int dim = _dim_evd[n1];

      // Rate Matrix
      Array2D<double> Rate(dim, dim);
      for (int i=0; i<dim; i++)
	{
	  for (int j=0; j<dim; j++)
	    {
	      Rate[i][j]=0;
	    }
	}
      
      for (int d0=0; d0<=n0; d0++)
	{
	  //cout << "d0 " << d0 << endl;
	  for (int d1=0; d1<=n1; d1++)
	    {
	      int irow = d0*(n1+1)+d1;	      
	      int nbmeth = d0+n1-d1;
	      double rate;
	      int icol;
	      int d0a, d1a;
	      
	      double sum=0;
	      
	      //cout << "_mutrate_context_m[" << nbmeth << "] " << _mutrate_context_m[nbmeth] << endl;
	      d0a = d0-1; d1a = d1; icol = d0a*(n1+1)+d1a;
	      if (d0a>=0)
		{
		  //  cout << irow << " " << icol << " " << d0 << " " << d1 << "/" << rate << "/ " << endl;
		  rate=d0*_mutrate_context_u[nbmeth];
		  sum+=rate;
		  Rate[irow][icol]=rate;
		}
	      d0a = d0+1; d1a = d1; icol = d0a*(n1+1)+d1a;
	      if (d0a<=n0)
		{
		  rate=(n0-d0)*_mutrate_context_m[nbmeth];
		  sum+=rate;
		  Rate[irow][icol]=rate;
		  //cout << irow << " " << icol << " " << d0 << " " << d1 << "/" << rate << "/ " << endl;
		}
	      d0a = d0; d1a = d1-1; icol = d0a*(n1+1)+d1a;
	      if (d1a>=0)
		{
		  //cout << irow << " " << icol << " " << d0 << " " << d1 << "/" << rate << "/ " << endl;
		  rate=d1*_mutrate_context_m[nbmeth];
		  sum+=rate;
		  Rate[irow][icol]=rate;
		}
	      d0a = d0; d1a = d1+1; icol = d0a*(n1+1)+d1a;
	      if (d1a<=n1)
		{
		  //cout << irow << " " << icol << " " << d0 << " " << d1 << "/" << rate << "/ " << endl;
		  rate=(n1-d1)*_mutrate_context_u[nbmeth];
		  sum+=rate;
		  Rate[irow][icol]=rate;
		}
	      Rate[irow][irow]=-sum;	
	      //cout << "rate " << rate << endl;
	    }
	}
      //cout << "Check Rate n1 " << n1 << endl;
      //cout << Rate << endl;

      //EVD
      Array2D<double> D(dim, dim);
      Array2D<double> P(dim, dim);
      Eigenvalue<double> Eig(Rate);
      Eig.getD(D);
      Eig.getV(P);
      LU<double> Lu(P);
      Array2D<double> I(dim,dim);
      for (int irow=0; irow<dim; irow++)
	{
	  for (int jcol=0; jcol<dim; jcol++)
	    {
	      I[irow][jcol]=0;
	    }
	  I[irow][irow]=1;
	}
      //cout << "1 I[0][0] " << I[0][0] << endl; 
      Array2D<double> Pinv = Lu.solve(I);
      //cout << "Check Prod Passage n1 " << n1 << endl;
      //cout << matmult(P,Pinv) << endl;

      //cout << "Check EVD n1 " << n1 << endl;
      //cout << matmult(P,matmult(D,Pinv)) << endl;
      
      //cout << "2 I[0][0] " << I[0][0] << endl; 
      for (int i=0; i<dim; i++)
	{
	  _eigenvalues[n1][i] = D[i][i];
	  for (int j=0; j<dim; j++)
	    {
	      _P[n1][i][j] = P[i][j];
	      _Pinv[n1][i][j] = Pinv[i][j];
	    }
	}
      
    }
  
}


double Mutmodel::probatranscontext(double length, int from_tag, int to_tag)
{
  length *= _pat_rate_fact;

  int n1 = _nbmethtag[from_tag];
  int eqclass = _equivtag[from_tag][to_tag];
  int dim = _dim_evd[n1];
  double proba=0;
  for (int i=0; i<dim; i++)
    {
      proba += _P[n1][0][i]*exp(_eigenvalues[n1][i]*length)*_Pinv[n1][i][eqclass];
      //  cout << i << " " << proba << " " << _eigenvalues[n1][i] << endl;
    }
  //cout << _nbtag_byeqclass_bytag[from_tag][eqclass] << endl;
  return proba/(double)_nbtag_byeqclass_bytag[from_tag][eqclass];
}

double Mutmodel::probatranscontext(double length, int* from_tag, int* to_tag)
{
//   cout << "probatranscontext(double length, int* from_tag, int* to_tag) " << endl;
//   for (int i=0; i<_ltag; i++)
//     {
//       cout << from_tag[i];
//     }
//   cout << endl;
//   int cod = this->codSeq(from_tag);
//   for (int i=0; i<_ltag; i++)
//     {
//       cout << _cod2tag[cod][i];
//     }
//   cout << endl;
  double val = this->probatranscontext(length, this->codSeq(from_tag), this->codSeq(to_tag));
  if (val < 0)
    {
      val = 1e-50;
    }
//   for (int i=0; i<_ltag; i++)
//     {
//       cout << from_tag[i];
//     }
//   cout << endl;
//   for (int i=0; i<_ltag; i++)
//     {
//       cout << to_tag[i];
//     }
//   cout << endl;
//   cout << val << " " << log(val) << " " << length << endl;
  return val;
}

double Mutmodel::probatranscontext(double length, int from_tag, int* to_tag)
{
  double val = this->probatranscontext(length, from_tag, this->codSeq(to_tag));
  if (val < 0)
    {
      val = 1e-50;
    }
  return val;
}

double Mutmodel::probatranscontext1(double length, int from_tag, int to_tag)
{
  length *= _pat_rate_fact;
  int n1 = _nbmethtag[to_tag];
  int eqclass = _equivtag[to_tag][from_tag];
  int dim = _dim_evd[n1];
  double proba=0;
  for (int i=0; i<dim; i++)
    {
      proba += _P[n1][eqclass][i]*exp(_eigenvalues[n1][i]*length)*_Pinv[n1][i][0];
    }
  return proba;
}

void Mutmodel::prepProbaTransContext(double* store, double length, int tag, bool fromb)
{
  length *= _pat_rate_fact;
  _from = fromb;
  if (_from)
    {
      _from_tag = tag; 
    } else {
      _to_tag = tag;
    }
  int n1 = _nbmethtag[tag];
  int dim = _dim_evd[n1];
  for (int i=0; i<dim; i++) 
    {
      _int_exp[i] = exp(_eigenvalues[n1][i]*length);
    }
  
  if (_from)
    {
      double checksum = 0;
      for (int w=0; w<dim; w++) 
	{
	  store[w] = 0;
	  for (int i=0; i<dim; i++)
	    {
	      store[w] += _P[n1][0][i]*_int_exp[i]*_Pinv[n1][i][w];
	    }
	  checksum += store[w];
	  store[w] /= (double)_nbtag_byeqclass_bytag[tag][w];
	}      
      // cout << "checksum " << checksum << endl;
    } else {
      for (int w=0; w<dim; w++) 
	{
	  store[w] = 0;
	  for (int i=0; i<dim; i++)
	    {
	      store[w] += _P[n1][w][i]*_int_exp[i]*_Pinv[n1][i][0];
	    }
	}            
    }
}

double Mutmodel::probatranscontext(int tag)
{
  int w;
  if (_from)
    {
      w = _equivtag[_from_tag][tag];
    } else {
      w = _equivtag[_to_tag][tag];    
    }
  return _int_ptrans1[w];
}

void Mutmodel::contextDepDrawSeq(double length, int* ancestor_seq, int* res_seq)
{
  int ancestor_cod = this->codSeq(ancestor_seq);
  this->prepProbaTransContext(_int_ptrans1, length, ancestor_cod, true);
  int n1 = _nbmethtag[ancestor_cod];
  int dim = _dim_evd[n1];
  
  double r = gsl_ran_flat(_rng, 0, 1);
  int w = 0;
  double sum_p = _int_ptrans1[w]*(double)_nbtag_byeqclass_bytag[ancestor_cod][w];    
  while(w+1<dim && sum_p < r) 
    {
      w++;
      sum_p += _int_ptrans1[w]*(double)_nbtag_byeqclass_bytag[ancestor_cod][w];
    }
  int rint =  gsl_rng_uniform_int(_rng, _nbtag_byeqclass_bytag[ancestor_cod][w]);
  int res_cod = _tag_byeqclass_bytag[ancestor_cod][w][rint];
  for (int i=0; i<_ltag; i++)
    {
      res_seq[i] = _cod2tag[res_cod][i];
    }
}

double Mutmodel::contextSumLikelihoodNode(double length_anc,double length_cle,
					  double length_cri,
					  int* seq_anc, int* seq_cle, int* seq_cri)
{
  int cod_anc = this->codSeq(seq_anc);
  int cod_cle = this->codSeq(seq_cle);
  int cod_cri = this->codSeq(seq_cri);
  bool allthesame = (cod_anc == cod_cle)&&(cod_anc == cod_cri);

  this->prepProbaTransContext(_int_ptrans1, length_anc, cod_anc, true);     
  this->prepProbaTransContext(_int_ptrans2, length_cle, cod_cle, false);     
  this->prepProbaTransContext(_int_ptrans3, length_cri, cod_cri, false);     

  if (!allthesame)
    {
      double sum=0;
      //   for (int i=0; i<_ltag; i++)
      //     {
      //       cout << seq_anc[i] << " ";
      //     }
      //   cout << endl;
      //cout << "Codes " << cod_anc << " " << cod_cle << " " << cod_cri << endl;
      for (int it=0; it<_ntag; it++)
	{
	  //cout << it << " " << _int_ptrans1[_equivtag[cod_anc][it]] << " " << _int_ptrans2[_equivtag[cod_cle][it]] << " " << _int_ptrans3[_equivtag[cod_cri][it]] << endl;
	  _int_postpr[it] = _int_ptrans1[_equivtag[cod_anc][it]]*_int_ptrans2[_equivtag[cod_cle][it]]*_int_ptrans3[_equivtag[cod_cri][it]];
	  sum += _int_postpr[it];
	}
      
      // By the way we scale
      for (int it=0; it<_ntag; it++)
	{
	  _int_postpr[it]/=sum;
	}
      return sum;
    } else {
      double sum = 0;
      int dim = _dim_evd[_nbmethtag[cod_anc]];
      for (int w=0; w<dim; w++)
	{
	  _int_postpr[w] = _int_ptrans1[w]*(double)_nbtag_byeqclass_bytag[cod_anc][w]*_int_ptrans2[w]*_int_ptrans3[w];
	  sum += _int_postpr[w];	  
	}      
      // By the way we scale
      for (int w=0; w<dim; w++)
	{
	  _int_postpr[w]/=sum;
	}
      
//       for (int w=0; w<dim; w++)
// 	{
// 	  cout << "(" << w << "," << _int_postpr[w] << ") ";
// 	}
      return sum;
    }
}

double Mutmodel::contextSumLikelihoodNodeType0(double length_anc,
					       int* seq_anc, int* seq_obs)
{
  int cod_anc = this->codSeq(seq_anc);

  this->prepProbaTransContext(_int_ptrans1, length_anc, cod_anc, true);     

  double sum=0;
  for (int it=0; it<_ntag; it++)
    {
      int ne=0;
      int n=0;
      for (int ipos=0; ipos<_ltag; ipos++) 
	{
	  if (seq_obs[ipos]>=0) 
	    {
	      n++;
	      if (seq_obs[ipos] != _cod2tag[it][ipos]) 
		{
		  ne++;
		}
	    }
	}
      _int_postpr[it] = _int_ptrans1[_equivtag[cod_anc][it]]*pow(_p_error,ne)*pow(1-_p_error,n-ne);
      sum += _int_postpr[it];
    }
      
  // By the way we scale
  for (int it=0; it<_ntag; it++)
    {
      _int_postpr[it]/=sum;
    }

  return sum;
}

void Mutmodel::contextPosteriorSimulNodeType0(double length_anc, 
					      int* seq_anc, int* seq_obs,
					      int* res_seq)
{
  this->contextSumLikelihoodNodeType0(length_anc,seq_anc,seq_obs);

  double r = gsl_ran_flat(_rng, 0, 1);

  int it = 0;
  double sum_p = _int_postpr[it];

  while(it+1<_ntag && sum_p < r) 
    {
      it++;
      sum_p += _int_postpr[it];
    }
  int restag = it;

  
  for (int i=0; i<_ltag; i++)
    {
      res_seq[i] =  _cod2tag[restag][i];
    }
}



void Mutmodel::contextPosteriorSimulNode(double length_anc, double length_cle,
					   double length_cri,
					   int* seq_anc, int* seq_cle, int* seq_cri,
					   int* res_seq)
{
//   cout << "origin" << endl;
//   for (int i=0; i<_ltag; i++)
//     {
//       cout << res_seq[i];
//     }
//   cout << endl;
  this->contextSumLikelihoodNode(length_anc,length_cle,length_cri,seq_anc,seq_cle,seq_cri);
  int cod_anc = this->codSeq(seq_anc);
  int cod_cle = this->codSeq(seq_cle);
  int cod_cri = this->codSeq(seq_cri);
  bool allthesame = (cod_anc == cod_cle)&&(cod_anc == cod_cri);
  double r = gsl_ran_flat(_rng, 0, 1);
  int restag;
  if (allthesame) 
    {
      int w = 0;
      double sum_p = _int_postpr[w];
      int cod_anc = this->codSeq(seq_anc);
      int dim = _dim_evd[_nbmethtag[cod_anc]];
      while(w+1<dim && sum_p < r) 
	{
	  w++;
	  sum_p += _int_postpr[w];
	}      
      int rint = gsl_rng_uniform_int(_rng, _nbtag_byeqclass_bytag[cod_anc][w]);
      restag = _tag_byeqclass_bytag[cod_anc][w][rint];      
    } else {
      int it = 0;
      double sum_p = _int_postpr[it];
//   for (int i=0; i<_ntag; i++)
//     {
//       cout << "(" << i << "," << _int_postpr[i] << ") ";
//     }
//   cout << endl;
      while(it+1<_ntag && sum_p < r) 
	{
	  it++;
	  sum_p += _int_postpr[it];
	}
      restag = it;
    }
  
  for (int i=0; i<_ltag; i++)
    {
      res_seq[i] =  _cod2tag[restag][i];
    }
//   cout << "simul" << endl;
//   for (int i=0; i<_ltag; i++)
//     {
//       cout << res_seq[i];
//     }
//   cout << endl;
}

void Mutmodel::printContextDepParams(ofstream& out)
{
  out << _muo << " ";
  out << _sigmao << " ";
  if (_contextdep)
    {
      for (int i=0; i<=_ltag; i++)
	{
	  out << _mutrate_context_m[i] << " ";
	}
      for (int i=0; i<=_ltag; i++)
	{
	  out << _mutrate_context_u[i] << " ";
	}
    } else {
      for (int i=0; i<_nb_mutclass; i++)
	{
	  out << _mutrate_m[i] << " ";
	}
      for (int i=0; i<_nb_mutclass; i++)
	{
	  out << _mutrate_u[i] << " ";
	}
    }
}

 
void Mutmodel::readContextDepParams(ifstream& in)
{
  if (_contextdep)
    {
      cout << "enter void Mutmodel::readContextDepParams(ifstream& in)" << endl;
      in >> _muo;
      cout << "_muo " << _muo << endl;
      in >> _sigmao;
      double rate;
      for (int i=0; i<= _ltag; i++)
	{
	  in >> rate;
	  _mutrate_context_m[i] = rate;
	  cout << rate << " ";
	}
      for (int i=0; i<= _ltag; i++)
	{
	  in >> rate;
	  _mutrate_context_u[i] = rate;
	  cout << rate << " ";
	}
      this->computeContextTrans();
      cout << "new _mutrate_context_m ";
      for (int i=0; i<= _ltag; i++)
	{
	  cout << _mutrate_context_m[i] << " ";
	}
      cout <<endl;
      cout << "new _mutrate_context_u ";
      for (int i=0; i<= _ltag; i++)
	{
	  cout <<_mutrate_context_u[i] << " ";
	}
      cout <<endl;
    } else {
      cout << "enter void Mutmodel::readContextDepParams(ifstream& in)" << endl;
      in >> _muo;
      cout << "_muo " << _muo << endl;
      in >> _sigmao;
      for (int i=0; i<_nb_mutclass; i++)
	{
	  in >> _mutrate_m[i];
	}
      for (int i=0; i<_nb_mutclass; i++)
	{
	  in >> _mutrate_u[i];
	}
      cout << "new _mutrate_m ";
      for (int i=0; i<_nb_mutclass; i++)
	{
	  cout << _mutrate_m[i] << " ";
	}
      cout << endl;
      cout << "new _mutrate_u ";
      for (int i=0; i<_nb_mutclass; i++)
	{
	  cout << _mutrate_u[i] << " ";
	}
      cout << endl;
    }
}


double Mutmodel::computeLogRatioLambdaChange(double lambda1, double lambda2)
{
  double loglik_1 = 0;
  double loglik_2 = 0;
  
  for (int i=0; i<_nbcontextdep_mutrate; i++)
    {
      double mu_m = _mutrate_context_m[_limit_contextdep_mutrate[i]];
      double mu_u = _mutrate_context_u[_limit_contextdep_mutrate[i]];

      loglik_1 += log(gsl_ran_lognormal_pdf(mu_m/lambda1, 0, _sigmao));
      loglik_1 += log(gsl_ran_lognormal_pdf(mu_u/lambda1, 0, _sigmao));
      loglik_2 += log(gsl_ran_lognormal_pdf(mu_m/lambda2, 0, _sigmao));
      loglik_2 += log(gsl_ran_lognormal_pdf(mu_u/lambda2, 0, _sigmao));
    }
  return loglik_2-loglik_1;
}

int Mutmodel::dimModel()
{
  return 2*_nbcontextdep_mutrate;
}


void Mutmodel::updateHyperParamSigmao(double lambda)
{
  double sigmao_1 = _sigmao;
  double sigmao_2 = sigmao_1 + gsl_ran_gaussian(_rng,sigmao_1/5);
  double log_proposal = 
    log(gsl_ran_gaussian_pdf(sigmao_2-sigmao_1,sigmao_2/5))
    -log(gsl_ran_gaussian_pdf(sigmao_2-sigmao_1,sigmao_1/5));
  double log_prior = 
    log(gsl_ran_exponential_pdf(sigmao_2,1)) 
    - log(gsl_ran_exponential_pdf(sigmao_1,1));
  double loglik_1 = 0;
  double loglik_2 = 0;
  if (_contextdep) {
    for (int i=0; i<_nbcontextdep_mutrate; i++)
      {
	double mu_m = _mutrate_context_m[_limit_contextdep_mutrate[i]];
	double mu_u = _mutrate_context_u[_limit_contextdep_mutrate[i]];
	
	loglik_1 += log(gsl_ran_lognormal_pdf(mu_m/lambda, 0, sigmao_1));
	loglik_1 += log(gsl_ran_lognormal_pdf(mu_u/lambda, 0, sigmao_1));
	loglik_2 += log(gsl_ran_lognormal_pdf(mu_m/lambda, 0, sigmao_2));
	loglik_2 += log(gsl_ran_lognormal_pdf(mu_u/lambda, 0, sigmao_2));
      }
  } else {
    for (int i=0; i<_nb_mutclass; i++)
      {
	double mu_m = _mutrate_m[i];
	double mu_u = _mutrate_u[i];
	
	loglik_1 += log(gsl_ran_lognormal_pdf(mu_m/lambda, 0, sigmao_1));
	loglik_1 += log(gsl_ran_lognormal_pdf(mu_u/lambda, 0, sigmao_1));
	loglik_2 += log(gsl_ran_lognormal_pdf(mu_m/lambda, 0, sigmao_2));
	loglik_2 += log(gsl_ran_lognormal_pdf(mu_u/lambda, 0, sigmao_2));
      }    
  }
    
  double r=gsl_ran_flat(_rng,0,1);
  if (log(r) < log_prior + log_proposal + loglik_2 - loglik_1)
    {
      _sigmao = sigmao_2;
    }  
}

double Mutmodel::proposeContextDepModelUpdate(double lambda, bool& ok)
{
  _alpha_diff_old = _alpha_diff;
  // Return log ratio proposal if ok==true
  ok=true;
  double log_proposal = 0;
  int rmov_type = gsl_rng_uniform_int(_rng, 2);
  if (rmov_type==0 && _nbcontextdep_mutrate>1)
    {
      // Change the limit of the breaks
      int r_break = 1;
      if (_nbcontextdep_mutrate>2)
	{
	  r_break = 1+gsl_rng_uniform_int(_rng, _nbcontextdep_mutrate-1);
	}
      int rmov_dir = gsl_rng_uniform_int(_rng, 2);
      int old_breakpos = _limit_contextdep_mutrate[r_break];
      int new_breakpos = _limit_contextdep_mutrate[r_break];
      //cout << "move 0 rbreak " << r_break << " rmov_dir " << rmov_dir << endl;
      if (rmov_dir==0)
	{
	  new_breakpos--;
	} else {
	  new_breakpos++;
	}
      if (new_breakpos <= _limit_contextdep_mutrate[r_break-1] 
	  || new_breakpos >= _limit_contextdep_mutrate[r_break+1])
	//	  || new_breakpos >= _ltag-2)
	{
	  ok = false;
	  return 0;
	}

      int *tmp_limit = _limit_contextdep_mutrate_old;
      _limit_contextdep_mutrate_old = _limit_contextdep_mutrate;
      _limit_contextdep_mutrate = tmp_limit;

      double* tmp__mutrate_context_m = _mutrate_context_m_old; 
      _mutrate_context_m_old = _mutrate_context_m;
      _mutrate_context_m = tmp__mutrate_context_m;
      double* tmp__mutrate_context_u = _mutrate_context_u_old; 
      _mutrate_context_u_old = _mutrate_context_u;
      _mutrate_context_u = tmp__mutrate_context_u;

      for (int i=0; i<_nbcontextdep_mutrate+1; i++)
	{
	  _limit_contextdep_mutrate[i] = _limit_contextdep_mutrate_old[i];
	}
      for (int i=0; i<=_ltag; i++)
	{
	  _mutrate_context_m[i] = _mutrate_context_m_old[i];
	  _mutrate_context_u[i] = _mutrate_context_u_old[i];
	}      

      _limit_contextdep_mutrate[r_break] = new_breakpos; 
      if (rmov_dir==0)
	{
	  _mutrate_context_m[old_breakpos-1] = _mutrate_context_m_old[old_breakpos];
	  _mutrate_context_u[old_breakpos-1] = _mutrate_context_u_old[old_breakpos];
	} else {
	  _mutrate_context_m[old_breakpos] = _mutrate_context_m_old[old_breakpos-1];
	  _mutrate_context_u[old_breakpos] = _mutrate_context_u_old[old_breakpos-1];
	}
    } else {
      int r_context = 0;
      if (_nbcontextdep_mutrate>1)
	{
	  r_context = gsl_rng_uniform_int(_rng, _nbcontextdep_mutrate);
	}
      int meth = gsl_rng_uniform_int(_rng, 2);
      double mutrate1m = _mutrate_context_m[_limit_contextdep_mutrate[r_context]];
      double mutrate1u = _mutrate_context_u[_limit_contextdep_mutrate[r_context]];
      double mutrate2m;
      double mutrate2u;
      double theta1;
      double theta2;
      if (meth==1) 
	{
	  theta1 = mutrate1m/lambda;
	  theta2 = theta1 + gsl_ran_gaussian(_rng, theta1/5);
	  mutrate2m = theta2*lambda;
	  mutrate2u = mutrate1u;
	} else {
	  theta1 = mutrate1u/lambda;
	  theta2 = theta1 + gsl_ran_gaussian(_rng, theta1/5);
	  mutrate2m = mutrate1m;	  
	  mutrate2u = theta2*lambda;
	}

      log_proposal = 
	log(gsl_ran_gaussian_pdf(theta2-theta1,theta2/5))
	- log(gsl_ran_gaussian_pdf(theta2-theta1,theta1/5)) 
	+ log(gsl_ran_lognormal_pdf(theta2, 0, _sigmao)) 
	- log(gsl_ran_lognormal_pdf(theta1, 0, _sigmao));

      int *tmp_limit = _limit_contextdep_mutrate_old;
      _limit_contextdep_mutrate_old = _limit_contextdep_mutrate;
      _limit_contextdep_mutrate = tmp_limit;
      
      double* tmp__mutrate_context_m = _mutrate_context_m_old; 
      _mutrate_context_m_old = _mutrate_context_m;
      _mutrate_context_m = tmp__mutrate_context_m;
      double* tmp__mutrate_context_u = _mutrate_context_u_old; 
      _mutrate_context_u_old = _mutrate_context_u;
      _mutrate_context_u = tmp__mutrate_context_u;

      for (int i=0; i<_nbcontextdep_mutrate+1; i++)
	{
	  _limit_contextdep_mutrate[i] = _limit_contextdep_mutrate_old[i];
	}
       for (int i=0; i<=_ltag; i++)
	{
	  _mutrate_context_m[i] = _mutrate_context_m_old[i];
	  _mutrate_context_u[i] = _mutrate_context_u_old[i];
	}
      for (int i=_limit_contextdep_mutrate[r_context]; i<_limit_contextdep_mutrate[r_context+1]; i++)
	{
	  _mutrate_context_m[i] = mutrate2m;
	  _mutrate_context_u[i] = mutrate2u;	  
	}      
    }
  
  double** tmp_eigenvalues = _eigenvalues_old;
  _eigenvalues_old = _eigenvalues;
  _eigenvalues = tmp_eigenvalues;
  double*** tmp_P = _P_old;
  _P_old = _P;
  _P = tmp_P;
  double*** tmp_Pinv = _Pinv_old;
  _Pinv_old = _Pinv;
  _Pinv = tmp_Pinv;
  this->computeContextTrans();
  return log_proposal;
}

double Mutmodel::proposeModelUpdate(double lambda, bool& ok)
{
  // Return log ratio proposal if ok==true
  ok=true;
  double log_proposal = 0;

  _alpha_diff_old = _alpha_diff;
  int r_mutclass = 0;
  if (_nb_mutclass>1)
    {
      r_mutclass = gsl_rng_uniform_int(_rng, _nb_mutclass);
    }

  if (_unique_rate) 
    {
      double mutrate1 = _mutrate[r_mutclass];
      double theta1 = mutrate1/lambda;
      double theta2 = theta1 + gsl_ran_gaussian(_rng, theta1/5);
      double mutrate2 = theta2*lambda;
      
      log_proposal = 
	log(gsl_ran_gaussian_pdf(theta2-theta1,theta2/5))
	- log(gsl_ran_gaussian_pdf(theta2-theta1,theta1/5)) 
	+ log(gsl_ran_lognormal_pdf(theta2, 0, _sigmao)) 
	- log(gsl_ran_lognormal_pdf(theta1, 0, _sigmao));
      
      double *mutrate_tmp = _mutrate;
      _mutrate = _mutrate_old;
      _mutrate_old = mutrate_tmp;
      for (int i=0; i<_nb_mutclass; i++)
	{
	  _mutrate[i] = _mutrate_old[i];
	}
      _mutrate[r_mutclass] = mutrate2;
      return log_proposal;
    } else {
      int meth = gsl_rng_uniform_int(_rng, 2);

      double mutrate1m = _mutrate_m[r_mutclass];
      double mutrate1u = _mutrate_u[r_mutclass];
      double mutrate2m;
      double mutrate2u;
      double theta1;
      double theta2;
      if (meth==1) 
	{
	  theta1 = mutrate1m/lambda;
	  theta2 = theta1 + gsl_ran_gaussian(_rng, theta1/5);
	  mutrate2m = theta2*lambda;
	  mutrate2u = mutrate1u;
	} else {
	  theta1 = mutrate1u/lambda;
	  theta2 = theta1 + gsl_ran_gaussian(_rng, theta1/5);
	  mutrate2m = mutrate1m;	  
	  mutrate2u = theta2*lambda;
	}
      
      log_proposal = 
	log(gsl_ran_gaussian_pdf(theta2-theta1,theta2/5))
	- log(gsl_ran_gaussian_pdf(theta2-theta1,theta1/5)) 
	+ log(gsl_ran_lognormal_pdf(theta2, 0, _sigmao)) 
	- log(gsl_ran_lognormal_pdf(theta1, 0, _sigmao));
      
      double *mutrate_tmp = _mutrate_m;
      _mutrate_m = _mutrate_m_old;
      _mutrate_m_old = mutrate_tmp;
      mutrate_tmp = _mutrate_u;
      _mutrate_u = _mutrate_u_old;
      _mutrate_u_old = mutrate_tmp;
      for (int i=0; i<_nb_mutclass; i++)
	{
	  _mutrate_m[i] = _mutrate_m_old[i];
	  _mutrate_u[i] = _mutrate_u_old[i];
	}
      _mutrate_m[r_mutclass] = mutrate2m;
      _mutrate_u[r_mutclass] = mutrate2u;
	
      return log_proposal;
    }
}

void Mutmodel::changeLambdaKeepTheta(double lambda1,double lambda2)
{
  _alpha_diff_old = _alpha_diff;
  _alpha_diff = _alpha_diff*(lambda1/lambda2);
  if (_contextdep)
    {
      int *tmp_limit = _limit_contextdep_mutrate_old;
      _limit_contextdep_mutrate_old = _limit_contextdep_mutrate;
      _limit_contextdep_mutrate = tmp_limit;
      
      double* tmp__mutrate_context_m = _mutrate_context_m_old; 
      _mutrate_context_m_old = _mutrate_context_m;
      _mutrate_context_m = tmp__mutrate_context_m;
      double* tmp__mutrate_context_u = _mutrate_context_u_old; 
      _mutrate_context_u_old = _mutrate_context_u;
      _mutrate_context_u = tmp__mutrate_context_u;
      
      for (int i=0; i<_nbcontextdep_mutrate+1; i++)
	{
	  _limit_contextdep_mutrate[i] = _limit_contextdep_mutrate_old[i];
	}
      for (int i=0; i<=_ltag; i++)
	{
	  _mutrate_context_m[i] = _mutrate_context_m_old[i]*(lambda2/lambda1);
	  _mutrate_context_u[i] = _mutrate_context_u_old[i]*(lambda2/lambda1);
	}

      double** tmp_eigenvalues = _eigenvalues_old;
      _eigenvalues_old = _eigenvalues;
      _eigenvalues = tmp_eigenvalues;
      double*** tmp_P = _P_old;
      _P_old = _P;
      _P = tmp_P;
      double*** tmp_Pinv = _Pinv_old;
      _Pinv_old = _Pinv;
      _Pinv = tmp_Pinv;
      this->computeContextTrans();
      return;
    } else {
      if (_unique_rate) 
	{
	  double *mutrate_tmp = _mutrate;
	  _mutrate = _mutrate_old;
	  _mutrate_old = mutrate_tmp;
	  for (int i=0; i<_nb_mutclass; i++)
	    {
	      _mutrate[i] = _mutrate_old[i]*(lambda2/lambda1);
	    }
	  return;
	} else if (!_unique_rate) {
	  double *mutrate_tmp = _mutrate_m;
	  _mutrate_m = _mutrate_m_old;
	  _mutrate_m_old = mutrate_tmp;
	  mutrate_tmp = _mutrate_u;
	  _mutrate_u = _mutrate_u_old;
	  _mutrate_u_old = mutrate_tmp;
	  for (int i=0; i<_nb_mutclass; i++)
	    {
	      _mutrate_m[i] = _mutrate_m_old[i]*(lambda2/lambda1);
	      _mutrate_u[i] = _mutrate_u_old[i]*(lambda2/lambda1);
	    }
	  return;
	}
    }
}


void Mutmodel::cancelContextDepModelUpdate()
{
  _alpha_diff = _alpha_diff_old;
  if (_contextdep)
    {
      int *tmp_limit = _limit_contextdep_mutrate;
      _limit_contextdep_mutrate = _limit_contextdep_mutrate_old;
      _limit_contextdep_mutrate_old = tmp_limit;
      
      double* tmp_mutrate_context_m = _mutrate_context_m; 
      _mutrate_context_m = _mutrate_context_m_old;
      _mutrate_context_m_old = tmp_mutrate_context_m;
      
      double* tmp_mutrate_context_u = _mutrate_context_u; 
      _mutrate_context_u = _mutrate_context_u_old;
      _mutrate_context_u_old = tmp_mutrate_context_u;
      
      double** tmp_eigenvalues = _eigenvalues;
      _eigenvalues = _eigenvalues_old;
      _eigenvalues_old = tmp_eigenvalues;
      double*** tmp_P = _P;
      _P = _P_old;
      _P_old = tmp_P;
      double*** tmp_Pinv = _Pinv;
      _Pinv = _Pinv_old;
      _Pinv_old = tmp_Pinv;
    } else {
       if (_unique_rate) 
	{
	  double *mutrate_tmp = _mutrate;
	  _mutrate = _mutrate_old;
	  _mutrate_old = mutrate_tmp;
	} else if (!_unique_rate) {
	  double *mutrate_tmp = _mutrate_m;
	  _mutrate_m = _mutrate_m_old;
	  _mutrate_m_old = mutrate_tmp;
	  mutrate_tmp = _mutrate_u;
	  _mutrate_u = _mutrate_u_old;
	  _mutrate_u_old = mutrate_tmp;
	}     
    }
}

double Mutmodel::averagemeth(double t)
{
  this->prepProbaTransContext(_int_ptrans1, t, 0, true);
  double sum = 0;
  double sump = 0;
  for (int itag=0; itag<_ntag; itag++)
    {
      double p = probatranscontext(itag);
      sum += p*_nbmethtag[itag];
      sump += p;
      //cout << itag << " " << _nbmethtag[itag] << endl;
    }
  //cout << "( " << sump << " , " << sum << ") ";
  //exit(1);
  return sum;
}

vector<double> Mutmodel::nbmethdist(double t)
{
  this->prepProbaTransContext(_int_ptrans1, t, 0, true);
  vector<double> dist(10,0);
  for (int itag=0; itag<_ntag; itag++)
    {
      double p = probatranscontext(itag);
      dist[_nbmethtag[itag]] += p;
    }
  return dist;
}
