/* Copyright (C) 2006 Pierre Nicolas
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef _MUTMODEL_H_
#define _MUTMODEL_H_

#include <vector>
#include <gsl/gsl_rng.h>
#include <cmath>
#include <iostream>
#include <fstream>
#include <gsl/gsl_rng.h>

using namespace std;

class Mutmodel
{

 protected:
  bool _contextdep;

  double *_mutrate;
  double *_mutrate_m;
  double *_mutrate_u;
  double *_mutrate_old;
  double *_mutrate_m_old;
  double *_mutrate_u_old;

  int _nb_mutclass;
  double** _switch_rate;

  double _p_error;
  bool _unique_rate;

  // Useful when contextdep == true
  int _ltag;
  int _nbcontextdep_mutrate;
  int* _limit_contextdep_mutrate;
  int _nbcontextdep_mutrate_old;
  int* _limit_contextdep_mutrate_old;
  
  double *_mutrate_context_m;
  double *_mutrate_context_u;
  double *_mutrate_context_m_old;
  double *_mutrate_context_u_old;
  int** _equivtag;
  int* _nbmethtag;
  int** _cod2tag;
  int _ntag;
  int* _codfact;
  int* _nbeqclass_bytag;
  int** _nbtag_byeqclass_bytag;
  int*** _tag_byeqclass_bytag;

  int* _dim_evd;
  double** _eigenvalues;
  double*** _P;
  double*** _Pinv;
  double** _eigenvalues_old;
  double*** _P_old;
  double*** _Pinv_old;

  double* _int_ptrans1;
  double* _int_ptrans2;
  double* _int_ptrans3;
  double* _int_postpr;
  double* _int_exp;
  int _from_tag;
  int _to_tag;
  bool _from;
  gsl_rng * _rng;

  double _muo; 
  double _sigmao;

  double _pat_rate_fact;
  double _alpha_diff;
  double _alpha_diff_old;

 public:
  Mutmodel(int nb_mutclass, vector<double> mutrate, 
	   vector<double> switch_rates_inc, vector<double> switch_rates_dec,
	   double p_error);
  Mutmodel(int nb_mutclass, vector<double> mutrate_m, vector<double> mutrate_u, 
	   vector<double> switch_rates_inc, vector<double> switch_rates_dec, 
	   double p_error, gsl_rng * rng);

  Mutmodel(int ltag, int nb_contextdep_mutrate, 
	   vector<double> mutrate_m, vector<double> mutrate_u, 
	   double p_error, gsl_rng * rng);
  
  ~Mutmodel();

  inline double computeBranchLength(double t1, double t2);
  inline int nmodal();
  inline double proba(double length, int imut, int from_allele, int to_allele);
  inline double probaswitch(double length, int imut, int jmut);
  inline void changeswitchrate(double newrate, int imut, int jmut);
  inline void changemutrate(int imut, double new_rate);
  inline double mutrate(int imut);
  inline void changemutrate_m(int imut, double new_rate);
  inline double mutrate_m(int imut);
  inline void changemutrate_u(int imut, double new_rate);
  inline double mutrate_u(int imut);
  inline bool unique_rate();
  inline void changeperror(double p_error);
  inline double perror();
  inline int nbmutclass();
  inline void set_pat_rate_factor(double fact);
  inline void set_alpha_diff(double alpha_diff);
  inline double alpha_diff();

  inline bool contextdep();
  void computeContextTrans();
  double probatranscontext(double length, int from_tag, int to_tag);
  double probatranscontext(double length, int* from_seq, int* to_seq);
  double probatranscontext(double length, int from_seq, int* to_seq);
  double probatranscontext1(double length, int from_tag, int to_tag);
  void prepProbaTransContext(double* store, double length, int tag, bool fromb);
  double probatranscontext(int tag);
  inline int codSeq(int* seq);
  void contextDepDrawSeq(double length, int* ancestor_seq, int* res_seq);
  double contextSumLikelihoodNode(double length_anc,double length_cle,
				  double length_cri,
				  int* seq_anc, int* seq_cle, int* seq_cri);
  double contextSumLikelihoodNodeType0(double length_anc,
				  int* seq_anc, int* seq_obs);

  void contextPosteriorSimulNode(double length_anc, double length_cle,
				   double length_cri,
				   int* seq_anc, int* seq_cle, int* seq_cri,
				   int* res_seq);
 void contextPosteriorSimulNodeType0(double length_anc, 
				     int* seq_anc, int* seq_obs,
				     int* res_seq);
 void printContextDepParams(ofstream& out);
  //void updateContextDepHyperParamMuo(double mutratemin, double mutratemax, bool inverse_prior);
  //void updateContextDepHyperParamSigmao(bool inverse_prior);
  double proposeContextDepModelUpdate(double lambda, bool& ok);
  double proposeModelUpdate(double lambda, bool& ok);

  void cancelContextDepModelUpdate();
  void readContextDepParams(ifstream& in);
  double averagemeth(double t);
  vector<double> nbmethdist(double t);
  double computeLogRatioLambdaChange(double lambda1, double lambda2);
  // void updateContextDepHyperParamSigmao(double lambda);
  void updateHyperParamSigmao(double lambda);
  void changeLambdaKeepTheta(double lambda1,double lambda2);
  int dimModel();
};

inline double Mutmodel::computeBranchLength(double t1, double t2)
{
  if (t1 > t2) 
    {
      cerr << "Error in Mutmodel::computeBranchLength" << endl;
      int *k=NULL;
      *k=1;
      exit(0);
    }
  if (t2 < 0) 
    {
      return (t2-t1)*_alpha_diff;
    } else if (t1 > 0) {
      return t2-t1;
    } else {
      return t2 - t1*_alpha_diff;
    }
}
  

inline void Mutmodel::set_pat_rate_factor(double fact)
{
  _pat_rate_fact = fact;  
}

inline int Mutmodel::codSeq(int* seq)
{
  int ret_val = 0;
  for (int i=0; i<_ltag; i++)
    {
      ret_val += _codfact[i]*seq[i];
    }
  return ret_val;
}

inline bool Mutmodel::contextdep()
{
  return _contextdep;
}

inline int Mutmodel::nbmutclass()
{
  return _nb_mutclass;
}

inline double Mutmodel::probaswitch(double length, int imut, int jmut)
{
  length *= _pat_rate_fact;
  //  cout << "length " << length << endl;
  //cout << "_switch_rate[0][1] " << _switch_rate[0][1] << endl;
  //cout << "_switch_rate[1][0] " << _switch_rate[1][0] << endl;
  if (_nb_mutclass == 1 && imut==0 && jmut==0)
    {
      return 1;
    } else if (_nb_mutclass == 2) {
      double exp_term = exp(-(_switch_rate[0][1]+_switch_rate[1][0])*length);
      //cout << "exp_term " << exp_term << endl;
      if (jmut==0)
	{
	  double proba = _switch_rate[1][0]/(_switch_rate[1][0]+_switch_rate[0][1]);  
	  if (imut==0) 
	    {
	      proba += (_switch_rate[0][1]/(_switch_rate[1][0]+_switch_rate[0][1]))*exp_term;
	    } else {
	      proba += -(_switch_rate[1][0]/(_switch_rate[1][0]+_switch_rate[0][1]))*exp_term;
	    } 
	  return proba;      
	} else {
	  double proba = _switch_rate[0][1]/(_switch_rate[1][0]+_switch_rate[0][1]);  
	  if (imut==1) 
	    {
	      proba += (_switch_rate[1][0]/(_switch_rate[1][0]+_switch_rate[0][1]))*exp_term;
	    } else {
	      proba += -(_switch_rate[0][1]/(_switch_rate[1][0]+_switch_rate[0][1]))*exp_term;
	    } 
	  return proba;      
	}
      
    } else {
      cout << "Nb of class greater than 2" << endl;
      exit(1);
    }
  
}

inline double Mutmodel::proba(double length, int imut, int from_allele, int to_allele)
{
  //cout << "length " << length << endl;
  length *= _pat_rate_fact;
  if (_unique_rate)
    {
      double proba = 0.5;  
      if (from_allele==to_allele) 
	{
	  proba += 0.5*exp(-2*_mutrate[imut]*length);
	} else {
	  proba -= 0.5*exp(-2*_mutrate[imut]*length);
	}
      return proba;
    } else {
      double exp_term = exp(-(_mutrate_m[imut]+_mutrate_u[imut])*length);
      //cout << "_mutrate_m["<<imut<<"] " << _mutrate_m[imut] << endl;
      //cout << "_mutrate_u["<<imut<<"] " << _mutrate_u[imut] << endl;
      //cout << "exp_term " << exp_term << endl;
      if (to_allele==0)
	{
	  double proba = _mutrate_u[imut]/(_mutrate_m[imut]+_mutrate_u[imut]);  
	  if (from_allele==0) 
	    {
	      proba += (_mutrate_m[imut]/(_mutrate_m[imut]+_mutrate_u[imut]))*exp_term;
	    } else {
	      proba += -(_mutrate_u[imut]/(_mutrate_m[imut]+_mutrate_u[imut]))*exp_term;
	    } 
	  return proba;      
	} else {
	  double proba = _mutrate_m[imut]/(_mutrate_m[imut]+_mutrate_u[imut]);  
	  if (from_allele==1) 
	    {
	      proba += (_mutrate_u[imut]/(_mutrate_m[imut]+_mutrate_u[imut]))*exp_term;
	    } else {
	      proba += -(_mutrate_m[imut]/(_mutrate_m[imut]+_mutrate_u[imut]))*exp_term;
	    } 	  
	  return proba;      
	}
    }
}
 
inline int Mutmodel::nmodal()
{
  return 2;
}

inline void Mutmodel::changemutrate(int imut, double new_rate)
{
  _mutrate[imut] = new_rate;
}

inline void Mutmodel::changeperror(double p_error)
{
  _p_error = p_error;
}

inline double Mutmodel::mutrate(int imut)
{
  return(_mutrate[imut]);
}

inline double Mutmodel::perror()
{
  return(_p_error);
}

inline void Mutmodel::changemutrate_m(int imut, double new_rate)
{
  //cout << "_mutrate_m["<<imut<<"] " << _mutrate_m[imut] << " -> " << new_rate << endl;
  _mutrate_m[imut] = new_rate;  
}

inline double Mutmodel::mutrate_m(int imut)
{
  return(_mutrate_m[imut]);  
}

inline void Mutmodel::changemutrate_u(int imut, double new_rate)
{
  _mutrate_u[imut] = new_rate;  
}

inline double Mutmodel::mutrate_u(int imut)
{
  return(_mutrate_u[imut]);
}

inline bool Mutmodel::unique_rate()
{
  return _unique_rate;
}

inline void Mutmodel::changeswitchrate(double newrate, int imut, int jmut)
{
  double oldrate = _switch_rate[imut][jmut];
  _switch_rate[imut][jmut] = newrate;
  _switch_rate[imut][imut] -= (newrate-oldrate);
}

inline void Mutmodel::set_alpha_diff(double alpha_diff)
{
  _alpha_diff = alpha_diff;
}

inline double Mutmodel::alpha_diff()
{
  return _alpha_diff;
}

#endif
