/* Copyright (C) 2006 Pierre Nicolas
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "params.h"

#include <iostream>
#include <fstream>
#include <cstdlib>
#include <cstdio>

using namespace std;

Params::Params(string paramfilename) 
{
  string word;
  ifstream input (paramfilename.c_str(), ios::in);

  _update_g = true;
  _update_alpha = true;
  _inverse_prior = true;
  _update_perror = 0;
  _update_coalrate = true;
  _update_N = 0;
  _sym_mutrate = false;
  _update_mutrate = true;

  _unique_Nlambda = false;
  _unique_mutmodel = false;
  _no_clusterupdate = false;
  _unique_perror = false;

  _lseq = -1;
  _contextdep_mutmodel = false;
  _simul = false;
  _nb_mutrate = 1;
  _mutratemin = 0.0001;
  _mutratemax = 2;
  _alphamin = 0;
  _alphamax = 1;
  _lambdamin = 0.005;
  _lambda = 1;
  _lambdamax = 2;
  _niter = 50000000;
  _g = 7;
  _gmin = 5;
  _gmax = 10;
  _alpha = 0.05;
  _nb_meth_range = 1;
  _nstemcells_prior_min = 1;
  _nstemcells_prior_max = 100;
  _nstemcells_diff_max = 100;
  _seed = -1;
  _N = -1;
  _nclusters = 1;
  _mu_m = -1;
  _mu_u = -1;
  _mu_m_vect.clear();
  _mu_u_vect.clear();
	
  while (!input.eof()) 
    {
      input >> word;
      if (word[0] == '#') 
	{
	  input.ignore(1000, '\n');
	  cout << "skipping end of line after #" << endl;
	} else {
	  cout << "keyword: \"" << word << "\" "<< endl;
	  if (word == string("mutratemin")) 
	    {
	      input >> _mutratemin;
	    } else if (word == string("mutratemax")) {
	      input >> _mutratemax;
	    } else if (word == string("lambdamin")) {
	      input >> _lambdamin;
	    } else if (word == string("lambdamax")) {
	      input >> _lambdamax;
	    } else if (word == string("gmin")) {
	      input >> _gmin;
	    } else if (word == string("gmax")) {
	      input >> _gmax;
	    } else if (word == string("alphamin")) {
	      input >> _alphamin;
	    } else if (word == string("alphamax")) {
	      input >> _alphamax;
	    } else if (word == string("nstemcells_prior_min")) {
	      input >> _nstemcells_prior_min;
	    } else if (word == string("nstemcells_prior_max")) {
	      input >> _nstemcells_prior_max;
	    } else if (word == string("nstemcells_diff_max")) {
	      input >> _nstemcells_diff_max;
	    } else if (word == string("seed")) {
	      input >> _seed;
	    } else if (word == string("niter")) {
	      input >> _niter;
	    } else if (word == string("nb_mutrate")) {
	      input >> _nb_mutrate;
	    } else if (word == string("N")) {
	      input >> _N;
	    } else if (word == string("lambda")) {
	      input >> _lambda;
	    } else if (word == string("g")) {
	      input >> _g;
	    } else if (word == string("contextdep_mutmodel")) {
	      input >> _contextdep_mutmodel;
	    } else if (word == string("sym_mutrate")) {
	      input >> _sym_mutrate;
	    } else if (word == string("unique_perror")) {
	      input >> _unique_perror;
	    } else if (word == string("nclusters")) {
	      input >> _nclusters;
	    } else if (word == string("simul")) {
	      input >> _simul;
	    } else if (word == string("nsimul")) {
	      input >> _nsimul;
	    } else if (word == string("p_error")) {
	      input >> _p_error;
	    } else if (word == string("mu_m")) {
	      input >> _mu_m;
	    } else if (word == string("mu_u")) {
	      input >> _mu_u;
	    } else if (word == string("alpha")) {
	      input >> _alpha;
	    } else if (word == string("update_alpha")) {
	      input >> _update_alpha;
	    } else if (word == string("update_coalrate")) {
	      input >> _update_coalrate;
	    }  else if (word == string("update_N")) {
	      input >> _update_N;
	    } else if (word == string("update_perror")) {
	      input >> _update_perror;
	    }  else if (word == string("update_mutrate")) {
	      input >> _update_mutrate;
	    } else if (word == string("update_g")) {
	      input >> _update_g;
	    } else if (word == string("inverse_prior")) {
	      input >> _inverse_prior;
	    } else if (word == string("lseq")) {
	      input >> _lseq;
	    } else if (word == string("nb_meth_range")) {
	      input >> _nb_meth_range;
	    } else if (word == string("unique_Nlambda")) {
	      input >> _unique_Nlambda;
	    } else if (word == string("unique_mutmodel")) {
	      input >> _unique_mutmodel;
	    } else if (word == string("no_clusterupdate")) {
	      input >> _no_clusterupdate;
	    } else if (word == string("unique_perror")) {
	      input >> _unique_perror;
	    } else if (word == string("mu_m_vect")) {
	      _mu_m_vect.clear();
	      if (_contextdep_mutmodel) 
		{		  
		  double mu;
		  for (int i=0; i<=_lseq; i++) 
		    {
		      input >> mu;
		      _mu_m_vect.push_back(mu);
		    }
		} else {
		  double mu;
		  for (int i=0; i<_nb_mutrate; i++) 
		    {
		      input >> mu;
		      _mu_m_vect.push_back(mu);
		    }
		}
	    } else if (word == string("mu_u_vect")) {
	      _mu_u_vect.clear();
	      if (_contextdep_mutmodel) 
		{		  
		  double mu;
		  for (int i=0; i<=_lseq; i++) 
		    {
		      input >> mu;
		      _mu_u_vect.push_back(mu);
		    }
		} else {
		  double mu;
		  for (int i=0; i<_nb_mutrate; i++) 
		    {
		      input >> mu;
		      _mu_u_vect.push_back(mu);
		    }		  
		}
	    } else {
	      cout << "unknown word \"" << word << "\" "<< endl;
	    }
	}
    }

  if (_mu_m >= 0 && _mu_m_vect.size()==0)
    {
      if (_contextdep_mutmodel) 
	{		  
	  for (int i=0; i<=_lseq; i++) 
	    {
	      _mu_m_vect.push_back(_mu_m);
	    }
	} else {
	  _mu_m_vect.push_back(_mu_m);
	}
    }
  if (_mu_u >= 0 && _mu_u_vect.size()==0)
    {
      if (_contextdep_mutmodel) 
	{		  
	  for (int i=0; i<=_lseq; i++) 
	    {
	      _mu_u_vect.push_back(_mu_u);
	    }
	} else {
	  _mu_u_vect.push_back(_mu_u);
	}
    }
  input.close();

}

Params::~Params() 
{
}
